
/*
 ******************************************************************************
 * Program: cdbs_sql_gen
 *
 * Purpose: This program reads a CDBS load file(s) and creates an SQL file
 *          which will load the data in the file(s) into the appropriate CDBS
 *          database tables
 *
 * Arguments: file specification for load file - can be list or wildcard spec
 *
 * Inputs: Load file(s)
 *
 * Output: A single SQL file with inserts into the database
 * 
 ******************************************************************************
 */ 
 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "ostdb.h"
#include "loadfile.h"
#include "loadutils.h"

#define MAX_NAME 32
#define MAX_MODE 30
#define ERROR -1
#define OK 0
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

typedef struct {
  char db_name[MAX_NAME];
  int col;
  int quoted;
}KeyList;

typedef struct {
  int mod;
  int sev;
}EquivClass;



/*
 ******************************************************************************
 *
 * Function:     get_delivery_num
 *
 * Purpose:      Determines the CDBS delivery number by querying the database
 *
 *               The following query is performed:
 *
 *               UPDATE delivery_number SET value = value + 1
 *               SELECT value FROM delivery_number
 *               
 *
 * Returns:      delivery number if successful
 *               ERROR if not successful
 *
 ******************************************************************************
*/
 
static int get_delivery_num(void){

  int delivery;

  ostdb_begin_transaction(STDB_CONN_1, "CDBS_SERVER", "CDBS_DB");
  ostdb_query_setup(STDB_CONN_1, "get_delivery_num", 1, STDB_SQL, 
		    "CDBS_SERVER", "CDBS_DB", NULL);
  stdb_bind_int(STDB_CONN_1, "value", &delivery);
  stdb_query(STDB_CONN_1, "update delivery_number set value = value + 1");
  stdb_query(STDB_CONN_1, "select value from delivery_number");

  while(ostdb_get_exec(STDB_CONN_1) == STDB_CONTINUE){
  }

  stdb_close_query(STDB_CONN_1);

  /*
   * If there was an error with the query - rollback
   */      
  if(ostdb_get_status(STDB_CONN_1) == OSTDB_ERROR){
    ostdb_abort_transaction(STDB_CONN_1);
    return(ERROR);
  }else{
    ostdb_end_transaction(STDB_CONN_1);
    return(delivery);
  }
}


/*
 ******************************************************************************
 *
 * Function:     get_new_equiv_class
 *
 * Purpose:      Gets the current equivalence class value from the database, 
 *               increments and updates that value in the database
 *               then returns the new moderate and severe values.
 *
 *               The following query is performed:
 *
 *               SELECT value 
 *               FROM equiv_class_value
 *               UPDATE equiv_class_value SET value = value + 2              
 *
 * Returns:      a pointer to an EquivClass structure containing the new 
 *               equivalence class info if successful
 *               NULL if not successful 
 *
 ******************************************************************************
*/

static EquivClass *get_new_equiv_class (void){

  int equiv_class_val, rec_count;
  EquivClass *new_equiv;
  new_equiv = (EquivClass *)malloc(sizeof(EquivClass));

  /*
   * Query the database for the equivalence class info
   */
  ostdb_query_setup(STDB_CONN_1, "get_new_equiv_class", 1, STDB_SQL, 
		    "CDBS_SERVER", "CDBS_DB", NULL);
  stdb_bind_int(STDB_CONN_1, "value", &equiv_class_val);
  stdb_query(STDB_CONN_1, "select value from equiv_class_value");
  stdb_query(STDB_CONN_1, "update equiv_class_value set value=value+2");
  while(ostdb_get_exec(STDB_CONN_1) == STDB_CONTINUE){
    ;
  }
  rec_count = stdb_record_count(STDB_CONN_1);

  stdb_close_query(STDB_CONN_1);
  
  if ((ostdb_get_status(STDB_CONN_1) != OSTDB_NORMAL) || (rec_count != 1)){
    free(new_equiv);
    return(NULL);
  }else{
    new_equiv->sev = equiv_class_val +1;
    new_equiv->mod = equiv_class_val +2;
    return(new_equiv);
  }
}


/*
 ******************************************************************************
 *
 * Function:     get_old_equiv_class
 *
 * Purpose:      Gets the old equivalence class information based on the 
 *               comparison file and the mode keywords
 *     
 *               The following database query is performed:
 *
 *               SELECT equivalence_class_severe, equivalence_class_moderate
 *               FROM <instrument>_row
 *               WHERE
 *               file_name = <comparison_file_name>
 *               AND
 *               <mode_keyword1> = <mode_value1>
 *               AND
 *               <mode_keyword2> = <mode_value2>
 *               ...
 *               AND
 *               <mode_keywordN> = <mode_valueN>
 *
 * Returns:      a pointer to an EquivClass structure containing the old 
 *               equivalence class info if successful
 *               NULL if not successful 
 *
 ******************************************************************************
 */

static EquivClass *get_old_equiv_class(
			      table *load_table,   /* i: load table          */
		              int    row,          /* i: row in load table   */
			      char  *row_table,    /* i: database table name */
			      char  *compare_file, /* i: comparison file name*/
		              KeyList *mode_keys,  /* i: structure containing 
  			                                 keyword names       */
		              int    num_mode,     /* i: # of mode keywords  */
			      int   *match         /* o: TRUE if a matching
						         row is found        */
			      ){

  char temp[1024];
  int j, sev, mod, rec_count;
  EquivClass *old_equiv;

  old_equiv = (EquivClass *)malloc(sizeof(EquivClass));
  *match = TRUE;

  /*
   * Make sure that an entry exists in the database for the comparison file.
   * If not, it is an error.
   */
  ostdb_query_setup(STDB_CONN_1, "get_old_equiv_class", 1, STDB_SQL, 
		    "CDBS_SERVER", "CDBS_DB", NULL);
  stdb_bind_int(STDB_CONN_1, "rec_count", &rec_count);
  sprintf(temp, "select count(*) rec_count from %s", row_table);
  stdb_query(STDB_CONN_1, temp); 
  sprintf(temp, "where file_name = '%s'", compare_file);
  stdb_query(STDB_CONN_1, temp);
  while(ostdb_get_exec(STDB_CONN_1) == STDB_CONTINUE){
  }

  if (rec_count < 1){
    free(old_equiv);
    printf("Error: the comparison file %s doesn't exist in the database\n",
	   compare_file);
    return(NULL);
  }

  stdb_close_query(STDB_CONN_1);

  /*
   * Query the database for the equivalence class info
   */
  ostdb_begin_transaction(STDB_CONN_1, "CDBS_SERVER", "CDBS_DB");
  ostdb_query_setup(STDB_CONN_1, "get_old_equiv_class", 2, STDB_SQL, 
		    "CDBS_SERVER", "CDBS_DB", NULL);
  stdb_bind_int(STDB_CONN_1, "equivalence_class_severe", &sev);
  stdb_bind_int(STDB_CONN_1, "equivalence_class_moderate", &mod);
  stdb_query(STDB_CONN_1, "select equivalence_class_severe,");
  stdb_query(STDB_CONN_1, "equivalence_class_moderate from");
  sprintf(temp, "%s where file_name='%s'", row_table, compare_file);
  stdb_query(STDB_CONN_1, temp);

  /*
   * Add mode keywords to the query
   */
  for (j=0; j<num_mode; j++){
    if (mode_keys[j].quoted){
      sprintf(temp, "and %s='%s'", mode_keys[j].db_name, 
	      get_tab_cell_bynum(load_table, mode_keys[j].col, row)); 
    }else{
      sprintf(temp, "and %s=%s", mode_keys[j].db_name, 
	      get_tab_cell_bynum(load_table, mode_keys[j].col, row)); 
    }
    stdb_query(STDB_CONN_1, temp);
  }
  
  while(ostdb_get_exec(STDB_CONN_1) == STDB_CONTINUE){
  }

  rec_count = stdb_record_count(STDB_CONN_1);

  stdb_close_query(STDB_CONN_1);

  /*
   * If there was an error with the query - rollback
   */      
  if (ostdb_get_status(STDB_CONN_1) != OSTDB_NORMAL){
    ostdb_abort_transaction(STDB_CONN_1);
    free(old_equiv);
    return(NULL);
  }

  ostdb_end_transaction(STDB_CONN_1);

  if (rec_count < 1){ 

    /*
     * The comparison file has no entry for the given modes.
     * Just return, the new equivalence class values will be used
     */
    printf("\nWarning: No comparison file records matched mode values ");
    printf("for row %-d.\n         New equivalence class values and", row+1);
    printf(" a SEVERE change_level were used\n");
    *match = FALSE;

  }else{ /* We did find a matching record */ 
    if((mod == -1) || (sev == -1)){

      /*
       * The comparison file has no evquivalence class values for the given 
       * modes. We need to get the values from the equiv_class_value table
       * and update the comparison file record in the database 
       * then return those same values for this file.
       */
      old_equiv = get_new_equiv_class();
      
      /*
       * Update the comparison file record with the new equiv class values
       */
      ostdb_begin_transaction(STDB_CONN_1, "CDBS_SERVER", "CDBS_DB");
      ostdb_query_setup(STDB_CONN_1, "get_old_equiv_class", 3, STDB_SQL, 
			"CDBS_SERVER", "CDBS_DB", NULL);
      sprintf(temp, "update %s set equivalence_class_severe = %-d,", 
	      row_table, old_equiv->sev);
      stdb_query(STDB_CONN_1, temp);
      sprintf(temp, "equivalence_class_moderate = %-d where file_name='%s'",
	      old_equiv->mod, compare_file);
      stdb_query(STDB_CONN_1, temp);
      
      /*
       * Add mode keywords to query
       */
      for (j=0; j<num_mode; j++){
	if (mode_keys[j].quoted){
	  sprintf(temp, "and %s='%s'", mode_keys[j].db_name, 
		  get_tab_cell_bynum(load_table, mode_keys[j].col, row)); 
	}else{
	  sprintf(temp, "and %s=%s", mode_keys[j].db_name, 
		  get_tab_cell_bynum(load_table, mode_keys[j].col, row)); 
	}
	stdb_query(STDB_CONN_1, temp);
      }
      
      while(ostdb_get_exec(STDB_CONN_1) == STDB_CONTINUE){
      }
      
      rec_count = stdb_record_count(STDB_CONN_1);
      
      stdb_close_query(STDB_CONN_1);
      
      /*
       * If there was an error with the query - rollback
       */      
      if (ostdb_get_status(STDB_CONN_1) != OSTDB_NORMAL) {
	ostdb_abort_transaction(STDB_CONN_1);
	free(old_equiv);
	return(NULL);
      }else{
	ostdb_end_transaction(STDB_CONN_1);
	printf("\nWarning: The matching comparison file record has no ");
	printf("equivalence class values\n         for row %-d.", row+1);
	printf(" New equivalence class values were used\n");
      }
    }else{   /* there is a matching entry */
      old_equiv->mod = mod;
      old_equiv->sev = sev;
    }
  }
  return (old_equiv);
}


/*
 ******************************************************************************
 *
 * Function:     set_equiv_class
 *
 * Purpose:      Sets the equivalence class information based on the
 *               level of change.
 *
 * Returns:      a pointer to an EquivClass structure containing the  
 *               equivalence class info.
 *
 ******************************************************************************
*/

static EquivClass *set_equiv_class(
				   char *change_level,
				   char *compare_file,
				   EquivClass *old_equiv,
				   EquivClass *new_equiv
				   ){
  EquivClass *equiv_class;

  equiv_class = (EquivClass *)malloc(sizeof(EquivClass));

  if ((strcmp(change_level, "SEVERE") == 0) ||
      (strcmp(compare_file, "(INITIAL)") == 0)){
    equiv_class->sev = new_equiv->sev;
    equiv_class->mod = new_equiv->mod;
  }else{
    if (strcmp(change_level, "MODERATE") == 0){
      equiv_class->sev = old_equiv->sev;
      equiv_class->mod = new_equiv->mod;
    }else{
      if (strcmp(change_level, "TRIVIAL") == 0){
	equiv_class->sev = old_equiv->sev;
	equiv_class->mod = old_equiv->mod;
      }else{
	printf("CHANGE_LEVEL is not set for all rows in the load file\n");
	return(NULL);
      }
    }
  }

  return(equiv_class);
}


/*
 ******************************************************************************
 *
 * Function:     load_unexpanded
 *
 * Purpose:      Generates SQL to load the contents of an unexpanded load file 
 *               into the database.
 *
 *               For each unexpanded load file, one or more row-level 
 *               statements is generated, followed by a single file-level 
 *               INSERT statement.
 *
 *               The row-level insert has the form:
 *
 *               INSERT  <instrument>_row (file_name,
 *                                         expansion_number
 *                                         <mode_keyword1>,
 *                                         <mode_keyword2>,
 *                                         ...
 *                                         <mode_keywordN>,
 *                                         equivalence_class_severe,
 *                                         equivalence_class_moderate,
 *                                         pedigree,
 *                                         observation_begin_date,
 *                                         observation_end_date,
 *                                         comment)
 *                                 VALUES (<file_name>,
 *                                         0,
 *                                         <mode_value1>,
 *                                         <mode_value2>,
 *                                         ...
 *                                         <mode_valueN>,
 *                                         <equivalence_class_severe>,
 *                                         <equivalence_class_moderate>,
 *                                         <pedigree>,
 *                                         <observation_begin_date>,
 *                                         <observation_end_date>,
 *                                         <comment>)
 *
 *               The file-level INSERT has the form:
 *
 *               INSERT  <instrument>_file (delivery_number,
 *                                          reference_file_type,
 *                                          file_name,
 *                                          expansion_number
 *                                          useafter_date,
 *                                          opus_flag,
 *                                          comparison_file_name,
 *                                          comment,
 *                                          reject_flag)
 *                                  VALUES (delivery_num,
 *                                          <reference_file_type>,
 *                                          <file_name>,
 *                                          0,
 *                                          <useafter_date>,
 *                                          <comparison_file>,
 *                                          <opus_flag>,
 *                                          <comment>,
 *                                          'N')
 *
 * Returns:      OK or ERROR
 *
 ******************************************************************************
*/

static int load_unexpanded(
			 table *load_table, /* i: table containing load data */
			 FILE *sql_file,    /* i: pointer to SQL file        */
			 int delivery       /* i: delivery number            */
			){
  
  char *file_name, *file_type, *instrument, *use_after, *opus_flag; 
  char *compare_file, *comment, *colname, *colval;
  char *pedigree, *begin_date, *end_date, *row_comment;  
  int num_mode, i, row, col;
  KeyList mode_keys[MAX_MODE];
  char row_table[MAX_NAME];
  char file_table[MAX_NAME];
  char *change_level;
  char expansion_num[2];
  int equiv_info, mode_info;
  EquivClass *old_equiv = NULL, *new_equiv = NULL, *equiv_class = NULL;
  int match;

  /* 
   * Get info from load file header
   */
  file_name = get_tab_key(load_table, "FILE_NAME");
  file_type = get_tab_key(load_table, "REFERENCE_FILE_TYPE");
  instrument = get_tab_key(load_table, "INSTRUMENT");

  use_after = get_tab_key(load_table, "USEAFTER_DATE");
  opus_flag = get_tab_key(load_table, "OPUS_FLAG");
  compare_file = get_tab_key(load_table, "COMPARISON_FILE");
  comment = get_tab_key(load_table, "COMMENT");

  /*
   * Determine table names
   */
  sprintf(file_table, "%s_file", LowCase(instrument));
  sprintf(row_table, "%s_row", LowCase(instrument));

  /*
   * Determine if we will want equivalence class info or not
   */
  if (((strcmp(instrument, "SYNPHOT") == 0) && 
       (strcmp(file_type, "COMPTAB") == 0)) ||
      (strcmp(instrument, "MULTI") == 0)){
    equiv_info = FALSE;
  }else{
    equiv_info = TRUE;
  }

  /*
   * Determine if we will need mode info or not
   */
  if ((strcmp(instrument, "SYNPHOT") == 0) && 
      (strcmp(file_type, "THROUGHPUT") != 0)){
    mode_info = FALSE;
  }else{
    mode_info = TRUE;
  }

  /* 
   * Determine column names for mode keywords
   */
  num_mode = 0;
  for (col=0; col<ncol_table(load_table); col++){
    colname = get_tab_colname(load_table, col);
    if ((strcmp(colname, "CHANGE_LEVEL") != 0) &&
	(strcmp(colname, "EXPANSION_NUMBER") != 0) &&
	(strcmp(colname, "PEDIGREE") != 0) &&
	(strcmp(colname, "OBSERVATION_BEGIN_DATE") != 0) &&
	(strcmp(colname, "OBSERVATION_END_DATE") != 0) &&
	(strcmp(colname, "COMMENT") != 0)){
      strcpy(mode_keys[num_mode].db_name, LowCase(colname));
      mode_keys[num_mode].col = col;
      mode_keys[num_mode].quoted = 
	QuotedField("CDBS_SERVER", "CDBS_DB", row_table, 
		    mode_keys[num_mode].db_name);
      num_mode++;
    }
  }

  /* 
   * Do row level inserts
   */
  for (row=0; row<nrow_table(load_table); row++){

    /*
     * get row level fixed keyword values
     */
    pedigree = get_tab_cell(load_table, "PEDIGREE", row); 
    begin_date = get_tab_cell(load_table, "OBSERVATION_BEGIN_DATE", row); 
    end_date = get_tab_cell(load_table, "OBSERVATION_END_DATE", row); 
    row_comment = get_tab_cell(load_table, "COMMENT", row);
    strcpy(expansion_num, "0");      /* set to zero for unexpanded files */

    /*
     * Determine equivalence class info
     */
    if (equiv_info){
      if (strcmp(compare_file, "(INITIAL)") != 0){
	if ((old_equiv = get_old_equiv_class(load_table, row, row_table, 
					     compare_file, mode_keys, 
					     num_mode, &match)) == NULL)
	{
	  printf("Error obtaining comparison file equivalence class values\n");
	  return(ERROR);
	}
      }

      if ((new_equiv = get_new_equiv_class()) == (EquivClass *)(NULL)){
	printf("Error obtaining new equivalence class values\n");
	return(ERROR);
      }

      /*
       * Assign new equivalence class values based on severity
       */
      if (match){
	change_level = get_tab_cell(load_table, "CHANGE_LEVEL", row);
	equiv_class = set_equiv_class(change_level, compare_file, old_equiv, 
				      new_equiv);
      }else{
	equiv_class = set_equiv_class("SEVERE", compare_file, old_equiv, 
				      new_equiv);
      }

      if (equiv_class == NULL){
	printf("An ERROR occured while determining equivalence class info\n");
	return(ERROR);
      }

      if (new_equiv != NULL){
	free(new_equiv);
      }
      if (old_equiv != NULL){
	free(old_equiv);
      }

    }

    /*
     * Do INSERT part
     */
    fprintf(sql_file, "INSERT %s\n\t( file_name", row_table);
    fprintf(sql_file, ",\n\t expansion_number");

    if (mode_info){
      for (i=0; i<num_mode; i++){
	colval = get_tab_cell_bynum(load_table, mode_keys[i].col, row);
	if (colval[0] != '\0'){
	  fprintf(sql_file, ",\n\t %s", mode_keys[i].db_name);
	}
      }
    }

    if(equiv_info){
      fprintf(sql_file, ",\n\t equivalence_class_severe");
      fprintf(sql_file, ",\n\t equivalence_class_moderate");
    }

    if ((pedigree != NULL) && (pedigree[0] != '\0')){
      fprintf(sql_file, ",\n\t pedigree");
    } 
    if ((begin_date != NULL) && (begin_date[0] != '\0')){
      fprintf(sql_file, ",\n\t observation_begin_date");
    }
    if ((end_date != NULL) && (end_date[0] != '\0')){
      fprintf(sql_file, ",\n\t observation_end_date");
    }
    if ((row_comment != NULL) && (row_comment[0] != '\0')){
      fprintf(sql_file, ",\n\t comment");
    }
    
    /*
     * Do VALUES part
     */
    fprintf(sql_file, ")\nVALUES ('%s'",file_name);
    fprintf(sql_file, ",\n\t%s", expansion_num); 

    if (mode_info){
      for (i=0; i<num_mode; i++){
	colval = get_tab_cell_bynum(load_table, mode_keys[i].col, row);
	if (colval[0] != '\0'){
	  if (mode_keys[i].quoted){
	    fprintf(sql_file, ",\n\t'%s'", colval); 
	  }else{
	    fprintf(sql_file, ",\n\t%s", colval); 
	  }
	}
      }
    }

    if(equiv_info){
      fprintf(sql_file, ",\n\t%-1d", equiv_class->sev);
      fprintf(sql_file, ",\n\t%-1d", equiv_class->mod);
    }

    if ((pedigree != NULL) && (pedigree[0] != '\0')){
      fprintf(sql_file, ",\n\t'%s'", pedigree);
    } 
    if ((begin_date != NULL) && (begin_date[0] != '\0')){
      fprintf(sql_file, ",\n\t'%s'", begin_date); 
    }
    if ((end_date != NULL) && (end_date[0] != '\0')){
      fprintf(sql_file, ",\n\t'%s'", end_date); 
    }
    if ((row_comment != NULL) && (row_comment[0] != '\0')){
      fprintf(sql_file, ",\n\t'%s'", row_comment); 
    }
    fprintf(sql_file, ")\nGO\n");
  }

  /*
   * Write out SQL to insert file-level record
   */

  /*
   * Do INSERT part
   */
  fprintf(sql_file, "INSERT %s\n\t", file_table);
  fprintf(sql_file, "(delivery_number,\n\t");
  fprintf(sql_file, "reference_file_type,\n\t");
  fprintf(sql_file, "file_name,\n\t");
  fprintf(sql_file, "expansion_number,\n\t");

  if ((use_after != NULL) && (use_after[0] != '\0')){
    fprintf(sql_file, "useafter_date,\n\t");
  }

  fprintf(sql_file, "comparison_file_name,\n\t");
  fprintf(sql_file, "opus_flag,\n\t");
  fprintf(sql_file, "comment,\n\t");
  fprintf(sql_file, "reject_flag)\n");

  /*
   * Do VALUES part
   */
  fprintf(sql_file, "VALUES (%-1d,\n\t", delivery);
  fprintf(sql_file, "'%s',\n\t", file_type);
  fprintf(sql_file, "'%s',\n\t", file_name);
  fprintf(sql_file, "0,\n\t");                  /* expansion number      */
  
  if ((use_after != NULL) && (use_after[0] != '\0')){
    fprintf(sql_file, "'%s',\n\t", use_after);  /* optional for nonopus */
  }
  fprintf(sql_file, "'%s',\n\t", compare_file);
  if ((opus_flag != NULL) && (opus_flag[0] != '\0')){
    fprintf(sql_file, "'%s',\n\t", opus_flag);
  }else{
    fprintf(sql_file, "'N',\n\t");              /* "N" for nonopus      */
  }
  fprintf(sql_file, "'%s',\n\t", comment);
  fprintf(sql_file, "'N')\nGO\n");              /* reject flag           */

  return(OK);
}


/*
 ******************************************************************************
 *
 * Function:     load_expanded
 *
 * Purpose:      Generates SQL to load the contents of an expanded load file 
 *               into the database. 
 *
 *               For each expansion number, there is a row-level
 *               INSERT statement.
 *
 *               The row level insert statement is of the form:
 *               INSERT  <instrument>_row (file_name,
 *                                         expansion_number,
 *                                         <mode_keyword1>,
 *                                         <mode_keyword2>,
 *                                         ...
 *                                         <mode_keywordN>,
 *                                         equivalence_class_severe,
 *                                         equivalence_class_moderate,
 *                                         pedigree,
 *                                         observation_begin_date,
 *                                         observation_end_date,
 *                                         comment)
 *                                 VALUES (<file_name>,
 *                                         <expansion_number>,
 *                                         <mode_value1>,
 *                                         <mode_value2>,
 *                                         ...
 *                                         <mode_valueN>,
 *                                         <equivalence_class_severe>,
 *                                         <equivalence_class_moderate>,
 *                                         <pedigree>,
 *                                         <observation_begin_date>,
 *                                         <observation_end_date>,
 *                                         <comment>)
 *
 *               There will also be a file-level insert statement generated
 *               for each expansion number of the form:
 *
 *               INSERT  <instrument>_file (delivery_number,
 *                                          reference_file_type,
 *                                          file_name,
 *                                          expansion_number,
 *                                          useafter_date,
 *                                          comparison_file_name,
 *                                          opus_flag,
 *                                          comment,
 *                                          reject_flag)
 *                                  VALUES (delivery_num,
 *                                          <reference_file_type>,
 *                                          <file_name>,
 *                                          <expansion_number>,
 *                                          <useafter_date>,
 *                                          <comparison_file>,
 *                                          <opus_flag>,
 *                                          <comment>,
 *                                          'N')
 *
 *               Values in <> come from the load file, others are determined 
 *               in this software.
 *
 *
 * Returns:      OK or ERROR
 *
 ******************************************************************************
*/

static int load_expanded(
			 table *load_table, /* i: table containing load data */
			 FILE *sql_file,    /* i: pointer to SQL file        */
			 int delivery       /* i: delivery number            */
			){

  char *file_name, *file_type, *instrument, *use_after, *opus_flag;
  char *compare_file, *comment, *colname, *colval; 
  char * expansion_num, *pedigree, *begin_date, *end_date, *row_comment;  
  int num_mode, i, row, col;
  KeyList mode_keys[MAX_MODE];
  char row_table[MAX_NAME];
  char file_table[MAX_NAME];
  char *change_level;
  EquivClass *old_equiv = NULL, *new_equiv = NULL, *equiv_class = NULL;
  int match;

  match = TRUE;

  /* 
   * Get info from load file header
   */
  file_name = get_tab_key(load_table, "FILE_NAME");
  file_type = get_tab_key(load_table, "REFERENCE_FILE_TYPE");
  instrument = get_tab_key(load_table, "INSTRUMENT");
  use_after = get_tab_key(load_table, "USEAFTER_DATE");
  opus_flag = get_tab_key(load_table, "OPUS_FLAG");
  compare_file = get_tab_key(load_table, "COMPARISON_FILE");
  comment = get_tab_key(load_table, "COMMENT");

  /*
   * Determine table names
   */
  sprintf(file_table, "%s_file", LowCase(instrument));
  sprintf(row_table, "%s_row", LowCase(instrument));

  /* 
   * Determine column names for mode keywords
   */
  num_mode = 0;
  for (col=0; col<ncol_table(load_table); col++){
    colname = get_tab_colname(load_table, col);
    if ((strcmp(colname, "CHANGE_LEVEL") != 0) &&
	(strcmp(colname, "EXPANSION_NUMBER") != 0) &&
	(strcmp(colname, "PEDIGREE") != 0) &&
	(strcmp(colname, "OBSERVATION_BEGIN_DATE") != 0) &&
	(strcmp(colname, "OBSERVATION_END_DATE") != 0) &&
	(strcmp(colname, "COMMENT") != 0)){
      strcpy(mode_keys[num_mode].db_name, LowCase(colname));
      mode_keys[num_mode].col = col;
      mode_keys[num_mode].quoted = 
	QuotedField("CDBS_SERVER", "CDBS_DB", row_table, 
		    mode_keys[num_mode].db_name);	  
      num_mode++;
    }
  }

  /* 
   * Do row level inserts
   */
  for (row=0; row<nrow_table(load_table); row++){
    /*
     * get row level fixed keyword values
     */
    pedigree = get_tab_cell(load_table, "PEDIGREE", row); 
    begin_date = get_tab_cell(load_table, "OBSERVATION_BEGIN_DATE", row); 
    end_date = get_tab_cell(load_table, "OBSERVATION_END_DATE", row); 
    row_comment = get_tab_cell(load_table, "COMMENT", row); 
    expansion_num = get_tab_cell(load_table, "EXPANSION_NUMBER", row);

    /*
     * Determine equivalence class info
     */
    if (strcmp(compare_file, "(INITIAL)") != 0){
      if ((old_equiv = get_old_equiv_class(load_table, row, row_table, 
					   compare_file, mode_keys, 
					   num_mode, &match)) == NULL){
	printf("Error obtaining comparison file equivalence class values\n");
	return(ERROR);
      }
    }
    
    if ((new_equiv = get_new_equiv_class()) == NULL){
      printf("Error obtaining new equivalence class values\n");
      return(ERROR);
    }
    
    /*
     * Assign new equivalence class values based on severity
     */
    if (match){
      change_level = get_tab_cell(load_table, "CHANGE_LEVEL", row);
      equiv_class = set_equiv_class(change_level, compare_file, old_equiv, 
				    new_equiv);
    }else{
      equiv_class = set_equiv_class("SEVERE", compare_file, old_equiv, 
				    new_equiv); 
    }

    if (equiv_class == NULL){
      printf("An ERROR occured while determining equivalence class info\n");
      return(ERROR);
    }

    if (new_equiv != NULL){
      free(new_equiv);
    }
    if (old_equiv != NULL){
      free(old_equiv);
    }

    /*
     * Do INSERT part
     */
    fprintf(sql_file, "INSERT %s\n\t( file_name", row_table);
    fprintf(sql_file, ",\n\t expansion_number");
    
    for (i=0; i<num_mode; i++){
      colval = get_tab_cell_bynum(load_table, mode_keys[i].col, row);
      if (colval[0] != '\0'){
	fprintf(sql_file, ",\n\t %s", mode_keys[i].db_name);
      }
    }
    
    fprintf(sql_file, ",\n\t equivalence_class_severe");
    fprintf(sql_file, ",\n\t equivalence_class_moderate");


    if ((pedigree != NULL) && (pedigree[0] != '\0')){
      fprintf(sql_file, ",\n\t pedigree");
    } 
    if ((begin_date != NULL) && (begin_date[0] != '\0')){
      fprintf(sql_file, ",\n\t observation_begin_date");
    }
    if ((end_date != NULL) && (end_date[0] != '\0')){
      fprintf(sql_file, ",\n\t observation_end_date");
    }
    if ((row_comment != NULL) && (row_comment[0] != '\0')){
      fprintf(sql_file, ",\n\t comment");
    }
    
    /*
     * Do VALUES part
     */
    fprintf(sql_file, ")\nVALUES ('%s'",file_name);
    fprintf(sql_file, ",\n\t%s", expansion_num); 

    for (i=0; i<num_mode; i++){
      colval = get_tab_cell_bynum(load_table, mode_keys[i].col, row);
      if (colval[0] != '\0'){
	if (mode_keys[i].quoted){
	  fprintf(sql_file, ",\n\t'%s'", colval); 
	}else{
	  fprintf(sql_file, ",\n\t%s", colval); 
	  }
      }
    }

    fprintf(sql_file, ",\n\t%-1d", equiv_class->sev);
    fprintf(sql_file, ",\n\t%-1d", equiv_class->mod);

    if ((pedigree != NULL) && (pedigree[0] != '\0')){
      fprintf(sql_file, ",\n\t'%s'", pedigree);
    } 
    if ((begin_date != NULL) && (begin_date[0] != '\0')){
      fprintf(sql_file, ",\n\t'%s'", begin_date); 
    }
    if ((end_date != NULL) && (end_date[0] != '\0')){
      fprintf(sql_file, ",\n\t'%s'", end_date); 
    }
    if ((row_comment != NULL) && (row_comment[0] != '\0')){
      fprintf(sql_file, ",\n\t'%s'", row_comment); 
    }
    fprintf(sql_file, ")\nGO\n");
    
    /*
     * Write out SQL to insert file-level record
     */

    /*
     * Do INSERT part
     */
    fprintf(sql_file, "INSERT %s\n\t", file_table);
    fprintf(sql_file, "(delivery_number,\n\t");
    fprintf(sql_file, "reference_file_type,\n\t");
    fprintf(sql_file, "file_name,\n\t");
    fprintf(sql_file, "expansion_number,\n\t");
    fprintf(sql_file, "useafter_date,\n\t");
    fprintf(sql_file, "comparison_file_name,\n\t");
    fprintf(sql_file, "opus_flag,\n\t");
    fprintf(sql_file, "comment,\n\t");
    fprintf(sql_file, "reject_flag)\n");    

    /*
     * Do VALUES part
     */
    fprintf(sql_file, "VALUES (%-1d,\n\t", delivery);
    fprintf(sql_file, "'%s',\n\t", file_type);
    fprintf(sql_file, "'%s',\n\t", file_name);
    fprintf(sql_file, "%s,\n\t", expansion_num);
    fprintf(sql_file, "'%s',\n\t", use_after);
    fprintf(sql_file, "'%s',\n\t", compare_file);
    if((opus_flag == NULL) || (opus_flag[0] == '\0')){
      fprintf(sql_file, "'N',\n\t");
    }else{
      fprintf(sql_file, "'%s',\n\t", opus_flag);
    }
    fprintf(sql_file, "'%s',\n\t", comment);
    fprintf(sql_file, "'N')\nGO\n");       /* reject flag           */
  }
  return(OK);
}


int main(int argc, char **argv){

  table *load_table;
  int delivery;
  char sql_filename[256];
  FILE *sql_file;
  int i;
  char *cdbs_db;
  char *cdbs_server;
  char auth_server[50];
  char password[50];

  if (argc < 2){
    printf("Must supply load file name as argument\n");
    return(ERROR);
  }

  /*
   * Make sure CDBS_SERVER and CDBS_DB environment variables are set
   */
  if ((cdbs_server = getenv("CDBS_SERVER")) == NULL){
    printf("CDBS_SERVER must be set to the server name\n");
    return (ERROR);
  }

  if ((cdbs_db = getenv("CDBS_DB")) == NULL){
    printf("CDBS_DB must be set to the CDBS database name\n");
    return (ERROR);
  }

  /*
   * Need to get password for calwrite user. We will fake out 
   * stdb_authenticate by passing it the server with CALWRITE appended
   * onto it. This will give us the password for CALWRITE from an
   * authorized user's password file.
   */
  sprintf(auth_server, "%sCALWRITE", cdbs_server);
  stdb_authenticate(auth_server, cuserid(NULL), password);

  ostdb_setup(STDB_CONN_1, STDB_USERNAME, "calwrite");
  ostdb_setup(STDB_CONN_1, STDB_PASSWORD, password);
  ostdb_setup(STDB_CONN_1, STDB_USERNAME, "calwrite");
  ostdb_setup(STDB_CONN_1, STDB_PASSWORD, password);

  /*
   * Get delivery number to use in file name
   */
  if ((delivery = get_delivery_num()) == ERROR){
    printf ("Error getting next delivery number");
    return(ERROR);
  } 

  /*
   * Open the output SQL script and write initial SQL
   */
  sprintf(sql_filename, "cdbs_delivery%-1d.sql", delivery);
  sql_file = fopen(sql_filename, "w");
  fprintf(sql_file, "USE %s\n", cdbs_db);
  fprintf(sql_file, "GO\n");

  for (i=1; i<argc; i++){
    /* 
     * Read in the load file into an internal table
     */
    printf("Processing %s ...\n",argv[i]); 
    if ((load_table = read_loadfile(argv[i])) == NULL){
      printf("Error reading the load file into internal table\n");
      return(ERROR);
    }
    
    /*
     * Check if this is an expanded or unexpanded load file and
     * process accordingly
     */
    if (get_tab_cell(load_table, "EXPANSION_NUMBER", 0) == NULL){
      if (load_unexpanded(load_table, sql_file, delivery) == ERROR){
	printf("Error processing the file: %s\n", argv[1]);
	return(ERROR);
      }
    }else{
      if (load_expanded(load_table, sql_file, delivery) == ERROR){
	printf("Error processing the file: %s\n", argv[1]);
	return(ERROR);
      }
    }
    free_table(load_table);
  }

  /*
   * Finish up
   */
  fclose(sql_file);
  printf("\nProcessing complete -- %s generated\n", sql_filename);
    
  return(OK);  

}
