#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "config.h"
#include "error.h"
#include "cdbslib.h"
#include "loadfile.h"
#include "template.h"

#define  NFIELDS	5
#define	 NRANGES	1000
#define  NPAIR		2
#define  SZ_VAL		63
#define  CERTIFY_INDEF	HUGE_VAL

#define  LOGLIST	"f|t|no|yes"
#define	 KEYLIST	"header|group|column"

#define	 HEADER		1
#define	 GROUP		2
#define	 COLUMN		3

#define  TYPELIST	"integer|real|logical|double|character"

#define  INTEGER	1
#define  REAL		2
#define  LOGICAL	3
#define  DOUBLE		4
#define  CHARACTER	5

#define	 OPTLIST	"exclude|optional|present|required"

#define  EXCLUDE	1
#define  OPTIONAL	2
#define  PRESENT	3
#define  REQUIRED	4

#define  FUNCLIST	"pedigree|anydate|slashdate|sybdate"

#define  PEDIGREE	1
#define  ANYDATE	2
#define  SLASHDATE	3
#define  SYBDATE	4

#define  PEDIGREE_LIST	"inflight|ground|model|dummy"

static char		*func_error;
static FILE		*logfile;

static int		anydate_rule (char *);
static int		cmp_num_rule (value, value);
static int		cmp_str_rule (value, value);
static int		cvt_dblval_rule (char *, double *);
static int		cvt_intval_rule (char *, double *);
static void		error_rule (reffile *, tpl_record *, char *);
static void		listval_rule (rule_rec *, char *);
static int		match_log_rule (rule_rec *, char *);
static int		match_num_rule (rule_rec *, char *);
static int		match_string_rule (rule_rec *, char *);
static int		pedigree_rule (char *);
static int		search_rule (rule_rec *, value, cmpfunc);
static void		sort_rule (rule_rec *, cmpfunc);
static int		slashdate_rule (char *);
static int		sybdate_rule (char *);

/* ANYDATE_RULE -- Check a date in slash or sybase format for correctness
 *
 * i: str	date string
 */

static int anydate_rule (char *str)
{
    if (get_sybase_date (str) < 0.0 &&
        get_slash_date (str) < 0.0 ) {
	func_error = "is not a legal format date";
	return (NO);
    }

    return (YES);
}

/* BODY_RULE -- Return YES if rule accesses the body of a file
 *
 * i: temp	template record containing rule
 */

int body_rule (tpl_record *temp)
{
    rule_rec	*rule = temp->rule;

    return (rule->keytype != HEADER ? YES : NO);
}

/* CHECK_RULE -- Check calibration data item against rule
 *
 * i: temp	template record containing rule
 * i: ref	reference file descriptor
 * i: row	row within reference file
 */

void check_rule (tpl_record *temp, reffile *ref, int row)
{
    char	val[SZ_VAL+1], name[SZ_FNAME+1];
    rule_rec	*rule = temp->rule;
    int		status, ic;

    /* Set skip switch for next time this function is called */

    if (rule->keytype == HEADER)
	temp->skip = YES;

    /* Read value from calibration file */

    if (rule->keytype == HEADER) {
	status = get_hdr_reffile (ref, rule->field, val, SZ_VAL);
    } else {
	status = get_body_reffile (ref, rule->field, row, val, SZ_VAL);
    }

    /* Run through the list of possible error conditions.
     * First check the presence field.
     */

    switch (rule->presence) {
    case EXCLUDE:
        temp->skip = YES;
        if (status != ERR)
	     error_rule (ref, temp, "field should not be present");
	return;

    case OPTIONAL:
        break;

    case PRESENT:
    case REQUIRED:
        if (status == ERR) {
	     temp->skip = YES;
	     error_rule (ref, temp, "field is missing");
	     return;
	}
        break;
    }

    /* A blank value is always a legal value for a field
     * unless the field is required
     */

    ic = first_char (val);
    if (val[ic] == EOS) {
	if (rule->presence == REQUIRED)
	     error_rule (ref, temp, "field is blank");

	return;
    }

    /* If existence test passed, compare value to values in rule */

    strlcase (val);
    strclean (val);

    if (rule->function != NULL) {
	status = (*rule->function) (val);

    } else if (rule->datatype == CHARACTER) {
	status = match_string_rule (rule, val);

    } else if (rule->datatype == LOGICAL) {
	status = match_log_rule (rule, val);
	    
    } else {
	status = match_num_rule (rule, val);
    }

    /* Print message if value does not match */

    if (status == NO) {
        error_rule (ref, temp, "");
	listval_rule (rule, val);
    }
}

/* CMP_NUM_RULE -- Compare two numeric values
 *
 * i: val1	first value
 * i: val2	second value
 */

static int cmp_num_rule (value val1, value val2)
{
    if (val1.num < val2.num)
	return (-1);

    if (val1.num > val2.num)
	return (1);

    return (0);
}

/* CMP_STR_RULE -- Compare two string values 
 * 
 * i: val1	first value
 * i: val2	second value
 */

static int cmp_str_rule (value val1, value val2)
{
    return (strcmp (val1.str, val2.str));
}


/* CVT_DBLVAL_RULE -- Convert a text string representing a double value
 *
 * i: val	text string representing value
 * o: num	double precision value
 */

static int cvt_dblval_rule (char *val, double *num)
{
    char	*rest;
    int		ic;

    if (strcmp (val, "indef") == 0) {
	*num = CERTIFY_INDEF;

    } else {
	*num = strtod (val, &rest);
	ic = first_char (rest);
	if (rest[ic] != EOS)
	    return (NO);
    }

    return (YES);
}

/* CVT_INTVAL_RULE -- Convert a text string representing an integer value
 *
 * i: val	text string representing value
 * o: num	double precision value
 */

static int cvt_intval_rule (char *val, double *num)
{
    char	*rest;
    int		ic;

    if (strcmp (val, "indef") == 0) {
	*num = CERTIFY_INDEF;

    } else {
	*num = strtol (val, &rest, 0);
	ic = first_char (rest);
	if (rest[ic] != EOS)
	    return (NO);
    }

    return (YES);
}

/* ERROR_RULE -- Print the error message if the rule does not check out
 *
 * i: ref	reference file descriptor
 * i: temp	template record containing rule
 * i: text	explanatory text to print
 */

static void error_rule (reffile *ref, tpl_record *temp, char *text)
{
    int		row;
    rule_rec	*rule = temp->rule;
    char	message[SZ_LINE+1];

    /* Print field name */

    if (temp->skip == NO && inbody_reffile(ref)) {
        row = row_reffile(ref);
	sprintf (message, "Error in %s[%d]: ", rule->field, row);

    } else {
	sprintf (message, "Error in %s: ", rule->field);
    }

    cdbs_message (WARN_MSG, message);

    if (text[0] != EOS) {
        cdbs_message (INFO_MSG, text);
	cdbs_message (INFO_MSG, "\n");
    }
}

/* FREE_RULE -- Free memory used by the rule
 *
 * i: rule	rule descriptor
 */

void free_rule (rule_rec *rule)
{
    int		ival;

    if (rule->datatype == CHARACTER || rule->datatype == LOGICAL) {
	for (ival = 0; ival < rule->nval; ival ++) {
	    if (rule->min[ival].str != rule->max[ival].str)
		free (rule->max[ival].str);

	    free (rule->min[ival].str);
	}
    }

    if (rule->min != NULL)
	free (rule->min);

    if (rule->max != NULL)
	free (rule->max);

    free (rule);
}

/* LISTVAL_RULE -- List the value and legal alternatives as the error text
 * i: rule	rule descriptor
 * i: val	value that failed check
 */

static void listval_rule (rule_rec *rule, char *val)
{
    int		ival, jval;
    char	message[SZ_LINE+1];

    if (rule->function != NULL) {
        sprintf (message, "\"%s\"\n%s\n", val, func_error);
	cdbs_message (INFO_MSG, message);

    } else if (strmatch (val, WILDCARD_LIST) != 0) {
	sprintf (message, 
		 "\"%s\" is not a legal value.\nMay need to run expload\n", 
		 val);
	cdbs_message (INFO_MSG, message);

    } else if (rule->nval == 0) {
	/* Zero length list, so must have wrong type */
	sprintf (message, "\"%s\" is wrong type\n", val);
	cdbs_message (INFO_MSG, message);

    } else {
	/* Value not in list */
	/* Print the faulty value and the list of values in the rule */

	if (rule->datatype == CHARACTER || rule->datatype == LOGICAL) {
	    sprintf (message, "\"%s\" is not in ", val);
	    cdbs_message (INFO_MSG, message);

	    jval = 4;
	    for (ival = 0; ival < rule->nval; ival ++) {
	        safecopy (message, rule->min[ival].str, SZ_LINE);

		if (rule->min[ival].str != rule->max[ival].str) {
		    safecat (message, ":", SZ_LINE);
		    safecat (message, rule->max[ival].str, SZ_LINE);
		}

		if (++ jval % 8 == 0) {
		    safecat (message, ",\n", SZ_LINE);
		} else if (ival < rule->nval - 1){
		    safecat (message, ", ", SZ_LINE);
		}

		cdbs_message (INFO_MSG, message);
	    }

	} else {
	    if (rule->datatype == INTEGER) {
		sprintf (message, "%d is not in ", atoi (val));
	    } else {
		sprintf (message, "%g is not in ", atof (val));
	    }
	    cdbs_message (INFO_MSG, message);

	    jval = 4;
	    for (ival = 0; ival < rule->nval; ival ++) {
	        safecopy (message, dtoa (rule->min[ival].num), SZ_LINE);

		if (rule->min[ival].num != rule->max[ival].num) {
		    safecat (message, ":", SZ_LINE);
		    safecat (message, dtoa (rule->max[ival].num), SZ_LINE);
		}

		if (++ jval % 8 == 0) {
		    safecat (message, ",\n", SZ_LINE);
		} else if (ival < rule->nval - 1){
		    safecat (message, ", ", SZ_LINE);
		}

		cdbs_message (INFO_MSG, message);
	    }
	}

	cdbs_message (INFO_MSG, "\n");
    }
}

/* MATCH_LOG_RULE -- See if a logical value matches the value in a rule
 *
 * i: rule	rule descriptor
 * i: val	value to be matched
 */

static int match_log_rule (rule_rec *rule, char * val)
{
    char	*torf;
    value	tempval;

    torf = nospaces (val);

    if (strmatch (torf, LOGLIST) == 0)
	return (NO);

    tempval.str = torf;
    return (search_rule (rule, tempval, cmp_str_rule));
}

/* MATCH_NUM_RULE -- See if a numeric value matches a number in the rule
 *
 * i: rule	rule descriptor
 * i: val	value to be matched
 */

static int match_num_rule (rule_rec *rule, char * val)
{
    value	tempval;
    double	dval;

    if (rule->datatype == INTEGER) {
	if (cvt_intval_rule (val, &dval) == NO)
	    return (NO);

    } else {
	if (cvt_dblval_rule (val, &dval) == NO)
	    return (NO);
    }

    tempval.num = dval;
    return (search_rule (rule, tempval, cmp_num_rule));
}

/* MATCH_STRING_RULE -- See if string value matches a string in the rule
 *
 * i: rule	rule descriptor
 * i: val	value to be matched
 */

static int match_string_rule (rule_rec *rule, char * val)
{
    value	tempval;

    tempval.str = val;
    return (search_rule (rule, tempval, cmp_str_rule));
}

/* MERGE_RULE -- Merge overlapping ranges
 *
 * i: rule	rule descriptor
 * i: compare	comparison function
 */

static void merge_rule (rule_rec *rule, cmpfunc compare)
{
    int		ival, jval;

    jval = 1;
    for (ival = 1; ival < rule->nval; ival ++) {
        if (ival > jval) {
	    rule->min[jval] = rule->min[ival];
	    rule->max[jval] = rule->max[ival];
	}

	if ((*compare) (rule->min[jval], rule->max[jval-1]) > 0) {
	    jval ++;
	} else if ((*compare) (rule->max[jval], rule->max[jval-1]) > 0) {
	    rule->max[jval-1] = rule->max[jval];
	}
    }

    rule->nval = jval;
}

/* PARSE_RULE -- Parse a string to create a new rule
 *
 * u: buf	string to be parsed
 */

rule_rec * parse_rule (char *buf)
{
    int		ival;
    double	dval;
    char 	*field[NFIELDS], *range[NRANGES], *pair[NPAIR];
    rule_rec 	*rule;
    
    /* Allocate structure and initialize its fields */

    rule = (rule_rec *) malloc (sizeof (rule_rec));

    rule->nval = 0;
    rule->min = NULL;
    rule->max = NULL;
    rule->sorted = NO;
    rule->function = NULL;

    /* Split the buffer on whitespace and set fields that do not
     * need further splitting: the keyname, key type, and datatype.
     */

    strlcase (buf);
    strsplit (' ', buf, field, NFIELDS);
    
    safecopy (rule->field, field[0], SZ_FIELD);
    rule->keytype = strabbrev (field[1], KEYLIST);
    rule->datatype = strabbrev (field[2], TYPELIST);
    rule->presence = strabbrev (field[3], OPTLIST);

    if (rule->keytype == 0 || rule->datatype == 0 || rule->presence == 0)
        cdbs_error (BADRULE, "%s", rule->field);
    
    if (rule->presence == EXCLUDE && field[4][0] != EOS)
        cdbs_error (BADRULE, "%s", rule->field);

    if (field[4][0] == '&') {
	/* If the value field starts with an ampersand, it means
	 * call a special function to do the checking
	 */

	switch (strmatch (&field[4][1], FUNCLIST)) {
	case PEDIGREE:	
	    rule->function = pedigree_rule;
	    break;
	case ANYDATE:
	    rule->function = anydate_rule;
	    break;
	case SLASHDATE:
	    rule->function = slashdate_rule;
	    break;
	case SYBDATE:
	    rule->function = sybdate_rule;
	    break;
	default:
	    cdbs_error (BADFUNC, "%s", field[4]);
	}

    } else {
	/* Otherwise we have a range or a comma separated list of ranges 
	 * Split into separate ranges
	 */

	strsplit (',', field[4], range, NRANGES);

	/* Count the number of ranges */

	for (ival = 0; range[ival][0] != EOS; ival ++)
	    ;

	/* Allocate memory for the min and max values */

	rule->nval = ival;
	if (ival > 0) {
	    rule->min = (value *) malloc (ival * sizeof (value));
	    rule->max = (value *) malloc (ival * sizeof (value));
	}

	/* Split each range on the colon. If no colon, set the maximum
	 * value to the minimum, treating it as a zero width range
	 */

	for (ival = 0; ival < rule->nval; ival ++) {
	    strsplit (':', range[ival], pair, NPAIR);

	    if (rule->datatype == CHARACTER || rule->datatype == LOGICAL) {
		/* Allocate memory and store string values */

		rule->min[ival].str = strclone (pair[0]);

		if (pair[1][0] == EOS) {
		    rule->max[ival].str = rule->min[ival].str;
		} else {
		    rule->max[ival].str = strclone (pair[1]);
		}

	    } else if (rule->datatype == INTEGER) {
		if (cvt_intval_rule (pair[0], &dval) == YES) {
		    rule->min[ival].num = dval;
		} else {
		    cdbs_error (BADRULE, "%s", pair[0]);
		}

		if (pair[1][0] == EOS) {
		    rule->max[ival].num = rule->min[ival].num;
		} else if (cvt_intval_rule (pair[1], &dval) == YES) {
		    rule->max[ival].num = dval;
		} else {
		    cdbs_error (BADRULE, "%s", pair[1]);
		}

	    } else {
		if (cvt_dblval_rule (pair[0], &dval) == YES) {
		    rule->min[ival].num = dval;
		} else {
		    cdbs_error (BADRULE, "%s", pair[0]);
		}

		if (pair[1][0] == EOS) {
		    rule->max[ival].num = rule->min[ival].num;
		} else if (cvt_dblval_rule (pair[1], &dval) == YES) {
		    rule->max[ival].num = dval;
		} else {
		    cdbs_error (BADRULE, "%s", pair[1]);
		}
	    }
	}
    }

    return (rule);
}

/* PEDIGREE_RULE -- Check a pedigree for validity
 *
 * u: str	string to be checked
 */

static int pedigree_rule (char *str)
{
    char	*dash, *value, *field[4];

    /* Copy string so original value is not altered */

    value = strclone (str);

    /* Remove dash between dates and split string 
     * into blank delimeted fields */

    if ((dash = strchr (value, '-')) != NULL)
	*dash = ' ';

    strsplit (' ', value, field, 4);

    /* Check fields for legal values */

    func_error = NULL;
    if (field[0][0] == EOS) {
	func_error = "pedigree string is blank";

    } else if (strmatch (field[0], PEDIGREE_LIST) == 0) {
	func_error = "is not a legal pedigree string";

    } else if (strcmp (field[0], "inflight") == 0 && field[1][0] == EOS) {
	func_error = "is missing the date field";
 
    } else if (field[1][0] != EOS && get_slash_date (field[1]) < 0.0) {
	func_error = "has an illegal date";

    } else if (field[2][0] != EOS && get_slash_date (field[2]) < 0.0) {
	func_error = "has an illegal date";

    } else if (field[3][0] != EOS) {
	func_error = "has extra info after dates";
    }

    /* Free memory and return status */

    free (value);

    if (func_error == NULL) {
	return (YES);
    } else {
	return (NO);
    }
}

/* SEARCH_RULE -- Search the rule for a match to the value
 *
 * i: rule	rule descriptor
 * i: val	value to be matched
 * i: compare	comparison function
 */

static int search_rule (rule_rec * rule, value val, cmpfunc compare)
{
    int		ival, status;
    int		lo, hi, mid;

    if (rule->nval == 0)
	return (YES);

    if (rule->sorted == NO) {
	/* Linear search */

	status = NO;
	for (ival = 0; ival < rule->nval; ival++) {
	    if ((*compare) (val, rule->min[ival]) >= 0 &&
		(*compare) (val, rule->max[ival]) <= 0   ) {
		status = YES;
		break;
	    }
	}

    } else {
	/* Binary search algorithm that handles equal keys */

	/* Set bounds on portion of array to check.  Find midpoint
	 * and make it the new high or low bound, depending on how
	 * the array value compares with the input
	 */

	lo = -1;
	hi = rule->nval;

	while (hi != lo + 1) {
	    mid = (hi + lo) / 2;
	    if ((*compare) (val, rule->max[mid]) > 0) {
		lo = mid;
	    } else {
		hi = mid;
	    }
	}

	/* Check to see if there is a match */

	mid = hi;
	if (mid == rule->nval) {
	    status = NO;
	} else if ((*compare) (val, rule->min[mid]) >= 0 &&
		   (*compare) (val, rule->max[mid]) <= 0   ) {
	    status = YES;
	} else {
	    status = NO;
	}
    }

    return (status);
}

/* SLASHDATE_RULE -- Check a date in slash format for correctness
 *
 * i: str	date string
 */

static int slashdate_rule (char *str)
{
    if (get_slash_date (str) < 0.0) {
	func_error = "is not a valid slash format date (dd/mm/yy)";
	return (NO);
    }

    return (YES);
}

/* SORT_RULE -- Sort the array of minimum values
 *
 * i: rule	rule descriptor
 * i: compare	comparison function
 */

static void sort_rule (rule_rec *rule, cmpfunc compare)
{
    value	minval, maxval, tempval;
    int		ival, jval;

    /* Insertion sort */

    for (ival = 0; ival < rule->nval - 1; ival ++) {
	minval = rule->min[ival+1];
	maxval = rule->max[ival+1];

	/* Make sure first value in array is smaller than the value
	 * we are sorting so the sort does not go past the start of
	 * the array.
	 */

	if ((*compare) (minval, rule->min[0]) < 0) {
	    tempval = minval;
	    minval = rule->min[0];
	    rule->min[0] = tempval;

	    tempval = maxval;
	    maxval = rule->max[0];
	    rule->max[0] = tempval;
	}

	/* Move all the values larger than the value we are sorting up
	 * so that there is room to insert the sort value
	 */

	for (jval = ival; (*compare) (minval, rule->min[jval]) < 0; jval --) {
	    rule->min[jval+1] = rule->min[jval];
	    rule->max[jval+1] = rule->max[jval];
	}

	/* Insert the sort value */

	rule->min[jval+1] = minval;
	rule->max[jval+1] = maxval;
    }
}

/* STUDY_RULE -- Sort rule values so binary search can be used
 *
 * i: temp	template record containing rule
 */

void study_rule (tpl_record *temp)
{
    rule_rec	*rule = temp->rule;

    if (rule->nval <= 1)
	return;

    if (rule->datatype == CHARACTER || rule->datatype == LOGICAL) {
	sort_rule (rule, cmp_str_rule);
	merge_rule (rule, cmp_str_rule);
    } else {
	sort_rule (rule, cmp_num_rule);
	merge_rule (rule, cmp_num_rule);
    }

    rule->sorted = YES;
}

/* SYBDATE_RULE -- Check a date in sybase format for correctness
 *
 * i: str	date string
 */

static int sybdate_rule (char *str)
{
    if (get_sybase_date (str) < 0.0) {
	func_error = "is not a valid Sybase format date";
	return (NO);
    }

    return (YES);
}
