
/*
 ******************************************************************************
 * Program: comments_sql_gen
 *
 * Purpose: This program reads a CDBS load file and creates an SQL file
 *          which update the comment fields in the appropriate CDBS
 *          database tables
 *
 * Arguments: file specification for load file 
 *
 * Inputs: Load file
 *
 * Output: A single SQL file with updates the database
 * 
 ******************************************************************************
 */ 
 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "ostdb.h"
#include "loadfile.h"
#include "loadutils.h"

#define MAX_NAME 32
#define MAX_MODE 30
#define ERROR -1
#define OK 0
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

typedef struct {
  char db_name[MAX_NAME];
  int col;
  int quoted;
}KeyList;

typedef struct {
  int mod;
  int sev;
}EquivClass;



/*
 ******************************************************************************
 *
 * Function:     NullRowComment
 *
 * Purpose:      Determines if the comment currently in the database is NULL
 *
 * Returns:      TRUE if blank, FALSE if not
 *
 ******************************************************************************
*/

static int NullRowComment( table *load_table, /* i: load file table         */
			   char *filename,     /* i: load file name          */
 			   char *table,        /* i: database table          */
			   KeyList *mode_keys, /* i: mode keywords           */
			   int num_mode,       /* 1: number of mode keywords */
			   int row             /* i: row number              */
			   ){
  char query[256];
  int tptr;
  char *colval;
  int i;
  
  ostdb_query_setup(STDB_CONN_1, "NullRowComment", 1, STDB_SQL, "CDBS_SERVER",
		    "CDBS_DB", NULL);

  stdb_bind_int(STDB_CONN_1, "tptr", &tptr);
  sprintf(query, 
	  "select textptr(comment) tptr from %s where file_name = '%s'",
	  table, filename);
  stdb_query(STDB_CONN_1, query);

  for (i=0; i<num_mode; i++){
    colval = UpCase(get_tab_cell_bynum(load_table, mode_keys[i].col, 
				       row));
    if (colval[0] != '\0'){
      if (mode_keys[i].quoted){
	sprintf(query, "and %s = '%s'", mode_keys[i].db_name, colval); 
      }else{
	sprintf(query, "and %s = %s", mode_keys[i].db_name, colval); 
      }
      stdb_query(STDB_CONN_1, query);
    }
  }

  while(ostdb_get_exec(STDB_CONN_1) == STDB_CONTINUE){
  }

  stdb_close_query(STDB_CONN_1);

  if (tptr == 0){
    return TRUE;
  }else{
    return FALSE;
  }

}



/*
 ******************************************************************************
 *
 * Function:     update_expaned
 *
 * Purpose:      Generates SQL to update the comments in commthe database 
 *               with the values in the updated expanded load file 
 *
 * Returns:      OK or ERROR
 *
 ******************************************************************************
*/

static int update_expanded(
			   table *load_table, /* i: load file table          */
			   FILE *sql_file     /* i: pointer to SQL file      */
			  ){
  
  char *file_name, *file_type, *instrument, *comment, *colname, *colval, 
    *row_comment;  
  char row_table[MAX_NAME];
  char file_table[MAX_NAME];

  /* 
   * Get info from load file header
   */
  file_name = UpCase(get_tab_key(load_table, "FILE_NAME"));
  instrument = UpCase(get_tab_key(load_table, "INSTRUMENT"));
  file_type = UpCase(get_tab_key(load_table, "REFERENCE_FILE_TYPE"));
  comment = get_tab_key(load_table, "COMMENT");
  if (isspace(comment[0]) && !isspace(comment[1])){
    comment++;
  }

  /*
   * Determine table names
   */
  sprintf(file_table, "%s_file", instrument);
  LowCase(file_table);
  sprintf(row_table, "%s_row", instrument);
  LowCase(row_table);

  /*
   * get the row level comment value for the first row
   */
  row_comment = get_tab_cell(load_table, "COMMENT", 0);
    if (isspace(row_comment[0]) && !isspace(row_comment[1])){
    row_comment++;
  }

  /*
   * Generate the update SQL for the row level records
   * All rows for this file will get the same comment as the first row 
   */
  if (NotBlank(row_comment)){
    fprintf(sql_file, "\n\nUPDATE %s\nSET comment = \"%s\"",
	    row_table, TransformQuotes(row_comment));
    fprintf(sql_file, "\nWHERE file_name = '%s'", file_name);

  }else{
    /*
     * If the current comment is NULL and the new comment is
     * blank, don't even update. If the current one is not NULL and the
     * new one is blank then update the database with NULL
     */
    if (!NullRowComment(load_table, file_name, row_table, NULL, 0, 0)){
      fprintf(sql_file, "\n\nUPDATE %s\nSET comment = NULL",
	      row_table);
      fprintf(sql_file, "\nWHERE file_name = '%s'", file_name);

    }   
  }
    
  /*
   * Generate the SQL to update file-level record
   */
  fprintf(sql_file, "\n\nUPDATE %s\nSET comment = \"%s\"", file_table, 
	  TransformQuotes(comment));
  fprintf(sql_file, "\nWHERE file_name = '%s'",file_name);

  return(OK);  
}

/*
 ******************************************************************************
 *
 * Function:     update_unexpaned
 *
 * Purpose:      Generates SQL to update the comments in the database 
 *               with the values in the updated unexpanded load file 
 *
 * Returns:      OK or ERROR
 *
 ******************************************************************************
*/

static int update_unexpanded(
			     table *load_table, /* i: load file table       */
			     FILE *sql_file     /* i: pointer to SQL file   */
			    ){
  
  char *file_name, *file_type, *instrument, *comment, *colname, *colval, 
    *row_comment;  
  int num_mode, i, row, col, mode_info;
  KeyList mode_keys[MAX_MODE];
  char row_table[MAX_NAME];
  char file_table[MAX_NAME];
  int update;

  /* 
   * Get info from load file header
   */
  file_name = UpCase(get_tab_key(load_table, "FILE_NAME"));
  instrument = UpCase(get_tab_key(load_table, "INSTRUMENT"));
  file_type = UpCase(get_tab_key(load_table, "REFERENCE_FILE_TYPE"));

  comment = get_tab_key(load_table, "COMMENT");
  if (isspace(comment[0]) && !isspace(comment[1])){
    comment++;
  }

  /*
   * Determine table names
   */
  sprintf(file_table, "%s_file", instrument);
  LowCase(file_table);
  sprintf(row_table, "%s_row", instrument);
  LowCase(row_table);

  /*
   * Determine if we will need mode info or not
   */
  if ((strcmp(instrument, "SYNPHOT") == 0) && 
      (strcmp(file_type, "THROUGHPUT") != 0)){
    mode_info = FALSE;
  }else{
    mode_info = TRUE;
  }

  /* 
   * Determine column names for mode keywords
   */
  if (mode_info){
    num_mode = 0;
    for (col=0; col<ncol_table(load_table); col++){
      colname = get_tab_colname(load_table, col);
      if ((strcmp(colname, "CHANGE_LEVEL") != 0) &&
	  (strcmp(colname, "EXPANSION_NUMBER") != 0) &&
	  (strcmp(colname, "PEDIGREE") != 0) &&
	  (strcmp(colname, "OBSERVATION_BEGIN_DATE") != 0) &&
	  (strcmp(colname, "OBSERVATION_END_DATE") != 0) &&
	  (strcmp(colname, "COMMENT") != 0)){
	strcpy(mode_keys[num_mode].db_name, LowCase(colname));
	mode_keys[num_mode].col = col;
	mode_keys[num_mode].quoted = 
	  QuotedField("CDBS_SERVER", "CDBS_DB", row_table, 
		      mode_keys[num_mode].db_name);
	num_mode++;
      }
    }
  }

  /* 
   * Do row level updates
   */
  for (row=0; row<nrow_table(load_table); row++){

    /*
     * get the row level comment value
     */
    row_comment = get_tab_cell(load_table, "COMMENT", row);
    if (isspace(row_comment[0]) && !isspace(row_comment[1])){
      row_comment++;
    }

    /*
     * Generate the update SQL
     */
    update = FALSE;
    if (NotBlank(row_comment)){
      fprintf(sql_file, "\n\nUPDATE %s\nSET comment = \"%s\"",
	      row_table, TransformQuotes(row_comment));
      update = TRUE;
    }else{
      /*
       * If the current comment is NULL and the new comment is
       * blank, don't even update. If the current one is not NULL and the
       * new one is blank then update the database with NULL
       */
      if (!NullRowComment(load_table, file_name, row_table, mode_keys, 
			   num_mode, row)){
	fprintf(sql_file, "\n\nUPDATE %s\nSET comment = NULL",
		row_table);
	update = TRUE;
      }   
    }

    if (update){
      fprintf(sql_file, "\nWHERE file_name = '%s'", file_name);

      if (mode_info){
	for (i=0; i<num_mode; i++){
	  colval = UpCase(get_tab_cell_bynum(load_table, mode_keys[i].col, 
					     row));
	  if (colval[0] != '\0'){
	    fprintf(sql_file, "\n\tAND %s = ", mode_keys[i].db_name);
	    if (mode_keys[i].quoted){
	      fprintf(sql_file, "'%s'", colval); 
	    }else{
	      fprintf(sql_file, "%s", colval); 
	    }
	  }
	}
      }
    }
  }

  /*
   * Generate the SQL to update file-level record
   */
  fprintf(sql_file, "\n\nUPDATE %s\nSET comment = \"%s\"", file_table, 
	  TransformQuotes(comment));
  fprintf(sql_file, "\nWHERE file_name = '%s'",file_name);

  return(OK);
}

int main(int argc, char **argv){

  table *load_table;
  char sql_filename[256];
  FILE *sql_file;
  char *cdbs_db;
  char *file_name, *table_name;
  char *cdbs_server;
  char auth_server[50];
  char password[50];


  if (argc < 2){
    printf("Must supply a load file name as argument\n");
    exit(1);
  }

  /*
   * Make sure CDBS_SERVER and CDBS_DB environment variables are set
   */
  if ((cdbs_server = getenv("CDBS_SERVER")) == NULL){
    printf("CDBS_SERVER must be set to the server name\n");
    exit(1);
  }

  if ((cdbs_db = getenv("CDBS_DB")) == NULL){
    printf("CDBS_DB must be set to the CDBS database name\n");
    exit(1);
  }

  /*
   * Need to get password for calwrite user. We will fake out 
   * stdb_authenticate by passing it the server with CALWRITE appended
   * onto it. This will give us the password for CALWRITE from an
   * authorized user's password file.
   */
  sprintf(auth_server, "%sCALWRITE", cdbs_server);
  stdb_authenticate(auth_server, cuserid(NULL), password);

  ostdb_setup(STDB_CONN_1, STDB_USERNAME, "calwrite");
  ostdb_setup(STDB_CONN_1, STDB_PASSWORD, password);
  ostdb_setup(STDB_CONN_1, STDB_USERNAME, "calwrite");
  ostdb_setup(STDB_CONN_1, STDB_PASSWORD, password);

  /* 
   * Read in the load file into an internal table
   */
  printf("Processing %s ...\n",argv[1]); 
  if ((load_table = read_loadfile(argv[1])) == NULL){
    printf("Error reading the load file into internal table\n");
    exit(1);
  }

  /*
   * Open the output SQL script and write initial SQL
   */
  file_name = UpCase(get_tab_key(load_table, "FILE_NAME"));
  sprintf(table_name, "%s_file", UpCase(get_tab_key(load_table, 
						    "INSTRUMENT")));

  sprintf(sql_filename, "comments_%s", argv[1] );
  sql_filename[strcspn(sql_filename, ".")] = '\0';
  strcat(sql_filename, ".sql");

  sql_file = fopen(sql_filename, "w");
  fprintf(sql_file, "USE %s\n", cdbs_db);
  fprintf(sql_file, "GO\n");
  fprintf(sql_file, "BEGIN TRANSACTION\n");
  
  /*
   * Check if this is an expanded or unexpanded load file and
   * process accordingly
   */
  if (get_tab_cell(load_table, "EXPANSION_NUMBER", 0) == NULL){
    if (update_unexpanded(load_table, sql_file) == ERROR){
      printf("Error processing the file: %s\n", argv[1]);
      exit(1);
    }
  }else{
    if (update_expanded(load_table, sql_file) == ERROR){
      printf("Error processing the file: %s\n", argv[1]);
      exit(1);
    }
  }
  
  free_table(load_table);

  /*
   * Finish up
   */
  fprintf(sql_file, "\nCOMMIT TRANSACTION");
  fprintf(sql_file, "\nGO\n");
  fclose(sql_file);
  printf("Processing complete -- %s generated\n", sql_filename);
    
  exit(0);  

}
