#include "getref.h"

/* GLOBAL.H -- Global variables used by getref 
 *
 * The global variables defined in this file control much of the operation
 * of getref. They describe the significant names of the database tables
 * and fields used by getref as well as the corresponding names of the
 * header keyword fields. If new instruments or reference files are added 
 * to the calibration pipeline, this file must be updated to track the 
 * changes. The variables in this file are structured as arrays of pointers
 * to character strings. The dimensions of the arrays are defined by macros 
 * in getref.h. Each array must be terminated by a zero length string so 
 * that the program using the array knows where it ends.
 *
 * The INSTRUMENT array contains the names of the instruments in alphabetical
 * order. This order is the same (and must remain the same) as the 
 * instruments and icodes macros defined in config.h. All other arrays in 
 * this file subscripted on instrument must also have the same number of
 * elements in the same order as the instrument array.
 *
 * The CDBSFILE array contains the name of the file level relation for
 * each instrument in the cdbs database. The CDBSROW array contains the name
 * of the corresponding row level relation and the DADSTAB array contains 
 * the name of the corresponding relation in the dads database. The OBSFIELD
 * array contains the name of the field in the dads relation containing the
 * name of the observation set name.
 *
 * The KEYWORD array contains the name of the header keywords containing 
 * the names of the calibration reference files. The names should be
 * in the same order that they occur in the header, though this is a problem
 * for the stis in that the header keywords depend on whether the ccds or
 * mamas were used. The BESTREF array contains the names of the fields in 
 * the dads database containing the recommended best reference files. The
 * order of the elements in the array matches the keyword array. The OLDREF
 * array contains the reference file originally used to process the 
 * observation. It, likewise, is in the same order as the keyword array.
 *
 * The CDBSMODE array contains the names of the observation mode fields as 
 * listed in the cdbs row level relation and the DADSMODE array contains
 * the names of the corresponding fields in the dads relation.
 *
 * The DETAIL_FILECOL array contains the fields in the cdbs file level 
 * relation used in the detail query. The DETAIL_ROWCOL array contains the
 * names of the fields in the cdbs row level relation, aside from the 
 * observation mode fields, which are used in the detail query.
 */

char *instrument[MAXINST] = {"foc", "fos", "hrs", "hsp", "nic",
"stis", "synphot", "wfpc2", "wfpc", ""};

char *cdbsfile[MAXINST] ={"foc_file", "fos_file", "hrs_file",
"hsp_file", "nic_file", "stis_file", "synphot_file", "wfpc2_file",
"wfpc_file", ""};

char *cdbsrow[MAXINST] ={"foc_row", "fos_row", "hrs_row", "hsp_row",
"nic_row", "stis_row", "synphot_row", "wfpc2_row", "wfpc_row", ""};

char *dadstab[MAXINST] = {"foc_ref_data", "fos_ref_data",
"hrs_ref_data", "hsp_ref_data", "nicmos_ref_data", "stis_ref_data",
"?", "wfpc2_ref_data", "wfpc_ref_data", ""};

char *obsfield[MAXINST] = {"fcr_data_set_name", "fsr_data_set_name",
"hsr_data_set_name", "hpr_data_set_name", "nsr_data_set_name",
"ssr_data_set_name", "?", "w2r_data_set_name", "wcr_data_set_name", ""};

char *keyword[MAXINST][MAXREF] = {
{"bachfile", "itfhfile", "blmhfile", "unihfile", "sdehfile", 	/* foc */
"geohfile", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", "", 
 "", "", "", "", ""},

{"bachfile", "fl1hfile", "fl2hfile", "iv1hfile", "iv2hfile", 	/* fos */
"aishfile", "rethfile", "ddthfile", "dq1hfile", "dq2hfile", 
"ccg2", "ccs0", "ccs1", "ccs2", "ccs3", 
"ccs4", "ccs5", "ccs6", "ccs7", "ccs8", 
"ccs9", "ccsa", "ccsb", "ccsc", "ccsd",
"", "", "", "", ""},

{"diohfile", "phchfile", "vighfile", "abshfile", "nethfile", 	/* hrs  */
"dqihfile", "saahfile", "ccr1", "ccr2", "ccr3", 
"ccr4", "ccr5", "ccr6", "ccr7", "ccr8", 
"ccr9", "ccra", "ccrb", "ccrc", "ccrd", 
"ccg2", "", "", "", "", 
"", "", "", "", ""},

{"ccp0", "ccp1", "ccp2", "ccp3", "ccp4", 			/* hsp */
"ccp5", "ccp7", "ccp8", "ccp9", "",
"", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", ""},

{"maskfile", "noisfile", "nlinfile", "darkfile", "flatfile",	/* nic */
"phottab", "backtab", "illmfile", "", "", 
"", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", ""},

{"bpixtab", "mlintab", "ccdtab", "atodtab", "biasfile", 	/* stis */
"darkfile", "pfltfile", "dfltfile", "lfltfile", "shadfile", 
"apertab", "apdestab", "sptrctab", "disptab", "inangtab", 
"sdstfile", "lamptab", "idctab", "sdctab", "phottab", 
"crrejtab", "xtractab", "mofftab", "pctab", "", 
"", "", "", "", ""},

{"", "", "", "", "", 						/* synphot */
"", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", ""},

{"maskfile", "atodfile", "biasfile", "darkfile", "flatfile", 	/* wfpc2 */
"shadfile", "graphtab", "comptab", "", "", 
"", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", ""},

{"maskfile", "atodfile", "biasfile", "preffile", "purgfile", 	/* wfpc */
"darkfile", "flatfile", "phottab", "", "", 
"", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", "", 
"", "", "", "", ""}
};

char *bestref[MAXINST][MAXREF] = {
{"fcr_best_bac", "fcr_best_itf", "fcr_best_blm", "fcr_best_unitab",  /* foc */
"fcr_best_sde", "fcr_best_geo", "",  "",  
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", ""},

{"fsr_best_bac", "fsr_best_flt1", "fsr_best_flt2", "fsr_best_ivs1", /* fos */
"fsr_best_ivs2", "fsr_best_ais", "fsr_best_ret", "fsr_best_ddt", 
"fsr_best_qin1", "fsr_best_qin2", "fsr_best_ccg2", "fsr_best_ccs0", 
"fsr_best_ccs1", "fsr_best_ccs2", "fsr_best_ccs3", "fsr_best_ccs4", 
"fsr_best_ccs5", "fsr_best_ccs6", "fsr_best_ccs7", "fsr_best_ccs8", 
"fsr_best_ccs9", "fsr_best_ccsa", "fsr_best_ccsb", "fsr_best_ccsc", 
"fsr_best_ccsd",  "", "", "", 
"", ""},

{"hsr_best_dio", "hsr_best_phc", "hsr_best_vig", "hsr_best_abs",   /* hrs */
"hsr_best_net", "hsr_best_qin", "hsr_best_saa", "hsr_best_ccr1", 
"hsr_best_ccr2", "hsr_best_ccr3", "hsr_best_ccr4", "hsr_best_ccr5", 
"hsr_best_ccr6", "hsr_best_ccr7", "hsr_best_ccr8", "hsr_best_ccr9", 
"hsr_best_ccra", "hsr_best_ccrb", "hsr_best_ccrc", "hsr_best_ccrd", 
"hsr_best_ccg2", "", "", "", 
"", "", "", "", 
"", ""},

{"hpr_best_ccp0", "hpr_best_ccp1", "hpr_best_ccp2", "hpr_best_ccp3", /* hsp */
"hpr_best_ccp4", "hpr_best_ccp5", "hpr_best_ccp7", "hpr_best_ccp8", 
"hpr_best_ccp9", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", ""},

{"nsr_best_maskfile", "nsr_best_noisfile", "nsr_best_nlinfile", /* nic */
 "nsr_best_darkfile", "nsr_best_flatfile", "nsr_best_phottab", 
 "nsr_best_backtab",  "nsr_best_illmfile", "", 
 "", "", "", 
 "", "", "", 
 "", "", "", 
 "", "", "", 
 "", "", "", 
 "", "", "", 
 "", "", ""},

{"ssr_best_bpixtab", "ssr_best_mlintab", "ssr_best_ccdtab", 	/* stis */
 "ssr_best_atodtab", "ssr_best_bias", "ssr_best_dark", 
 "ssr_best_pflt", "ssr_best_dflt",  "ssr_best_lflt", 
 "ssr_best_shad", "ssr_best_apertab", "ssr_best_apdestab", 
 "ssr_best_sptrctab", "ssr_best_disptab", "ssr_best_inangtab", 
 "ssr_best_sdst", "ssr_best_lamptab", "ssr_best_idctab", 
 "ssr_best_sdctab", "ssr_best_phottab", "ssr_best_crrejtab", 
 "ssr_best_xtractab", "ssr_best_mofftab", "ssr_best_pctab", 
 "", "", "", 
 "", "", ""},

{"", "", "", "", 						/* synphot */
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", ""},

{"w2r_best_msk", "w2r_best_a2d", "w2r_best_bas", "w2r_best_drk",  /* wfpc2 */
"w2r_best_flt", "w2r_best_shd", "w2r_best_graphtab", "w2r_best_comptab", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", ""},

{"wcr_best_msk", "wcr_best_a2d", "wcr_best_bas", "wcr_best_prf",  /* wfpc */
"wcr_best_pur", "wcr_best_drk", "wcr_best_flt", "wcr_best_phottab", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", ""}
};

char *oldref[MAXINST][MAXREF] = {
{"fcr_bachfile", "fcr_itfhfile", "fcr_blmhfile", "fcr_unihfile",   /* foc */
"fcr_sdehfile", "fcr_geohfile", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"",  "", "", "", 
"", ""},

{"fsr_bachfile", "fsr_fl1hfile", "fsr_fl2hfile", "fsr_iv1hfile",   /* fos */
"fsr_iv2hfile", "fsr_aishfile", "fsr_rethfile", "fsr_ddthfile", 
"fsr_dq1hfile", "fsr_dq2hfile", "fsr_ccg2", "fsr_ccs0", 
"fsr_ccs1", "fsr_ccs2", "fsr_ccs3", "fsr_ccs4", 
"fsr_ccs5", "fsr_ccs6", "fsr_ccs7", "fsr_ccs8", 
"fsr_ccs9", "fsr_ccsa", "fsr_ccsb", "fsr_ccsc", 
"fsr_ccsd", "", "", "", 
"", ""},

{"hsr_diohfile", "hsr_phchfile", "hsr_vighfile", "hsr_abshfile",   /* hrs */
"hsr_nethfile", "hsr_dqihfile", "hsr_saahfile", "hsr_ccr1", 
"hsr_ccr2", "hsr_ccr3", "hsr_ccr4", "hsr_ccr5", 
"hsr_ccr6", "hsr_ccr7", "hsr_ccr8", "hsr_ccr9", 
"hsr_ccra", "hsr_ccrb", "hsr_ccrc", "hsr_ccrd", 
"hsr_ccg2", "", "", "", 
"", "", "", "", 
"", ""},

{"hpr_ccp0", "hpr_ccp1", "hpr_ccp2", "hpr_ccp3", 		   /* hsp */
"hpr_ccp4", "hpr_ccp5", "hpr_ccp7", "hpr_ccp8", 
"hpr_ccp9", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", ""},

{"nsr_maskfile", "nsr_noisfile", "nsr_nlinfile", "nsr_darkfile", /* nic */
"nsr_flatfile", "nsr_phottab", "nsr_backtab", "nsr_illmfile", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", ""},

{"ssr_bpixtab", "ssr_mlintab", "ssr_ccdtab", "ssr_atodtab", 	/* stis */
"ssr_biasfile", "ssr_darkfile", "ssr_pfltfile", "ssr_dfltfile", 
"ssr_lfltfile", "ssr_shadfile", "ssr_apertab", "ssr_apdestab", 
"ssr_sptrctab", "ssr_disptab", "ssr_inangtab", "ssr_sdstfile", 
"ssr_lamptab", "ssr_idctab", "ssr_sdctab", "ssr_phottab", 
"ssr_crrejtab", "ssr_xtractab", "ssr_mofftab", "ssr_pctab", 
"", "", "", "", 
"", ""},

{"", "", "", "", 						/* synphot */
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", ""},

{"w2r_maskfile", "w2r_atodfile", "w2r_biasfile", "w2r_darkfile",  /* wfpc2 */
"w2r_flatfile", "w2r_shadfile", "w2r_graphtab", "w2r_comptab", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", ""},

{"wcr_maskfile", "wcr_atodfile", "wcr_biasfile", "wcr_preffile", /* wfpc */
"wcr_purgfile", "wcr_darkfile", "wcr_flatfile", "wcr_phottab", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", "", "", "", 
"", ""}
};

char *cdbsmode[MAXINST][MAXMODE] = {
{"pxformt", "cammode", "optcrly", "smmmode", "optelt1", 	/* foc */
"optelt2", "samppln", "linepfm", "sampoff", "lineoff", 
 "wavelength", "", "", "", ""},

{"detector", "overscan", "aper_id", "polar_id", "fgwa_id",	/* fos */
 "aper_pos", "pass_dir", "wavelength", "", "",
 "", "", "", "", ""},

{"detector", "grating", "", "", "",				/* hrs */
 "", "", "", "", "",
 "", "", "", "", ""},

{"", "", "", "", "",						/* hsp */
 "", "", "", "", "",
 "", "", "", "", ""},

{"nread", "camera", "filter", "readout", "samp_seq",		/* nic */
 "", "", "", "", "",
 "", "", "", "", ""},

{"ccdamp", "sclamp", "ccdgain", "cenwave", "obstype",		/* stis */
 "aperture", "binaxis1", "binaxis2", "ccdoffst", "detector",
 "opt_elem", "", "", "", ""},

{"", "", "", "", "",						/* synphot */
 "", "", "", "", "",
 "", "", "", "", ""},

{"mode", "atodgain", "filter1", "filter2", "serials", 		/* wfpc2 */
 "shutter", "", "", "", "",
 "", "", "", "", ""},

{"camera", "mode", "filter1", "filter2", "filtnam1", 		/* wfpc */
"filtnam2", "shutter", "clock", "", "",
 "", "", "", "", ""}
};

char *dadsmode[MAXINST][MAXMODE] = {

{"fcr_pxformt", "fcr_cammode", "fcr_optcrly", "fcr_smmmode",	/* foc */
 "fcr_optelt1", "fcr_optelt2", "fcr_samppln", "fcr_linepfm",
 "fcr_sampoff", "fcr_lineoff", "fcr_wavelength", "", 
 "", "", ""},

{"fsr_detector", "fsr_overscan", "fsr_aper_id", "fsr_polar_id",	/* fos */
  "fsr_fgwa_id", "fsr_aper_pos", "fsr_pass_dir", "fsr_wavelength", 
 "",  "", "", "", 
 "", "", ""},

{"hsr_detector", "hsr_grating", "", "", 			/* hrs */
 "", "", "", "", 
 "", "", "", "", 
 "", "", ""},

{"", "", "", "", 						/* hsp */
 "", "", "", "", 
 "", "", "", "", 
 "", "", ""},

{"nsr_mode", "nsr_atodgain", "nsr_filter1", "nsr_filter2",	/* nic */
 "nsr_serials", "nsr_shutter", "", "", 
 "", "", "", "", 
 "", "", ""},


{"ssr_camera", "ssr_mode", "ssr_filter1", "ssr_filter2",	/* stis */
"ssr_filtnam1", "ssr_filtnam2", "ssr_shutter", "ssr_clock", 
"", "", "", "", ""},

{"", "", "", "", 						/* synphot */
 "", "", "", "", 
 "", "", "", "", 
 "", "", ""},

{"w2r_mode", "w2r_atodgain", "w2r_filter1", "w2r_filter2",	/* wfpc2 */
 "w2r_serials", "w2r_shutter", "", "", 
 "", "", "", "", 
 "", "", ""},

{"wcr_camera", "wcr_mode", "wcr_filter1", "wcr_filter2",	/* wfpc */
 "wcr_serials", "wcr_shutter", "", "", 
 "", "", "", "", 
 "", "", ""}
};

char *detail_filecol[MAXDETAIL] = {"file_name", "reference_file_type",
"useafter_date", "general_availability_date", "opus_load_date", 
"archive_date", "reject_flag", "comparison_file_name", "", ""};

char *detail_rowcol[MAXDETAIL] = {"pedigree", "observation_begin_date", 
"observation_end_date", "", "", "", "", "" "", ""};

