#include <stdlib.h>
#include "loadfile.h"
#define  CELL(tab,row,col)	(tab->ncol * row + col + 1)

/* The functions in this file implement an in-memory ascii table. The table
 * has header keywords accessed by name and cells accessed by column name
 * and row number. The values returned by the get functions are pointers to
 * the contents of the table and should not be modified by the calling
 * function. Also, doing a put to the table may invalidate the pointers
 * returned if memory is shuffled. Copy the results of a get to another
 * string before doing a put. Table access is oriented towards random 
 * access, although sequential access is also supported. 
 *
 * Random access read routines:
 *
 * str = get_tab_key (tab, keyname)		Read header keyword
 * str = get_tab_cell (tab, colname, row)	Read table cell
 *
 * Sequential access:
 *
 * nkey = nkey_table (tab)			Num keywords in table (macro)
 * nrow = nrow_table (tab)			Numb rows in table (macro)
 * ncol = ncol_table (tab)			Num columns in table (macro)
 *
 * colnum = get_tab_colnum (tab, colname)	Column number from name
 * str = get_tab_colname (tab, colnum)		Column name from number
 * str = get_tab_keyname_bynum (tab, keynum)	Read header keyword name
 * str = get_tab_keyval_bynum (tab, keynum)	Read header keyword value
 * str = get_tab_cell_bynum (tab, colnum, row)	Read table cell
 *
 * The read functions return a value of NULL if the read does not succeed.
 *
 * Most often a table is created by reading the contents of a loadfile into
 * it. However, it is possible to create a table in memory or modify the
 * contents of a table read from a loadfile. The use of the add column 
 * routine is optional, as columns are implicitly added when the first row 
 * of the table is written. The add column routine can be used when the
 * table column names are known before the table is created. The add column
 * routine returns the number of the column created or -1 if the column
 * cannot be created because the table is already written to.
 *
 * Create and destroy a table:
 *
 * tab = create_table ()			Create an empty table
 * col = add_tab_col (tab, colname)		Add a column to an empty table
 * trunc_table (tab, nrow)			Truncate table to nrow rows
 * free_table (tab)				Free memory used by table
 *
 * Write to table:
 *
 * ok = put_tab_key (tab, keyname, value)	Write a header keyword
 * ok = put_tab_cell (tab, colname, row, value)	Write a table cell
 * ok = cat_tab_key (tab, colname, value)	Catenate string to header key
 * ok = cat_tab_cell (tab, colname, row, value)	Catenate string to table cell
 *
 * The write functions return a value of -1 if the write does not succeed 
 * and 0 if it does. In addition to the error codes returned by the read 
 * and write functions, an error message is stored in the global variable 
 * lod_error.
 */

static char *table_message[] = {
    "Header keyword not found",			/* 0 */
    "Table column not found",			/* 1 */
    "Table row not found",			/* 2 */
    "Cannot create table column",		/* 3 */
    "Cannot create table row"			/* 4 */
};

/* ADD_TAB_COL -- Add a new column to a table
 *
 * i: tab	pointer to table
 * i: name	column name
 */

int add_tab_col (table *tab, char *name)
{
    int		col;

    /* Check to see if more than one row has been written */
    if (tab->nrow > 1) {
	lod_error = table_message[3];
	return (BAD);
    }

    /* Make sure we do not add the same column twice */
    col = find_text (&tab->colname, name);

    /* Add the column and set its value to the empty cell */
    if (col == BAD) {
	col = add_text (&tab->colname, name);
	tab->ncol = col + 1;

	copy_text (&tab->cell, 0, CELL(tab,0,col));
	tab->nrow = 1;
    }

    return (col);
}

/* CAT_TAB_CELL -- Concatenate a string to the contents of a table cell
 *
 * u: tab	pointer to table
 * i: name	column name
 * i: row	row number
 * i: value	string to concatenate
 */

int cat_tab_cell (table *tab, char *name, int row, char *value)
{
    int		col;

    col = find_text (&tab->colname, name);
    if (col == BAD) {
	lod_error = table_message[1];
	return (BAD);
    }

    if (row >= tab->nrow) {
	lod_error = table_message[2];
	return (BAD);
    }

    cat_text (&tab->cell, CELL(tab,row,col), value);
    return (GOOD);
}

/* CAT_TAB_KEY -- Concatenate a string to the contents of a header keyword
 *
 * u: tab	pointer to table
 * i: name	column name
 * i: value	string to concatenate
 */

int cat_tab_key (table *tab, char *name, char *value)
{
    int		line;

    line = find_text (&tab->keyname, name);

    if (line == BAD) {
	lod_error = table_message[0];
	return (BAD);

    } else {
	cat_text (&tab->keyval, line, value);
	return (GOOD);
    }
}

/* CREATE_TABLE -- Create a new, empty table
 */

table *create_table (void)
{
    table	*tab;

    tab = (table *) malloc (sizeof (table));

    init_text (&tab->keyname, 10, 100);
    init_text (&tab->keyval, 10, 400);
    init_text (&tab->colname, 10, 100);
    init_text (&tab->cell, 100, 4000);

    add_text (&tab->cell, "");
    tab->nrow = 0;
    tab->ncol = 0;

    return (tab);
}

/* FREE_TABLE -- Free memory associated with a table
 *
 * u: tab	pointer to table
 */
void free_table (table *tab) 
{
    free_text (&tab->keyname);
    free_text (&tab->keyval);
    free_text (&tab->colname);
    free_text (&tab->cell);

    free (tab);
}

/* GET_TAB_CELL -- Get a cell from the table
 *
 * i: tab	pointer to table
 * i: name	column name
 * i: row	table row
 */

char *get_tab_cell (table *tab, char *name, int row)
{
    int 	col;

    col = find_text (&tab->colname, name);
    if (col == BAD) {
	lod_error = table_message[1];
	return (NULL);
    }

    if (row >= tab->nrow) {
	lod_error = table_message[2];
	return (NULL);
    }

    return (get_text (&tab->cell, CELL(tab,row,col)));
}
 
/* GET_TAB_CELL_BYNUM -- Get a cell from the table by position
 *
 * i: tab	pointer to table
 * i: col	column number
 * i: row	table row
 */

char *get_tab_cell_bynum (table *tab, int col, int row)
{

    if (col >= tab->ncol) {
	lod_error = table_message[1];
	return (NULL);
    }

    if (row >= tab->nrow) {
	lod_error = table_message[2];
	return (NULL);
    }

    return (get_text (&tab->cell, CELL(tab,row,col)));
}
 
/* GET_TAB_COLNAME -- Get a column name given the column number
 *
 * i: tab	pointer to table
 * i: col	column number
 */

char *get_tab_colname (table *tab, int col)
{
    char	*str;

    str = get_text (&tab->colname, col);
    if (str == NULL)
	lod_error = table_message[0];

    return (str);
}

/* GET_TAB_COLNUM -- Get a column number from the table
 *
 * i: tab	pointer to table
 * i: name	column name
 */

int get_tab_colnum (table *tab, char *name)
{
    int 	col;

    col = find_text (&tab->colname, name);
    if (col == BAD)
	lod_error = table_message[1];

    return (col);
}

/* GET_TAB_KEY -- Get a header keyword from the table
 *
 * i: tab	pointer to table
 * i: name	keyword name
 */

char *get_tab_key (table *tab, char *name)
{
    int		line;

    line = find_text (&tab->keyname, name);
    if (line == BAD) {
	lod_error = table_message[0];
	return (NULL);
    }

    return (get_text (&tab->keyval, line));
}

/* GET_TAB_KEYNAME_BYNUM -- Get a header keyword name by its position
 *
 * i: tab	pointer to table
 * i: line	keyword line
 */

char *get_tab_keyname_bynum (table *tab, int line)
{
    char	*str;

    str = get_text (&tab->keyname, line);
    if (str == NULL)
	lod_error = table_message[0];

    return (str);
}

/* GET_TAB_KEYVAL_BYNUM -- Get a header keyword by its position
 *
 * i: tab	pointer to table
 * i: line	keyword line
 */

char *get_tab_keyval_bynum (table *tab, int line)
{
    char	*str;

    str = get_text (&tab->keyval, line);
    if (str == NULL)
	lod_error = table_message[0];

    return (str);
}

/* PUT_TAB_CELL -- Write a cell to the table
 *
 * u: tab	table descriptor
 * i: name	column name
 * i: row	row number
 * i: value	cell value
 */

int put_tab_cell (table *tab, char *name, int row, char *value)
{
    int		col, icol;

    /* Columns may be implicitly added when the first row of the table
     * is written. 
     */

    col = find_text (&tab->colname, name);
    if (col == BAD) {
	if (tab->nrow <= 1) {
	    col = add_text (&tab->colname, name);
	    tab->ncol = col + 1;
	} else {
	    lod_error = table_message[3];
	    return (BAD);
	}
    }

    /* To add a new row to the table, fill all the cells with the
     * empty cell, which is allocated when the table is created.
     */

    if (row > tab->nrow) {
	lod_error = table_message[4];
	return (BAD);

    } else if (row == tab->nrow) {
	for (icol = 0; icol < tab->ncol; icol++)
	    copy_text (&tab->cell, 0, CELL(tab,row,icol));

	tab->nrow = row + 1;
    }

    put_text (&tab->cell, CELL(tab,row,col), value);
    return (GOOD);
}


/* PUT_TAB_KEY -- Put a keyword in the table header
 *
 * u: tab	table descriptor
 * i: name	header keyword name
 * i: value	header keyword value
 */

int put_tab_key (table *tab, char *name, char *value)
{
    int		line;

    line = find_text (&tab->keyname, name);

    if (line == BAD) {
	line = add_text (&tab->keyname, name);
	add_text (&tab->keyval, value);

    } else {
	put_text (&tab->keyval, line, value);
    }

    return (GOOD);
}

/* TRUNC_TABLE -- Truncate table to specified number of rows
 *
 * i: tab	table descriptor
 * i: nrow	number of rows in truncated table
 */

void trunc_table (table *tab, int nrow)
{
    if (nrow < tab->nrow)
	tab->nrow = nrow;
}
