/*
*******************************************************************************
*
* Module:     opus_sql_gen.c
*
* File Name:  opus_sql_gen.c
*
* Purpose:    Generate a script file opus_n.sql containing insert/delete sql
*             for use by opus, where n is the delivery number.
*               
* Usage:      opus_sql_gen [p1]
*               where p1 = deliver number, 
*                          if ommit, use delivery_number.value
*
* Logical:    setenv CDBS_SERVER <server> 
*             setenv CDBS_DB <dbname> 
*               
* Design:     1) If user did not provide p1 (delivery_number), use
*                delivery_number.value as default delivery_number
*             2) The script file start with a begin transcation
*             3) Create local variable @load_date to save current time
*                "dd-MMM-yyyy hh:mm:ss" for use by load_time fields in OPUS
*             4) Write message "LOAD_DATE_USED=<ccurrent datetime>"
*                              "DELIVERY_NUMBER_USED=<deliverynumber>"
*             5) Loop through <opus_log> table to get <file_name>, 
*                <expansion_number>, <instrument>, <reference_file_type>, 
*                <operation> data having delivery_number = <delivery_number>
*             6) get <opus_table> name from relation_info having instrument,
*                reference_file_type equals to the data retrieved from 5)
*             7) create map records which contains cdbs table, fields name,
*                corresponding to opus table, fields name by :
*                select <cdbs_field>, <cdbs_table>, <opus_field> from 
*                opus_fields where opus_table = <opus_table> and cdbs_table 
*                like "<instrument>%"
*                to get filed data type by :
*                select column_name = c.name, type = t.name
*                from dbo.syscolumns c, dbo.systypes t, dbo.sysobjects o
*                where o.name = <cdbs_table> and c.name = <cdbs_field>
*                and c.id = o.id and o.type = 'U' and c.usertype *= t.usertype
*             8) based on map array, create a string <sl> for the select-list:
*                <map[0].cdbs_table_name>.<map[0].cdbs_field_name> s0,
*                <map[1].cdbs_table_name>.<map[1].cdbs_field_name> s1, ...
*                and run query to get cdbs data:
*                select distinct <sl> from <instrument>_file, <instrument>_row
*                where <instrument>_file.file_name = <instrument>_row.file_name
*                      = <file_name>
*                and <instrument>_file.expansion_number = 
*                    <instrument>_row.expansion_number = <expansion_number>
*             9) if <operation> data (from 5) = "I" then generate insert
*                statement and write to output: 
*                (note: <instrument> data is for opus "caltable" only)
*                insert <opus_table> 
*                 (map[i].opus_field..., load_time[, instrument]) values
*                 (map[i].field_value,..., @load_date[, '<instrument_abbrev>')
*            10) if operation data = "R" or "D" then generate a delete 
*                statement and write to output:
*                delete from <opus_table> 
*                where map[i].opus_field = map[i].field_value ...
*                  [and instrument = '<instrument_abbrev>']
*            11) go back to 5) until no more recored returned from <opus_log>
*            12) Write message "OPUS_MODS_OK" and "commit transaction" to 
*                output and close database connection and output file
*
* Members:
*            get_delivery_number
*            instrument_abbrev
*            init_output
*            create_sql_file
*            determine_opus_table
*            copy_mode
*            cdbs_field_type
*            map_table_field
*            create_opus_field_list
*            write_opus_query
*            bind_cdbs_columns
*            create_cdbs_query
*            insert_opus_data
*            delete_opus_data
*            use_after_date
*            delete_file_prefix
*            write_command
*               
*
* Modification History:
*
*   Date       OPR      Who                       Reason
* --------   -------    ---   -------------------------------------------------
* 05/95                  sk   Original
*
*******************************************************************************
*/

#include "ostdb.h"
#include "stdb_utility_pkg.i"  /* for stdb_string_upper, stdb_string_lower */

#define TRUE       1
#define FALSE      0
#define STR        30
#define LINE       80

/* map_type.field_type */
#define CDBS_SMALLINT    1
#define CDBS_INT         2
#define CDBS_FLOAT       3
#define CDBS_STRING      4
#define CDBS_DATE        5

FILE *out;                    /* output file pointer                       */

int  db_proc1 = 0, db_proc2 = 0, db_proc3 = 0;  /* database process number */
int  error_out = FALSE;       /* error flag, TRUE or FALSE                 */

static char daytab[2][13] = { /* straight from Kernighan and Ritchie       */
   {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
   {0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31} };

static char   *month_array[] = {"Illegal month", "Jan", "Feb", "Mar",
       "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

struct map_type {
    char cdbs_field[51];
    char cdbs_table[51];
    char opus_field[51];
    char opus_table[51];
    int  field_type;                /* see above #define section  */
    union {                         /* field data                 */
       short  sval;                 /* smallint                   */
       int    ival;                 /* int                        */
       double dval;                 /* float                      */
       char   cval[51];             /* varchar/char               */
       STDB_DATEREC_TYPE use_after; /* datetime                   */
    }  field_value;
};

struct map_type map[50];            /* cdbs, opus table-field map */

/*
*******************************************************************************
*
* Function:     setup_db_connection
*
* Purpose:      setup 3 connections to the same database
*
* Note:         connection relationship:
*
*               -- db_proc3 is used for independent query or nested inside 
*                  db_proc2
*               -- db_proc2 is nested inside db_proc1 which gets cdbs data 
*                  (combine with db_proc3) to generate update opus sql
*               -- db_proc1 is used to process opus_log data
*
*               example:
*
*               setup db_proc3 ...
*               while stdb_get_exec (db_proc3) ...
*
*               setup db_proc1 ...  //  process opus_log data 
*               while stdb_get_exec (db_proc1) ...
*               {
*                  setup db_proc3 ...
*                  while stdb_get_exec (db_proc3) ...
*
*                  setup db_proc2 ... 
*                  while stdb_get_exec (db_proc2) ...  // get cdbs data 
*                  {
*                     setup db_proc3 ... //  get cdbs data 
*                     while stdb_get_exec (db_proc3) ...
*                  }
*               }
*
*******************************************************************************
*/

void setup_db_connection(
   char *server,        /* i: server name   */
   char *dbname)        /* i: database name */

{
   /* setup 3 connection */
   if ((stdb_connect_dbms(STDB_SYBASE, dbname, server, "", "",
            &db_proc1) == STDB_ERROR) ||
       (stdb_connect_dbms(STDB_SYBASE, dbname, server, "", "",
            &db_proc2) == STDB_ERROR) ||
       (stdb_connect_dbms(STDB_SYBASE, dbname, server, "", "",
            &db_proc3) == STDB_ERROR))
   {
      printf("ERROR -- Cannot connect to database\n" );
      exit(1);
   }

}   /* setup_db_connection */


/*
*******************************************************************************
*
* Function:     get_delivery_number
*
* Purpose:      get latest delivery_number by :
*               select value from delivery_number
*
* Returns:      delivery_number.value 
*
*******************************************************************************
*/

/* FunctionDefBegin */

int get_delivery_number()

/* FunctionDefEnd */

{
   char cmd_buf[500];             /* command buffer          */
   int  value;

   if (stdb_query_setup(db_proc3, "get_delivery_number", 1) == STDB_ERROR)
       exit(1);
  
   if (stdb_bind_int(db_proc3, "value", &value) == STDB_ERROR)
       exit(1);
  
   strcpy (cmd_buf, "select value from delivery_number");
   if (stdb_query(db_proc3, cmd_buf) == STDB_ERROR)
      exit(1);

   while (stdb_get_exec(db_proc3) == STDB_CONTINUE)
      ;

   stdb_close_query(db_proc3);

   return(value);   

}  /* get_delivery_number */

/*
*******************************************************************************
*
* Function:     instrument_abbrev
*
* Purpose:      get instrument's abbreviation
*
* Returns:      instrument abbreviation
*
*******************************************************************************
*/

/* FunctionDefBegin */

char *instrument_abbrev (
   const int delivery_number)     /* i: delivery_number */

/* FunctionDefEnd */
{
   char abbreviation[2];             
   char cmd_buf[200];

   if (stdb_query_setup(db_proc3, "instrument_abbrev", 1) == STDB_ERROR)
       exit(1);

   if  ((stdb_int_param(db_proc3, "delivery_number", delivery_number) == STDB_ERROR) ||
       (stdb_bind_string(db_proc3, "abbreviation", abbreviation, 0) == STDB_ERROR))
       exit(1);

   strcpy (cmd_buf, "select distinct abbreviation from ");
   strcat (cmd_buf, "instrument_abbreviations i, opus_log o where ");
   strcat (cmd_buf, "i.instrument = o.instrument and o.delivery_number = ");
   strcat (cmd_buf, "@delivery_number");
   if (stdb_query(db_proc3, cmd_buf) == STDB_ERROR)
      exit(1);

   while (stdb_get_exec(db_proc3) == STDB_CONTINUE)
      ;

   stdb_close_query(db_proc3);

   stdb_string_upper(abbreviation);

   if (strcmp(abbreviation, "G") == 0)   /* OPUS use "M" for "synphot" */
     strcpy(abbreviation, "M");

   return(abbreviation);

}  /* instrument_abbrev */

/*
*******************************************************************************
*
* Function:     init_output
*
* Purpose:      open output files and write to output file:
*               begin transaction, create local variable @load_date and assign
*               getdate() to @load_date in the form "dd-MMM-yyyy hh:mm:ss"
*               we'll use it's data for insert opus field load_date later and
*               message LOAD_DATE_USED=<ccurrent datetime>
*                       DELIVERY_NUMBER_USED=<deliverynumber>
*
* Returns:      none
*
*******************************************************************************
*/

/* FunctionDefBegin */
                        
void init_output (
   char *output_file,
   int  delivery_number)

/* FunctionDefEnd */

{
   if ((out = fopen(output_file, "w")) == NULL)
   {
      printf("\nERROR --  Cannot create output file: %s",
             output_file);
      exit(1);
   }
   else
   {
      fprintf(out, "begin tran\n\n");
      fprintf(out, "/* store current time to local variable @load_date ");
      fprintf(out, "for use by load_time  */\n");
      fprintf(out, "/* fields in OPUS, the format is ");
      fprintf(out, "\"dd-MMM-yyyy hh:mm:ss\" */\n");
      fprintf(out, "declare @load_date    varchar(20)\n");
      fprintf(out, "declare @dbtime       varchar(20)\n");
      fprintf(out, "declare @current_time datetime\n");
      fprintf(out, "select @current_time = getdate()\n");
      fprintf(out, "select @dbtime=convert(char(11), @current_time, 106)+' '+\n");
      fprintf(out, "       convert(char(8), @current_time, 108)\n");
      fprintf(out, "select @load_date = substring(@dbtime, 1, 2) + '-' + substring(@dbtime, 4, 1) +\n");
      fprintf(out, "       upper(substring(@dbtime, 5, 1)) + upper(substring(@dbtime, 6, 1)) +\n");
      fprintf(out, "       '-' + substring(@dbtime, 8, 13)\n\n\n");

      fprintf(out, "print \"LOAD_DATE_%%1!= %%2!\", \"USED\", @current_time\n");
      fprintf(out, "print \"DELIVERY_NUMBER_%%1!= %%2!\", \"USED\", \"%d\"\n\n", 
                   delivery_number);

   }

}  /* init_output */

/*
*******************************************************************************
*
* Function:     determine_opus_table 
*
* Purpose:      get opus_table name from relation_info having instrument, 
*               reference_file_type equals to the given data
*      
* Returns:      none
*
*******************************************************************************
*/

/* FunctionDefBegin */

void determine_opus_table (
   char *reference_file_type,    /* i:   reference file type */
   char *instrument,             /* i:   instrument name     */
   char *type,                   /* i/o: I=image, T=table    */
   char *opus_table)             /* i/o: opus table name     */

/* FunctionDefEnd */

{
   char cmd_buf[500];             /* command buffer          */

   if (stdb_query_setup(db_proc3, "determine_opus_table", 1) == STDB_ERROR)
       exit(1);

   if ((stdb_string_param (db_proc3, "reference_file_type", reference_file_type) == STDB_ERROR) ||
       (stdb_string_param (db_proc3, "instrument", instrument) == STDB_ERROR) ||
       (stdb_bind_string(db_proc3, "type", type, 0) == STDB_ERROR) ||
       (stdb_bind_string(db_proc3, "opus_table", opus_table, 0) == STDB_ERROR))
       exit(1);

   strcpy (cmd_buf, "select opus_table, type from relation_info where ");
   strcat (cmd_buf, "reference_file_type = @reference_file_type ");
   strcat (cmd_buf, "and instrument = @instrument");
   if (stdb_query(db_proc3, cmd_buf) == STDB_ERROR)
      exit(1);

   while (stdb_get_exec(db_proc3) == STDB_CONTINUE)
      ;   /* should only get one record returned */

   if (stdb_record_count(db_proc3) == 0)
   {
      printf("No opus table matches %s's reference_file_type '%s'\n",
              instrument, reference_file_type);
      error_out = TRUE;
   }

   stdb_close_query(db_proc3);

}  /* determine_opus_table */

/*
*******************************************************************************
*
* Function:     copy_mode
*
* Purpose:
*
* Returns:      TRUE = data goes to opus
*               FALSE = not opus data
*
*******************************************************************************
*/

/* FunctionDefBegin */

int copy_mode (
   char *instrument,          /* i:   instrument name      */
   char *reference_file_type, /* i:   reference_file_type  */
   char *cdbs_field)          /* i:   cdbs field           */

/* FunctionDefEnd */

{
   char file_select[2];
   char cmd_buf[200];

   strcpy(file_select, " ");

   if (stdb_query_setup(db_proc3, "copy_mode", 1) == STDB_ERROR)
       exit(1);

   if ((stdb_string_param (db_proc3, "instrument", instrument) == STDB_ERROR) ||
       (stdb_string_param (db_proc3, "reference_file_type", reference_file_type) == STDB_ERROR) ||
       (stdb_string_param (db_proc3, "cdbs_field", cdbs_field) == STDB_ERROR) ||
       (stdb_bind_string(db_proc3, "file_select", file_select, 0) == STDB_ERROR))
       exit(1);

   strcpy (cmd_buf, "select file_select from cdbs_mode where instrument=");
   strcat (cmd_buf, "@instrument and reference_file_type=@reference_file_type");
   strcat (cmd_buf, " and modes=@cdbs_field");

   if (stdb_query(db_proc3, cmd_buf) == STDB_ERROR)
      exit(1);

   while (stdb_get_exec(db_proc3) == STDB_CONTINUE)
      ;

   stdb_close_query(db_proc3);

   if (strcmp(file_select, "Y") == 0)
      return(TRUE);
   else
      return(FALSE);

}  /* copy_mode */


/*
*******************************************************************************
*
* Function:     cdbs_field_type
*
* Purpose:      determine cdbs_table.field data type
*
* Query:        select column_name = c.name, type = t.name
*                 from dbo.syscolumns c, dbo.systypes t, dbo.sysobjects o
*                where o.name = @cdbs_table and c.name = @cdbs_field 
*                  and c.id = o.id and o.type = 'U' 
*                  and c.usertype *= t.usertype
*
* Support CDBS type: char, varchar, datetime, smallint, int, float
*
* Note:         all cdbs data type should match opus data type, except cdbs 
*               datetime fields were defines as varchar in opus table
*
* Returns:      defined cdbs field type
*
*******************************************************************************
*/

/* FunctionDefBegin */
             
int cdbs_field_type (
   char *cdbs_table,    /* i: cdbs table name */
   char *cdbs_field)    /* i: cdbs field name */

/* FunctionDefEnd */

{
   char cmd_buf[500];   /* query                                   */
   char type[31];       /* cdbs field type                         */
   int  column_type;    /* define cdbs field int type, for detail, */
                        /* see #define section map_type.field_type */
   
   if (stdb_query_setup(db_proc3, "cdbs_field_type", 1) == STDB_ERROR)
       exit(1);

   if ((stdb_string_param (db_proc3, "cdbs_table", cdbs_table) == STDB_ERROR) ||
       (stdb_string_param (db_proc3, "cdbs_field", cdbs_field) == STDB_ERROR) ||
       (stdb_bind_string(db_proc3, "type", type, 0) == STDB_ERROR))
       exit(1);

   strcpy (cmd_buf, "select t.name type from dbo.syscolumns c, ");
   strcat (cmd_buf, "dbo.systypes t, dbo.sysobjects o where o.name =");
   strcat (cmd_buf, "@cdbs_table and c.name = @cdbs_field and c.id = o.id ");
   strcat (cmd_buf, "and o.type = 'U' and c.usertype *= t.usertype");
   if (stdb_query(db_proc3, cmd_buf) == STDB_ERROR)
      exit(1);

   /* should only get one record returned */
   while (stdb_get_exec(db_proc3) == STDB_CONTINUE)
   {
      if (strcmp(type, "smallint") == 0)
         column_type = CDBS_SMALLINT;
      else if (strcmp(type, "int") == 0)
         column_type = CDBS_INT;
      else if (strcmp(type, "float") == 0)
         column_type = CDBS_FLOAT;
      else if (strcmp(type, "datetime") == 0)
         column_type = CDBS_DATE;
      else  /* char, varchar */
         column_type = CDBS_STRING;
   }

   stdb_close_query(db_proc3);

   return (column_type);

}   /* cdbs_field_type */

/*
*******************************************************************************
*
* Function:     map_table_field 
*
* Purpose:      create map records which contains cdbs table name, fields name
*               and corresponding opus table, fields name and filed data type
*
*               to determine cdbs table name, fields name and opus fields :
*               select cdbs_field, cdbs_table, opus_field from opus_fields 
*               where opus_table = given opus table name and cdbs_table like
*               instrument%               
*
*               call function to get field type
*
* Returns:      none
*
*******************************************************************************
*/

/* FunctionDefBegin */

void map_table_field (
   char *opus_table,          /* i:   opus table name      */
   char *instrument,          /* i:   instrument name      */
   char *type,                /* i:   image or table       */
   char *reference_file_type, /* i:   reference_file_type  */
   int  *idx)                 /* i/o: map array index      */

/* FunctionDefEnd */

{
   char cmd_buf[500];     /* command buffer                          */
   char cdbs_field[51];   /* cdbs field name                         */
   char cdbs_table[51];   /* cdbs table name which map to opus table */
   char cdbs_mode[2];     /* 'Y' = field is mode, otherwise = 'N'    */
   char opus_field[51];   /* opus table field maps cdbs table field  */
   char instr[51];        /* lowercase instrument                    */
   int  opus_data;

   if (stdb_query_setup(db_proc2, "map_table_field", 1) == STDB_ERROR)
       exit(1);

   if ((stdb_string_param (db_proc2, "opus_table", opus_table) == STDB_ERROR) ||
       (stdb_bind_string(db_proc2, "cdbs_field", cdbs_field, 0) == STDB_ERROR) ||
       (stdb_bind_string(db_proc2, "cdbs_table", cdbs_table, 0) == STDB_ERROR) ||
       (stdb_bind_string(db_proc2, "cdbs_mode", cdbs_mode, 0) == STDB_ERROR) ||
       (stdb_bind_string(db_proc2, "opus_field", opus_field, 0) == STDB_ERROR))
       exit(1);

   strcpy(instr, instrument);
   stdb_string_lower(instr);   /* convert to lowercase */
   strcpy (cmd_buf, "select cdbs_field, cdbs_table, cdbs_mode, opus_field ");
   strcat (cmd_buf, "from opus_fields ");
   strcat (cmd_buf, "where opus_table = @opus_table and cdbs_table like '");
   strcat (cmd_buf, instr);
   strcat (cmd_buf, "%'");
   if (stdb_query(db_proc2, cmd_buf) == STDB_ERROR)
      exit(1);

   /* should only get one record returned */
   while (stdb_get_exec(db_proc2) == STDB_CONTINUE)
   {
      if (strcmp(cdbs_mode, "N") == 0)
      {
         opus_data = TRUE;
      }
      else
      {
         opus_data = copy_mode(instrument,reference_file_type,cdbs_field);
      }

      if (opus_data)
      {
         strcpy (map[*idx].cdbs_field, cdbs_field);
         strcpy (map[*idx].cdbs_table, cdbs_table);
         strcpy (map[*idx].opus_table, opus_table);
         strcpy (map[*idx].opus_field, opus_field);
         map[*idx].field_type = cdbs_field_type (cdbs_table, cdbs_field);
         *idx = *idx + 1;
      }
   }

   stdb_close_query(db_proc2);

}   /* map_table_field */

/*
*******************************************************************************
*
* Function:     create_opus_field_list
*
* Purpose:      loop through map array to get list of opus fild name and
*               create list string for later insert into opus table use
*
* Returns:      none
*
*******************************************************************************
*/

/* FunctionDefBegin */

void create_opus_field_list (
   char  *opus_field_list,           /* i/o: list of opus table fields      */
   const int  idx,                   /* i:   number of record in map array  */
   const char *opus_table)           /* i:   opus table name                */

/* FunctionDefEnd */

{
   int  i;

   strcpy (opus_field_list, "(");
   i = 0;
   while (i < idx)
   {
      strcat (opus_field_list, map[i].opus_field);
      i++;
      strcat (opus_field_list, ", ");
   }

   /* add load_time, and instrument (for caltable only) */
   if (strcmp(opus_table, "caltable") == 0)
      strcat (opus_field_list, "load_time, instrument)");
   else
      strcat (opus_field_list, "load_time)"); 

   /* opus_field_list = (map[i].opus_field, ..., load_time[, instrument]) */

}  /* create_opus_field_list */


/*
*******************************************************************************
*
* Function:     bind_cdbs_columns 
*
* Purpose:      call stdb_bind_<type> to bind cdbs field for retrieving 
*               cdbs data
*               
* Note:         Since the retrieved columns veries from the cdbs tables, 
*               this function use alias to bind the retrieved columns, so
*               the bind call looks like:
*               stdb_bind_<type>(db_proc2, "xi", &map[i].field_value.val) 
*               the query looks like: 
*               select cdbs_field xi, ... from cdbs_table (see function
*               create_cdbs_query for detail)
*
* Returns:      none
*
*******************************************************************************
*/

/* FunctionDefBegin */

void bind_cdbs_columns (
   const int  idx)     /* i: number of record in map array */

/* FunctionDefEnd */

{
   int  i;
   char column[4];     /* cdbs bind field's alias = x<i>, i is map index */

   i = 0;
   while (i < idx)
   {
      sprintf (column, "x%-d", i);
      switch (map[i].field_type)
      {
         case CDBS_SMALLINT:
            if (stdb_bind_short(db_proc2, column, &map[i].field_value.sval) 
                == STDB_ERROR)
               exit(1);
            break;
 
         case CDBS_INT:
            if (stdb_bind_int(db_proc2, column, &map[i].field_value.ival) 
                == STDB_ERROR)
               exit(1);
            break;

         case CDBS_FLOAT:
            if (stdb_bind_double(db_proc2, column, &map[i].field_value.dval) 
                == STDB_ERROR)
               exit(1);
            break;

         case CDBS_STRING:
            if (stdb_bind_string(db_proc2, column, map[i].field_value.cval, 0) 
                 == STDB_ERROR)
               exit(1);
            break;

         case CDBS_DATE:
            if (stdb_bind_datetime(db_proc2, column, 
                   &map[i].field_value.use_after) == STDB_ERROR)
               exit(1);
            break;

      }   /* swith */      
      i++;
   }

}  /* bind_cdbs_columns */

/*
*******************************************************************************
*
* Function:     create_cdbs_query
*
* Purpose:      loop through map array to get cdbs column name to form the query
*
* Query:        select <cdbs_table>.<cdbs_field> x<i>, ... 
*                 from <instrucment>_file, <instrucment>_row
*                where <instrucment>_file.file_name = opus_log.file_name
*                  and <instrucment>_row.file_name = opus_log.file_name
*                  and <instrucment>_file.expansion_number = opus_log.expansion_number
*                  and <instrucment>_row.expansion_number = opus_log.expansion_number
*
*               x<i> -- where i = map array index
*
* Returns:      none
*
*******************************************************************************
*/

/* FunctionDefBegin */

void create_cdbs_query (
   char  *command_buffer,         /* i/o: select cdbs data query        */
   const int  idx,                /* i:   number of record in map array */
   const char *instrument,        /* i:   instrument name               */
   const char *file_name,         /* i:   file name                     */
   const int  expansion_number)   /* i:   expansion number              */

/* FunctionDefEnd */

{
   int i;
   char from_where[500];   /* part of query, "from <table> where <condition> */
   char field[110];        /* cdbs field name and alias */

   strcpy (command_buffer, "select distinct ");
   i = 0;
   while (i < idx)
   {
      sprintf (field, "%s.%s x%-d", map[i].cdbs_table, map[i].cdbs_field, i);
      strcat (command_buffer, field);
      i++;
      if (i < idx)
         strcat (command_buffer, ",");
   }

   sprintf (from_where, " from %s_file, %s_row where %s_file.file_name = \
'%s' and %s_file.expansion_number = %d and %s_row.file_name = '%s' and \
%s_row.expansion_number = %d", instrument, instrument, instrument, file_name, 
         instrument, expansion_number, instrument, file_name, instrument, 
         expansion_number);

   strcat (command_buffer, from_where);

}  /* create_cdbs_query */

/*
*******************************************************************************
*
* Function:     use_after_date
*
* Purpose:      This routine convert datetime "yyyy:mm:dd:hh:mi:ss"
*               to opus use_after format "yyyy.ddd:hh:mm:ss"
*
* Returns:      none
*
* Note:         based on David Silberberg's dmsdate function
*
*******************************************************************************
*/

/* FunctionDefBegin */

void use_after_date (
   const char *cdbs_date,  /* i:   datetime format "yyyy:mm:dd:hh:mi:ss"     */
   char  *use_after)       /* i/o: opus use_after format "yyyy.ddd:hh:mm:ss" */

/* FunctionDefEnd */

{
   int   i;
   int   day, month, year, leap, daynum;
   char  buff[10];

   use_after[0] = '\0';

   /* copy year */
   strncat(use_after, cdbs_date, 4);
   year = atoi(use_after);
   leap = year%4 == 0 && year%100 != 0 || year%400 == 0;
   strcat(use_after, ".");    /* "yyyy." */

   /* get month */
   buff[0] = '\0';
   strncat(buff, cdbs_date+5, 2);
   month = atoi(buff);
   
   /* get day of month */
   buff[0] = '\0';
   strncat(buff, cdbs_date+8, 2);
   day = atoi(buff);
 
   /* convert month/day to julian date */
   for (daynum = 0, i = 1; i < month; i++) 
       daynum += daytab[leap][i];
   daynum += day;
   sprintf(buff, "%d", daynum);
   
   /* julian date : 3 digit, righ justify, 0 filled */
   if (daynum < 100) 
      strcat(use_after, "0");
   if (daynum < 10) 
      strcat(use_after, "0");
   strcat(use_after, buff);
   strcat(use_after, ":");    /* "yyyy.ddd:" */

   /* copy hh:mi:ss */
   strncat(use_after, cdbs_date+11, 8);

}  /* use_after_date */

/*
*******************************************************************************
*
* Function:     delete_file_prefix
*
* Purpose:      this function is called when reference_file_type is THROUGHPUT.
*               in this case, filename contains IRAG logical prefix (filename=
*               log$file.ext), which is not accept by OPUS, this function 
*               overwrite filename by delete IRAG logical prefix.
*
* Returns:      none
*
*******************************************************************************
*/

/* FunctionDefBegin */

void delete_file_prefix (
   char *filename)          /* i/o: filename */

/* FunctionDefEnd */

{
   char *tmpfile;
   char *pos;

   pos = strchr(filename, '$');        /* locate $ sign in filename   */
   if (pos != NULL)                    /* found $ sign                */
   {
      tmpfile = pos+1;                 /* filename start after $ sign */
      strcpy(filename, tmpfile);       /* overwrite filename          */
   }

}  /* delete_file_prefix */

/*
*******************************************************************************
*
* Function:     write_command
*
* Purpose:      write sql command to output file, 80 characters/line
*
* Returns:      none
*
*******************************************************************************
*/

/* FunctionDefBegin */

void write_command (
   const char *command_buffer)   /* i: sql command */

/* FunctionDefEnd */
{
   char output_buffer[LINE+1];   /* text to write to output file         */
   int  cmd_size;                /* size if command                      */
   int  pos;                     /* index(position) of command           */
   int  nchar;                   /* number of characters to be copied    */

   cmd_size = strlen(command_buffer);
   pos = 0;

   /* process command buffer by copying 80 characters to output file each time,
      or  < 80 characters (up to last word in line, we don't want to break 
      word into 2 lines)      */  
   while (pos < cmd_size)
   {
      if ((cmd_size - pos) > LINE)
      {
         nchar = LINE;
         strncpy(output_buffer, command_buffer+pos, nchar);

         nchar--;
         /* find last delimiter */
         while ((output_buffer[nchar] != ' ') &&
                (output_buffer[nchar] != ',') &&
                (output_buffer[nchar] != ')'))
            nchar--;
         output_buffer[nchar+1] = '\0';
         pos = pos + nchar + 1;
         fprintf(out, "%s\n", output_buffer);
      }
      else
      {
         strcpy(output_buffer, command_buffer+pos);
         fprintf(out, "%s\n", output_buffer);
         pos = cmd_size;
      }
   }

   fprintf(out, "\n");
}   /* write_command */


/*
*******************************************************************************
*
* Function:     insert_opus_data
*
* Purpose:      generate insert sql to insert data to opus table and write
*               to output file
*
* Returns:      none
*
*******************************************************************************
*/

/* FunctionDefBegin */

void insert_opus_data (
   const int  idx,                   /* i: number of reccord in map array */
   const char *reference_file_type,  /* i: reference_file_type */
   const char *instrument,           /* i: instrument name     */
   const char *opus_table,           /* i: oput table name     */
   const char *opus_field_list,      /* i: list of opus fields */
   const char *instr_abbrev)

/* FunctionDefEnd */

{
   char command_buffer[2000];  /* query                                    */
   char field_data[100];       /* retrieved cdbs column value              */
   char tmp_file[100];         /* file_name value                          */
   char cdbs_date[30];         /* convert cdbs date to yyyy:mm:dd:hh:mi:ss */
   char use_after[30];         /* convert above date to yyyy.ddd:hh:mi:ss  */
   char str_fmt[10];           /* string format for print file_name        */
   int  i;                     /* map array index                          */

   sprintf (command_buffer, "insert %s ", opus_table);
   strcat (command_buffer, opus_field_list);
   strcat (command_buffer, " values (");
   i = 0;
   while (i < idx)
   {
      switch (map[i].field_type)
      {
         /* map[i].field_value were retrieved from function write_opus_query */

         case CDBS_SMALLINT:
            sprintf (field_data, "%d, ", map[i].field_value.sval);
            break;

         case CDBS_INT:
            sprintf (field_data, "%d, ", map[i].field_value.ival);
            break;

         case CDBS_FLOAT:
            sprintf (field_data, "%f, ", map[i].field_value.dval);
            break;

         case CDBS_DATE:
            if ((strcmp(map[i].opus_field, "use_after") == 0) &&
                (strcmp(map[i].cdbs_field, "useafter_date") == 0))
            {
               /* convert cdbs_field to string "yyyy:mm:dd:hh:mi:ss" */
               sprintf (cdbs_date, "%4d:%02d:%02d:%02d:%02d:%02d", 
                        map[i].field_value.use_after.dateyear,  
                        map[i].field_value.use_after.datemonth+1, 
                        map[i].field_value.use_after.datedmonth,
                        map[i].field_value.use_after.datehour, 
                        map[i].field_value.use_after.dateminute,
                        map[i].field_value.use_after.datesecond);

               /* convert mm:dd to julian date */
               use_after_date (cdbs_date, use_after);
               sprintf (field_data, "'%s', ", use_after);
            }
            break;
         case CDBS_STRING:
            if (strcmp(map[i].cdbs_field, "file_name") == 0)
            {
               strcpy(tmp_file, map[i].field_value.cval);

               if (strcmp(map[i].opus_field, "data_file") == 0)
               {  
                  /* data_file attribute in ccaa_wfc_ref, ccau_wf2_ref,
                     ccan_nicmos_ref, ccao_stis_ref are header file name, 
                     rather than the data file name */
                  if ((strcmp(opus_table, "ccaa_wfc_ref") != 0) &&
                      (strcmp(opus_table, "ccau_wf2_ref") != 0) &&
                      (strcmp(opus_table, "ccan_nicmos_ref") != 0) &&
                      (strcmp(opus_table, "ccao_stis_ref") != 0))
                  {
                     /* replace last file extension character to D */
                     tmp_file[strlen(map[i].field_value.cval)-1] = 'D';
                  }
               }
               else if (strcmp(map[i].opus_field, "qual_file") == 0)
               {  /* replace first file extension character to B */
                  tmp_file[strlen(map[i].field_value.cval)-3] = 'B';
               }

               /* delete IRAF logical from synphot "throughput" filename */
               if ((strcmp(instr_abbrev, "M") == 0) &&  /* synphot */
                   (strcmp(reference_file_type, "THROUGHPUT") == 0))
               {
                  delete_file_prefix(tmp_file);
               }

               sprintf (field_data, "'%s', ", tmp_file);
            }
            else
               sprintf (field_data, "'%s', ", map[i].field_value.cval);

            break;

      }   /* swith */

      strcat (command_buffer, field_data);
      i++;
   }

   /* includes @load_date, and instrument data (for caltable only) */
   if (strcmp(opus_table, "caltable") == 0)
   {
      /* get instrument abbreviation for insert to caltable.instrument field */
      sprintf(field_data, "@load_date, '%s')", instr_abbrev);
      strcat (command_buffer, field_data);
   }
   else
      strcat (command_buffer, "@load_date)");

   /* write insert opus table data statement to output file.
      80 characters per line */
   write_command (command_buffer); 
  
}   /* insert_opus_data */

/*
*******************************************************************************
*
* Function:     delete_opus_data
*
* Purpose:      generate delete record sql from opus table and write to 
*               output file
*
* Returns:      none
*
*******************************************************************************
*/

/* FunctionDefBegin */

void delete_opus_data (
   const int  idx,                /* i: number of reccord in map array */
   const char *instrument,        /* i: instrument name     */
   const char *opus_table,        /* i: opud table name     */
   const char *reference_file_type,  /* i: reference_file_type */
   const char *instr_abbrev)

/* FunctionDefEnd */

{
   char command_buffer[2000];  /* query                                    */
   char field_data[100];       /* retrieved cdbs column value              */
   char tmp_file[100];         /* file_name value                          */
   char cdbs_date[30];         /* convert cdbs date to yyyy:mm:dd:hh:mi:ss */
   char use_after[30];         /* convert above date to yyyy.ddd:hh:mi:ss  */
   char str_fmt[15];           /* string format for print file_name        */
   int  i; 

   sprintf (command_buffer, "delete from %s where ", opus_table);
   i = 0;
   while (i < idx)
   {
      strcat (command_buffer, map[i].opus_field);
      switch (map[i].field_type)
      {
         /* map[i].field_value were retrieved from function write_opus_query */
         case CDBS_SMALLINT:
            sprintf (field_data, " = %d", map[i].field_value.sval);
            break;

         case CDBS_INT:
            sprintf (field_data, " = %d", map[i].field_value.ival);
            break;

         case CDBS_FLOAT:
            sprintf (field_data, " = %f", map[i].field_value.dval);
            break;

         case CDBS_STRING:
            if (strcmp(map[i].cdbs_field, "file_name") == 0)
            {
               strcpy(tmp_file, map[i].field_value.cval);

               if (strcmp(map[i].opus_field, "data_file") == 0)
               {  
                  /* data_file attribute in ccaa_wfc_ref, ccau_wf2_ref
                     ccan_nicmos_ref, ccao_stis_ref are header file name, 
                     rather than the data file name */
                  if ((strcmp(opus_table, "ccaa_wfc_ref") != 0) &&
                      (strcmp(opus_table, "ccau_wf2_ref") != 0) &&
                      (strcmp(opus_table, "ccan_nicmos_ref") != 0) &&
                      (strcmp(opus_table, "ccao_stis_ref") != 0))
                  {
                     /* replace last file extension character to D */
                     tmp_file[strlen(map[i].field_value.cval)-1] = 'D';
                  }
               }
               else if (strcmp(map[i].opus_field, "qual_file") == 0)
               {  /* replace first file extension character to B */
                  tmp_file[strlen(map[i].field_value.cval)-3] = 'B';
               }

               /* delete IRAF logical from synphot "throughput" filename */
               if ((strcmp(instr_abbrev, "M") == 0) &&  /* synphot */
                   (strcmp(reference_file_type, "THROUGHPUT") == 0))
               {
                  delete_file_prefix(tmp_file);
               }

               sprintf (field_data, " = '%s'", tmp_file);
            }
            else
               sprintf (field_data, " = '%s'", map[i].field_value.cval);

            break;

         case CDBS_DATE:  /* useafter_date */
            /* convert cdbs_field to string "yyyy:mm:dd:hh:mi:ss" */
            sprintf (cdbs_date, "%4d:%02d:%02d:%02d:%02d:%02d",
                     map[i].field_value.use_after.dateyear,
                     map[i].field_value.use_after.datemonth+1,
                     map[i].field_value.use_after.datedmonth,
                     map[i].field_value.use_after.datehour,
                     map[i].field_value.use_after.dateminute,
                     map[i].field_value.use_after.datesecond);

            /* convert mm:dd to julian date */
            use_after_date (cdbs_date, use_after);
            sprintf (field_data, " = '%s'",  use_after);
            break;

      }   /* swith */

      strcat (command_buffer, field_data);
      i++;
      if (i < idx)
         strcat (command_buffer, " and ");
   }

   if (strcmp(opus_table, "caltable") == 0)
   {
      sprintf(field_data, " and instrument = '%s'", instr_abbrev);
      strcat (command_buffer, field_data);
   } 

   /* write delete opus table data statement to output file.
      80 characters per line */
   write_command (command_buffer); 
  
}   /* delete_opus_data */

/*
*******************************************************************************
*
* Function:     write_opus_query
*
* Purpose:      retrieve cdbs fields data, create sql statement to update
*               opus table and write sql to output file
*               
*               o call function bind_cdbs_columns to bind cdbs fields
*               o call function create_cdbs_query to generate query
*               o execute the query to retrieve cdbs fields data
*               o if retrieved operation data = 'I' call insert_opus_data to
*                 generate insert sql and write to output file, if operation
*                 data = 'R' or 'D' call delete_opus_data to generate delete
*                 sql and write to output file
*
* Returns:      none
*
*******************************************************************************
*/

/* FunctionDefBegin */

void write_opus_query (
   const int  idx,                  /* i: number of record in map array       */
   const char *instrument,          /* i: instrument name                     */
   const char *file_name,           /* i: file name                           */
   const int  expansion_number,     /* i: expansion number                    */
   const char *operation,           /* i: operation flag, I=insert;R,D=delete */
   const char *reference_file_type, /* i: reference_file_type                 */
   const char *opus_table,          /* i: opus table name                     */
   const char *opus_field_list,     /* i: list of opus columns for query      */
   const char *instr_abbrev)

/* FunctionDefEnd */

{
   char command_buffer[2000];

   if (stdb_query_setup(db_proc2, "write_opus_query", 1) == STDB_ERROR)
       exit(1);

   /* bind cdbs fields */
   bind_cdbs_columns (idx);

   /* generate query and write the query to command_buffer */
   create_cdbs_query (command_buffer, idx, instrument, file_name, 
                      expansion_number);

   if (stdb_query(db_proc2, command_buffer) == STDB_ERROR)
      exit(1);

   while (stdb_get_exec(db_proc2) == STDB_CONTINUE)
   {
      if (strcmp(operation, "I") == 0)
          insert_opus_data (idx, reference_file_type, instrument, 
                            opus_table, opus_field_list, instr_abbrev);
      else if ((strcmp(operation, "R") == 0) || (strcmp(operation, "D") == 0))
          delete_opus_data (idx, instrument, opus_table, reference_file_type,
                            instr_abbrev);
   }

   stdb_close_query(db_proc2);

}   /* write_opus_query */


/*
*******************************************************************************
*
* Function:     create_sql_file
*
* Purpose:      o retrieve file_name, expansion_number, instrument,
*                 reference_file_type, operation from opus_log table having
*                 delivery_number equals to the given number
*               o call function determine_opus_table to get opus_table name
*               o call function map_table_field to create map array which stores
*                 cdbs table name, fields name and corresponding opus table,
*                 fields name and filed data type
*               o call function create_opus_field_list, write_opus_query to
*                 retrieve cdbs fields data, create sql statement to update
*                 opus table and write sql to output file
*
* Returns:      none
*
*******************************************************************************
*/

/* FunctionDefBegin */

void create_sql_file (
   const int  delivery_number,         /* i:   delivery number */
   const char *instr_abbrev)

/* FunctionDefEnd */

{
   int  idx;                     /* number of reccord in map array */
   int  expansion_number;        /* opus_log.expansion_number      */
   char file_name[51];           /* opus_log.file_name             */
   char instrument[51];          /* opus_log.instrument            */
   char reference_file_type[51]; /* opus_log.reference_file_type   */
   char operation[2];            /* opus_log.operation             */
   char type[2];                 /* I=image, T=type                */
   char opus_table[51];          /* relation_info.opus_table       */
   char opus_field_list[500];    /* list of opus table fields      */
   char cmd_buf[500];            /* command buffer                 */

   char old_reference_file_type[51]; /* last used reference_file_type   */
   char old_instrument[51];          /* last used instrument            */

   old_reference_file_type[0] = '\0';
   old_instrument[0] = '\0';

   if (stdb_query_setup(db_proc1, "create_sql_file", 1) == STDB_ERROR)
       exit(1);

   if ((stdb_int_param(db_proc1, "delivery_number", delivery_number) == STDB_ERROR) ||
       (stdb_bind_int(db_proc1, "expansion_number", &expansion_number) == STDB_ERROR) ||
       (stdb_bind_string(db_proc1, "file_name", file_name, 0) == STDB_ERROR) ||
       (stdb_bind_string(db_proc1, "instrument", instrument, 0) == STDB_ERROR) ||
       (stdb_bind_string(db_proc1, "reference_file_type", reference_file_type, 0) == STDB_ERROR) ||
       (stdb_bind_string(db_proc1, "operation", operation, 0) == STDB_ERROR))
       exit(1);

   strcpy(cmd_buf, "select file_name, expansion_number, instrument,");
   strcat(cmd_buf, "reference_file_type, operation from opus_log ");
   strcat(cmd_buf, "where delivery_number = @delivery_number");

   if (stdb_query(db_proc1, cmd_buf) == STDB_ERROR)
      exit(1);

   while (stdb_get_exec(db_proc1) == STDB_CONTINUE)
   {   
      /* get opus table name from <relation_info> */
      determine_opus_table(reference_file_type, instrument, type, opus_table);

      if (!error_out)
      {
         /* create map array which get opus<->cdbs map fields from */
         /* <opus_fields> */
         if ( (strcmp(reference_file_type, old_reference_file_type) != 0) |
              (strcmp(instrument, old_instrument) != 0) )
         {
            idx = 0;
            map_table_field (opus_table,instrument,type,reference_file_type,
                             &idx);
            strncpy(old_reference_file_type, reference_file_type,
                    strlen(reference_file_type)+1);
            strncpy(old_instrument, instrument, strlen(instrument)+1);
         }

         if (idx != 0)
         {
            create_opus_field_list (opus_field_list, idx, opus_table);

            /* retrieve cdbs data, create/write insert/delete query to output */
            write_opus_query (idx, instrument, file_name, expansion_number,
              operation, reference_file_type, opus_table, opus_field_list, 
              instr_abbrev);
         }
         else
         {
            printf("Failed on map cdbs<->opus table and/or field \n");
            error_out = TRUE;
         }
      }
      else
      {
         break;
      }
   }

   if (stdb_record_count(db_proc1) == 0)
   {
      printf("No record in opus_log table match delivery_number %d\n",
              delivery_number );
      error_out = TRUE;
   }

   stdb_close_query(db_proc1);

}   /* create_sql_file */

int main (
   int   argc,
   char *argv[])
{
   char server[STR+1];        /* p1, server name                      */
   char dbname[STR+1];        /* p2, database name                    */
   int  delivery_number;      /* p3, delivery number                  */
   char output_file[LINE+1];  /* output file name opus_n.sql          */
                              /* where n = delivery number            */
   char instr_abbrev[2];      /* instrument abbreviation              */
   char cmd_buf[500];         /* command buffer                       */
   char *tmp;

   /* get environment variables : CDBS_SERVER(server), CDBS_DB(dbname) */
   if ((tmp = getenv("CDBS_SERVER")) != NULL)
   {
      strcpy (server, tmp);
      stdb_string_upper(server);   /* convert server to uppercase */
   }
   else
   {
      printf("CDBS_SERVER must be set to the server name\n");
      exit(1);
   }

   if ((tmp = getenv("CDBS_DB")) != NULL)
   {
      strcpy(dbname, tmp);
      stdb_string_lower(dbname);   /* convert dbname to lowercase */
   }
   else
   {
      printf("CDBS_DB must be set to the CDBS database name\n");
      exit(1);
   }

   setup_db_connection (server, dbname);

   /* determine delivery number */
   if (argc < 2)
   {
      delivery_number = get_delivery_number();
   }
   else
   {
      delivery_number = atoi(argv[1]);
   }

   /* determine output file name */
   strcpy(instr_abbrev, instrument_abbrev(delivery_number));
   stdb_string_lower(instr_abbrev);   /* convert to lowercase */
   sprintf(output_file, "opus_%d_%s.sql", delivery_number, instr_abbrev);

   /* open output file and save getdate() to @load_date */
   init_output(output_file, delivery_number);

   /* create load opus sql file */
   stdb_string_upper(instr_abbrev);   /* convert to lowercase */
   create_sql_file(delivery_number, instr_abbrev);

   /* disconnect from database */
   stdb_disconnect_dbms(db_proc1);
   stdb_disconnect_dbms(db_proc2);
   stdb_disconnect_dbms(db_proc3);

   if (error_out)
   {
      /* on error, remove incompete output file */
      fclose (out);
      sprintf(cmd_buf, "rm %s", output_file);
      error_out = system(cmd_buf);
   }
   else
   {
      /* commit transaction and close output file */
      fprintf(out, "print \"OPUS_MODS_%%1!\", \"OK\"\n\n");
      fprintf(out, "commit tran\n");
      fprintf(out, "go\n");
      fclose(out);
      printf("\ncreated %s\n", output_file);
   }

   return 0;
}
