/*
 ******************************************************************************
 * Program: recreate_load_file
 *
 * Purpose: This program queries the database for information about a 
 *          given reference file and generates a load file containing that
 *          information. This is intended to be used to allow for editing
 *          the load file comment fields for updating the database.
 *
 * Arguments: Reference file name
 *
 * Inputs: None
 *
 * Output: A load file
 * 
 ******************************************************************************
 */ 

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "ostdb.h"
#include "loadutils.h"

#define OK 0
#define ERROR -1
#define NORECS -2

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#define DATELEN 30
#define TEXTLEN 120000
#define MAXMODE 30
#define MAXSTR  256

#define OPS_SERVER "CATLOG"
#define OPS_DB "cdbs_ops"
#define CDBS_USER "calread"
#define CDBS_PASSWORD "caliblook"

typedef struct{
  char  name[31];
  char  value[MAXSTR];
  int   isquoted;
}Mode;

char server[21];
char db[21];


/*
*****************************************************************************
*
* Function: GetInst
*
* Purpose: Queries the database table instrument_abbreviations to determine
*          the instrument with which a CDBS file is associated. 
*
*****************************************************************************
*/

char *GetInst(char *filename){

 static char instrument[31];

  /*
   * Get the instrument name from the database
   */
  ostdb_query_setup(STDB_CONN_1, "GetInst", 1, STDB_SQL, server,
		    db, NULL);
  stdb_stored_command(STDB_CONN_1, "instrument_name");
  stdb_string_param(STDB_CONN_1, "file", filename);
  stdb_bind_string(STDB_CONN_1, "instrument", instrument, 0);


  while(ostdb_get_exec(STDB_CONN_1) == STDB_CONTINUE){
  }
 
  stdb_close_query(STDB_CONN_1);
  if (ostdb_get_status(STDB_CONN_1) != OSTDB_NORMAL){
    return NULL;
  }

  if (strcmp(instrument, "UNKNOWN") == 0){
    return NULL;
  }
    
  return instrument;
}


/*
*****************************************************************************
*
* Function: GetModes
*
* Purpose: Queries the database table cdbs_mode to determine which modes are
*          used for a particular reference file type. It allocates a structure
*          to hold the name and value of the keyword and inserts its address
*          into the mode_key array that was passed in.
*
* Returns: The number of mode keywords for the reference file type
*
*****************************************************************************
*/
int GetModes(char *instrument, char *reference_file_type, Mode *mode_key[]){

  char table[31];
  char modes[31];
  int i;

  sprintf(table, "%s_row", instrument);
  LowCase(table);
  
  /*
   * Get the column names in the instrument row table
   */
  ostdb_query_setup(STDB_CONN_1, "GetModes", 1, STDB_SQL, server,
		    db, NULL);
  stdb_string_param(STDB_CONN_1, "instrument", instrument);
  stdb_string_param(STDB_CONN_1, "reference_file_type", reference_file_type);
  stdb_bind_string(STDB_CONN_1, "modes", modes, 0);
  stdb_query(STDB_CONN_1,"select modes from cdbs_mode");
  stdb_query(STDB_CONN_1, "where upper(instrument) = @instrument");
  stdb_query(STDB_CONN_1, "and upper(reference_file_type) = @reference_file_type");

  i = 0;
  while(ostdb_get_exec(STDB_CONN_1) == STDB_CONTINUE){
    mode_key[i] = (Mode *)malloc(sizeof(Mode));
    strcpy(mode_key[i]->name, UpCase(modes));
    mode_key[i]->isquoted = QuotedField(server, db, table, LowCase(modes));
    i++;
  }

  stdb_close_query(STDB_CONN_1);

  if (ostdb_get_status(STDB_CONN_1) != OSTDB_NORMAL){
    return ERROR;
  }

  return i;
}


/*
*****************************************************************************
*
* Function: GetChangeLevel
*
* Purpose: Determines the CHANGE_LEVEL of a load file row by comparing its
*          equivalence class info to that of its comparison file
*
*****************************************************************************
*/

char *GetChangeLevel(int equiv_sev, int equiv_mod, char *comparison_file, 
		     char *instrument, Mode *mode_key[], int num_modes){

  char table[31];
  int old_equiv_sev;
  int old_equiv_mod;
  int i;
  char tmp[256];

  sprintf(table, "%s_row", instrument);
  LowCase(table);

  /*
   * Get the equivalence class info for the comparison file
   */
  ostdb_query_setup(STDB_CONN_2, "GetChangeLevel", 1, STDB_SQL, server,
		    db, NULL);


  stdb_bind_int(STDB_CONN_2, "equivalence_class_severe", &old_equiv_sev);
  stdb_bind_int(STDB_CONN_2, "equivalence_class_moderate", &old_equiv_mod);

  stdb_query(STDB_CONN_2,"select equivalence_class_severe,");
  stdb_query(STDB_CONN_2,"equivalence_class_moderate");
  sprintf(tmp, "from %s where file_name = '%s'", table, comparison_file);
  stdb_query(STDB_CONN_2,tmp);  

  for (i=0; i<num_modes; i++){
    if (mode_key[i]->isquoted){
      sprintf(tmp, "and %s = '%s'", mode_key[i]->name, mode_key[i]->value);
    }else{
      sprintf(tmp, "and %s = %s", mode_key[i]->name, mode_key[i]->value);
    }
    stdb_query(STDB_CONN_2, tmp);
  }

  while(ostdb_get_exec(STDB_CONN_2) == STDB_CONTINUE){
  }

  if (stdb_record_count(STDB_CONN_2) < 1){
    stdb_close_query(STDB_CONN_2);
    return NULL;
  }

  stdb_close_query(STDB_CONN_2);
  if (ostdb_get_status(STDB_CONN_2) != OSTDB_NORMAL){
    return NULL;
  }

  /*
   * Compare the row's equivalence class info with that from the 
   * comparison file to determine the change level
   */

  if ((equiv_sev != old_equiv_sev) && (equiv_mod != old_equiv_mod)){
    return "SEVERE";
  }
  if ((equiv_sev == old_equiv_sev) && (equiv_mod != old_equiv_mod)){
    return "MODERATE";
  }
  if ((equiv_sev == old_equiv_sev) && (equiv_mod == old_equiv_mod)){
    return "TRIVIAL";
  }

  return NULL;

}


/*
******************************************************************************
* Function: PrintComment
*
* Purpose: Print a comment in the load file, wrapping at column 75
*  
******************************************************************************
*/

void PrintComment(char *comment, FILE *load_file){

  char *start = comment;
  char *next;
  int chunk;
  char tmp [76];
   
  chunk = 65;
  next = start;
    

  fprintf(load_file, "COMMENT = ");
  while(strlen(start) > chunk){
    for (next=start+chunk; next>start; next--){
      /*
       * Break the line at white space
       */
      if(isspace(*next)){
	break;
      }
    }

    /*
     * Got back to the beginning with no white space
     */
    if(next == start){ 
      strncpy(tmp, start, chunk);
      next = start + chunk;
      tmp[chunk] = '\0';
    }
    else{
      strncpy(tmp, start, next-start);
      tmp[next-start] = '\0';
    }

    chunk = 75;
    fprintf(load_file, "%s\n", tmp);
    start = next +1;
  }

  fprintf(load_file, "%s\n", next);

}


/*
******************************************************************************
* Function: GetPutFileLevel
*
* Purpose: Get the file-level keywords from the database and output them
*          to the load file
*  
******************************************************************************
*/

int  GetPutFileLevel(char *instrument, char *file_name, FILE *load_file, 
		     char *reference_file_type, char *comparison_file){

  char useafter_date[DATELEN+1];
  char opus_flag[2];
  char comment[TEXTLEN+1];
  int clen;

  char table[31];

  sprintf(table, "%s_file", instrument);
  LowCase(table);

  ostdb_query_setup(STDB_CONN_1, "GetPutFileLevel", 1, STDB_SQL, server,
                    db, NULL);

  stdb_string_param(STDB_CONN_1, "file_name", file_name);

  stdb_bind_string(STDB_CONN_1, "reference_file_type", reference_file_type, 0);
  stdb_bind_string(STDB_CONN_1, "comparison_file_name", comparison_file, 0);
  stdb_bind_string(STDB_CONN_1, "useafter_date", useafter_date, DATELEN);
  stdb_bind_string(STDB_CONN_1, "opus_flag", opus_flag, 0);
  stdb_bind_text(STDB_CONN_1, "comment", comment, &clen);

  stdb_query(STDB_CONN_1, "select reference_file_type, comparison_file_name");
  stdb_query(STDB_CONN_1, ",useafter_date, opus_flag, comment from");
  stdb_query(STDB_CONN_1, table);
  stdb_query(STDB_CONN_1, "where file_name = @file_name");

  while(ostdb_get_exec(STDB_CONN_1) == STDB_CONTINUE){
  }
  if (stdb_record_count(STDB_CONN_1) < 1){
    stdb_close_query(STDB_CONN_1);
    return NORECS;
  };

  stdb_close_query(STDB_CONN_1);
  if (ostdb_get_status(STDB_CONN_1) != OSTDB_NORMAL){
    return ERROR;
  }

  comment[clen] = '\0';

  /*
   * Output file-level keywords
   */
  fprintf(load_file, "FILE_NAME = %s\n", file_name);
  fprintf(load_file, "INSTRUMENT = %s\n", instrument);
  fprintf(load_file, "REFERENCE_FILE_TYPE = %s\n", reference_file_type);
  useafter_date[11] = '\0'; /* get rid of time portion */
  fprintf(load_file, "USEAFTER_DATE = %s\n", useafter_date);
  fprintf(load_file, "COMPARISON_FILE = %s\n", comparison_file);
  fprintf(load_file, "OPUS_FLAG = %s\n", opus_flag);
  PrintComment(comment, load_file);
  fprintf(load_file, "ENDHEADER\n\n");

  return OK;
}


/*
******************************************************************************
* Function: GetPutRowLevel
*
* Purpose: Get the row-level keywords from the database and output them
*          to the load file
*  
******************************************************************************
*/

int  GetPutRowLevel(char *instrument, char * file_name, FILE *load_file,
		    char *reference_file_type, char *comparison_file){

  Mode *mode_key[MAXMODE];
  int  expansion_number;
  char pedigree[51]; 
  char observation_begin_date[DATELEN+1];
  char observation_end_date[DATELEN+1];
  int  equivalence_class_severe;
  int  equivalence_class_moderate;
  char comment[TEXTLEN+1];
  int clen;
  char table[31];
  int num_modes; 
  int i;

  /*
   * Determine row-level mode keywords
   */
  num_modes = GetModes(instrument, reference_file_type, mode_key);

  sprintf(table, "%s_row", instrument);
  LowCase(table);

  ostdb_query_setup(STDB_CONN_1, "GetPutRowLevel", 1, STDB_SQL, server,
		    db, NULL);

  stdb_string_param(STDB_CONN_1, "file_name", file_name);
 
  stdb_bind_int(STDB_CONN_1, "expansion_number", &expansion_number);
  stdb_bind_int(STDB_CONN_1, "equivalence_class_severe", 
		&equivalence_class_severe);
  stdb_bind_int(STDB_CONN_1, "equivalence_class_moderate", 
		&equivalence_class_moderate);
  stdb_bind_string(STDB_CONN_1, "pedigree", pedigree, 0);
  stdb_bind_string(STDB_CONN_1, "observation_begin_date", 
		   observation_begin_date, DATELEN);
  stdb_bind_string(STDB_CONN_1, "observation_end_date", 
		   observation_end_date, DATELEN);
  stdb_bind_text(STDB_CONN_1, "comment", comment, &clen);

  /*
   * Bind mode keywords
   */
  for (i=0; i<num_modes; i++){
    stdb_bind_string(STDB_CONN_1, mode_key[i]->name, mode_key[i]->value, 0);
  }
      
  stdb_query(STDB_CONN_1, "select expansion_number,");
  stdb_query(STDB_CONN_1, "equivalence_class_severe,");
  stdb_query(STDB_CONN_1, "equivalence_class_moderate,");
  stdb_query(STDB_CONN_1, "pedigree, observation_begin_date,");
  stdb_query(STDB_CONN_1, "observation_end_date,");

  /*
   * Add mode keywords into query
   */
  for (i=0; i<num_modes; i++){
    stdb_query(STDB_CONN_1, mode_key[i]->name);
    stdb_query(STDB_CONN_1, ",");
  }

  stdb_query(STDB_CONN_1, "comment from");
  stdb_query(STDB_CONN_1, table);
  stdb_query(STDB_CONN_1, "where file_name = @file_name");

  while(ostdb_get_exec(STDB_CONN_1) == STDB_CONTINUE){
    
    if (expansion_number > 0){
      fprintf(load_file, "EXPANSION_NUMBER = %-d\n", expansion_number);
    }

    if ((strcmp(comparison_file, "(INITIAL)") == 0) || 
	(strcmp(comparison_file, "UNKNOWN") == 0) ||
	(comparison_file[0] == '\0')){
      fprintf(load_file, "CHANGE_LEVEL = SEVERE\n");
    }else{
      if (strcmp(instrument, "MULTI") == 0){
      fprintf(load_file, "CHANGE_LEVEL = \n");
      }else{
	fprintf(load_file, "CHANGE_LEVEL = %s\n", 
		GetChangeLevel(equivalence_class_severe, 
			       equivalence_class_moderate, comparison_file, 
			       instrument, mode_key, num_modes));
      }
    }
    fprintf(load_file, "PEDIGREE = %s\n", pedigree);
    observation_begin_date[11] = '\0'; /* get rid of time portion */
    fprintf(load_file, "OBSERVATION_BEGIN_DATE = %s\n", 
	    observation_begin_date);
    observation_end_date[11] = '\0'; /* get rid of time portion */
    fprintf(load_file, "OBSERVATION_END_DATE = %s\n", 
	    observation_end_date);

    for (i=0; i<num_modes; i++){
      fprintf(load_file, "%s = %s\n", mode_key[i]->name,
	      mode_key[i]->value);
    }

    comment[clen] = '\0';

    PrintComment(comment, load_file);
    
    fprintf(load_file, "ENDROW\n\n");
  }

  if (stdb_record_count(STDB_CONN_1) < 1){
    stdb_close_query(STDB_CONN_1);
    return NORECS;
  };

  stdb_close_query(STDB_CONN_1);
  if (ostdb_get_status(STDB_CONN_1) != OSTDB_NORMAL){
    return ERROR;
  }
  
  return OK;
}


/*
******************************************************************************
*
* Function: main
*  
******************************************************************************
*/
  
main (int argc, char **argv){

  char file_name[51];
  char *instrument;
  char load_file_name[51];
  FILE *load_file;
  char reference_file_type[51];
  char comparison_file[51];
  int status;
  char *env;
  int len;

  if (argc < 2){
    printf("Filename: ");
    gets(file_name);
  }else{
    strcpy(file_name, UpCase(argv[1]));
  }
  UpCase(file_name);

  /*
   * Use CDBS_SERVER and CDBS_DB environment variables if set, otherwise
   * default to OPS version of db and server
   */
  if ((env = getenv("CDBS_SERVER")) != NULL){
    strcpy(server, env);
  }else{
    strcpy(server, OPS_SERVER);
  }

  if ((env = getenv("CDBS_DB")) != NULL){
    strcpy(db, env);
  }else{
    strcpy(db, OPS_DB);
  }

  /*
   * set up STDB so that all of the database connections are for a set 
   * username and password with readonly access
   */
  ostdb_setup(STDB_CONN_1, STDB_USERNAME, CDBS_USER);
  ostdb_setup(STDB_CONN_1, STDB_PASSWORD, CDBS_PASSWORD);
  ostdb_setup(STDB_CONN_2, STDB_USERNAME, CDBS_USER);
  ostdb_setup(STDB_CONN_2, STDB_PASSWORD, CDBS_PASSWORD);
  ostdb_setup(STDB_CONN_3, STDB_USERNAME, CDBS_USER);
  ostdb_setup(STDB_CONN_3, STDB_PASSWORD, CDBS_PASSWORD);

  printf("Processing file %s...\n", file_name);

  if ((instrument = GetInst(file_name)) == NULL){
    printf("Unable to determine the instrument from the input filename\n");
    exit(1);
  }    

  strcpy(load_file_name, file_name);
  load_file_name[strcspn(load_file_name, ".")] = '\0';
  strcat(load_file_name, ".lod");
  LowCase(load_file_name);

  load_file = fopen(load_file_name, "w");

  /*
   * Get and output the row-level keywords
   */
  if ((status = GetPutFileLevel(instrument, file_name, load_file,
				reference_file_type, comparison_file)) != OK){
    if (status == NORECS){
      printf("Cannot find file-level entry in the database\n");
    }else{
      printf("Error querying the database for file-level record\n");
    }
    exit(1);
  }

  /*
   * Get and output row-level keywords
   */
  if ((status = GetPutRowLevel(instrument, file_name, load_file,
				reference_file_type, comparison_file))!= OK){
    if (status == NORECS){
      printf("Cannot find row-level entry in the database\n");
    }else{
      printf("Error querying the database for row-level record\n");
    }
    exit(1);
  }  

  fprintf(load_file, "ENDFILE\n");

  printf ("Load file: %s has been generated\n", load_file_name);

  exit(0);

}

