#include <time.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "config.h"
#include "error.h"
#include "cdbslib.h"

#define  SZ_DATE	63
#define  AM		1
#define  PM		2

typedef struct {
int	year;
int	month;
int	day;
int	hour;
int	min;
int	sec;
} calendar;

static char *month_name[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
			     "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

static char *month_list = "january|february|march|april|may|june|july|august\
|september|october|november|december";

static char *meridian_list = "am|pm";

static int 	check_date (calendar *);
static int	field_date (char *);
static void 	inverse_julian_date (double, calendar *); 
static double 	julian_date (calendar *);

/* CHECK_DATE -- Check date to see if it is in the valid range
 *
 * i: clock	time structure
 */

static int check_date (calendar *clock)
{
    static int mdays[2][12] = {
	{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
	{31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
    };

    int		leap;

    /* year	two or four digit year in CE
     * month	month number, 1 - 12
     * day	day of month, 1 - 31
     * hour	hour on 24 hour clock, 0 - 23
     * min	minute, 0 - 59
     * sec	second, 0 - 59
     */

    if (clock->year < 1901 || clock->year > 2099)
	return (ERR);

    if (clock->month < 1 || clock->month > 12)
	return (ERR);

    leap = (clock->year % 4) == 0;

    if (clock->day > mdays[leap][clock->month-1])
	return (ERR);

    if (clock->hour < 0 || clock->hour > 23)
	return (ERR);

    if (clock->min < 0 || clock->min > 59)
	return (ERR);

    if (clock->sec < 0 || clock->sec > 59)
	return (ERR);

    return (OK);
}

/* FIELD_DATE -- Compute the integer value of a date field
 *
 * i: field	string field, all digits
 */

static int field_date (char *field)
{
    char	*ch;
    int		value;

    value = 0;
    for (ch = field; *ch != EOS; ch ++) {
        if (isdigit (*ch)) {
	    value = 10 * value + (*ch - '0');
	} else {
	    return (-1);
	}
    }

    return (value);
}

/* INVERSE_JULIAN_DATE -- Convert MJD back to calendar date
 *
 * i: mjd	modified julian date
 * o: clock	calendar date
 */

static void inverse_julian_date (double mjd, calendar *clock) 
{
    double	a, b, c, d, e, z, alpha;

    if (mjd < 0.0)
	cdbs_error (NODATE, "%g", mjd);

    z = floor (mjd);
    alpha = floor ((z + 532784.75) / 36524.25);

    clock->sec = 86400.0 * (mjd - z);
    clock->min = clock->sec / 60;
    clock->sec = clock->sec - 60 * clock->min;
    clock->hour = clock->min / 60;
    clock->min = clock->min - 60 * clock->hour;

    /* Inverse MJD formula from chapter 7 of "Astronomical Agorithms" 
     * by Jean Meeus. Formula adjusted from JD to MJD. Variable names 
     * follow those in book.
     */

    a = z + 2400002.0 + alpha - floor (alpha / 4.0);
    b = a + 1524.0;
    c = floor ((b - 122.1) / 365.25);
    d = floor (365.25 * c);
    e = floor ((b - d) / 30.6001);

    clock->day = b - d - floor (30.6001 * e);
    clock->month = e - 1.0;
    clock->year = c - 4716.0;

    if (clock->month > 12) {
	clock->month = clock->month - 12;
	clock->year = clock->year + 1;
    }
}

/* JULIAN_DATE -- Compute the modified julian date 
 *
 *
 * u: clock	time structure
 */

static double julian_date (calendar *clock)
{
    int		century, gregory;
    double	mjd;

    /* Convert two digit year to four digits */

    if (clock->year >= 0 && clock->year < 50) {
	clock->year = clock->year + 2000; 
    } else if (clock->year < 100) {
	clock->year = clock->year + 1900;
    }

    /* Check date for legality */

    if (check_date (clock) == ERR)
	return (-1.0);

    /* Formula for modified julian date, 1901 to 2099
    /* Formula from chapter 7 of "Astronomical Agorithms" 
     * by Jean Meeus. Formula adjusted from JD to MJD.
     */

    if (clock->month <= 2) {
	clock->year = clock->year - 1;
	clock->month = clock->month + 12;
    }

    mjd = floor (365.25 * (clock->year + 4716)) + 
	  floor (30.6001 * (clock->month + 1)) +
	  clock->day - 2401538.0;

    mjd = mjd + (((clock->hour * 60.0) + clock->min) * 60.0 + clock->sec) /
	        86400.0;
    return (mjd);
}

/* GET_CDBS_DATE -- Convert a date from old CDBS format to MJD
 *
 * i: date	date string in yyyymmdd:hhmmss format (old cdbs date format)
 */

double get_cdbs_date (const char *date)
{
    char	ch;
    char	str[SZ_DATE+1];
    calendar	clock;

    if (strlen (date) < 15)
	return (-1.0);

    if (date[8] != ':')
        return (-1.0);

    safecopy (str, date, SZ_DATE);

    ch = str[4];
    str[4] = EOS;
    clock.year = field_date (str);
    str[4] = ch;

    ch = str[6];
    str[6] = EOS;
    clock.month = field_date (&str[4]);
    str[6] = ch;

    ch = str[8];
    str[8] = EOS;
    clock.day = field_date (&str[6]);

    ch = str[11];
    str[11] = EOS;
    clock.hour = field_date (&str[9]);
    str[11] = ch;

    ch = str[13];
    str[13] = EOS;
    clock.min = field_date (&str[11]);
    str[13] = ch;

    ch = str[15];
    str[15] = EOS;
    clock.sec = field_date (&str[13]);

    return (julian_date (&clock));
}

/* GET_NOW -- Get the current time as a Julian date
 *
 */

double get_now (void)
{
    time_t	tick;
    struct tm	*now;
    calendar	clock;

    tick = time (NULL);
    now = localtime (&tick);

    clock.year = now->tm_year + 1900;
    clock.month = now->tm_mon + 1;
    clock.day = now->tm_mday;
    clock.hour = now->tm_hour;
    clock.min = now->tm_min;
    clock.sec = now->tm_sec;

    return (julian_date (&clock));
}

/* GET_SLASH_DATE -- Convert a date in dd/mm/yy format to 
 * a modified julian date 
 *
 * i: date	date string in dd/mm/yy format (fits date format)
 */

double get_slash_date (const char *date) 
{
    char	*time[3];
    char	str[SZ_DATE+1];
    int		day, month, year;
    calendar	clock;

    safecopy (str, date, SZ_DATE);
    strsplit ('/', str, time, 3);

    clock.day = field_date (time[0]);
    clock.month = field_date (time[1]);
    clock.year = field_date (time[2]);
    clock.hour = 0;
    clock.min = 0;
    clock.sec = 0;

    return (julian_date (&clock));
}

/* GET_SYBASE_DATE -- Converth a date in Sybase format to mjd
 *
 * i: date	date string in (month dd yyyy hh:mm:ss am|pm) format
 */

double get_sybase_date (const char *date)
{
    char	*time[5], *piece[3];
    char	str[SZ_DATE+1];
    calendar 	clock;
    int		meridian;

    safecopy (str, date, SZ_DATE);
    strsplit (' ', str, time, 5);

    clock.month = strabbrev (time[0], month_list);
    if (clock.month == 0)
        return (-1.0);

    clock.day = field_date (time[1]);
    clock.year = field_date (time[2]);

    if (time[3][0] == EOS) {
	clock.hour = 0;
	clock.min = 0;
	clock.sec = 0;

    } else {
	strsplit (':', time[3], piece, 3);
	clock.hour = field_date (piece[0]);
	clock.min = field_date (piece[1]);
	clock.sec = field_date (piece[2]);
    }

    if (time[4][0] != EOS) {
	meridian = strmatch (time[4], meridian_list);
	switch (meridian) {
	case AM:
	    if (clock.hour == 12)
		clock.hour = clock.hour - 12;
	    break;
	case PM:
	    if (clock.hour > 12)
		clock.hour = clock.hour + 12;
	    break;
	default:
	    clock.hour = -1;
	}
    }

    return (julian_date (&clock));
}

/* PUT_SHORT_DATE -- Convert a date to short format (no hh:mm:ss)
 *
 * i: mjd	modified julian date
 * o: date	Sybase date string (month dd yyyy)
 * i: maxch	max length of date string
 */

void put_short_date (double mjd, char *date, int maxch)
{
    calendar	clock;
    char	buf[SZ_DATE+1];

    inverse_julian_date (mjd, &clock);
    sprintf (buf, "%s %d %d", month_name[clock.month-1], 
	     clock.day, clock.year);

    safecopy (date, buf, maxch);
}

/* PUT_SYBASE_DATE -- Convert a date from MJD to Sybase format
 *
 * i: mjd	modified julian date
 * o: date	Sybase date string (month dd yyyy)
 * i: maxch	max length of date string
 */

void put_sybase_date (double mjd, char *date, int maxch)
{
    calendar	clock;
    char	buf[SZ_DATE+1], meridian[3];

    inverse_julian_date (mjd, &clock);
    sprintf (buf, "%s %d %d", month_name[clock.month-1], 
	     clock.day, clock.year);

    safecopy (date, buf, maxch);

    if (mjd - floor (mjd) > 0.0) {
	if (clock.hour < 12) {
	    safecopy (meridian, "am", 2);
	    if (clock.hour < 1)
		clock.hour = clock.hour + 12;
	} else {
	    safecopy (meridian, "pm", 2);
	    if (clock.hour > 12)
		clock.hour = clock.hour - 12;
	}

	sprintf (buf, " %02d:%02d:%02d %s", clock.hour, clock.min, 
		 clock.sec, meridian);

	safecat (date, buf, maxch);
    }
}
