#include <stdio.h>
#include <string.h>
#include "error.h"
#include "config.h"
#include "cdbslib.h"
#include "reffile.h"
#include "loadfile.h"

#define  NKEY		4
#define  SZ_KEYVAL	63
#define  MAXFIELD	10
#define  NPAIR		2

#define  MATCH_ZIP	0
#define  MATCH_EXT	1
#define  MATCH_TYPE	2
#define  MATCH_KEY	3

static char	filename[] = "filename";
static char	extname[] = "extension";
static char	typename[] = "filetype";
static char	*keyname[] = {"keyname1", "keyname2", "keyname3", "keyname4"};
static char	*valname[] = {"keyval1", "keyval2", "keyval3", "keyval4"};

/* CLOSE_LOOKUP -- Close a lookup table
 *
 * i: look	table descriptor
 */

void close_lookup (table * look)
{
    free_table (look);
}

/* FIND_LOOKUP -- Search a table for the row matching a reference file
 *
 * i: look	table descriptor
 * i: fname	reference file name
 */

int find_lookup (table *look, char *fname)
{
    char	dir[SZ_FNAME+1], root[SZ_FNAME+1];
    char	filext[SZ_FNAME+1], fileval[SZ_KEYVAL+1];
    char	*lookext, *looktype, *lookkey, *lookval;
    int		extcol, typecol, keycol[NKEY], valcol[NKEY];
    int		ikey, irow, status;
    reffile	*ref;
    FILE	*fd;

    /* Check for the existence of the file */

    if ((fd = fopen (fname, "r")) == NULL) {
        cdbs_warn (NOTOPEN, "%s", fname);
	return (ERR); 
    }

    fclose (fd);

    /* Get the column numbers of the columns used to search the table */

    if ((extcol = get_tab_colnum (look, extname)) == BAD)
	cdbs_error (NOCOL, "%s", extname);

    if ((typecol = get_tab_colnum (look, typename)) == BAD)
	cdbs_error (NOCOL, "%s", typename);

    for (ikey = 0; ikey < NKEY; ikey ++) {
	if ((keycol[ikey] = get_tab_colnum (look, keyname[ikey])) == BAD)
	    cdbs_error (NOCOL, "%s", keyname[ikey]);

	if ((valcol[ikey] = get_tab_colnum (look, valname[ikey])) == BAD)
	    cdbs_error (NOCOL, "%s", valname[ikey]);
    }

    /* Search for the row which matches the reference file
     * There are two matching criteria: the filname extension
     * and keywords in the reference file header.
     */

    ref = NULL;
    lod_error = NULL;
    status = MATCH_ZIP;
    breakfname (fname, dir, root, filext, SZ_FNAME);

    for (irow = 0; irow < nrow_table(look); irow ++) {
	/* Check the reference file extension against the lookup table */

	lookext = get_tab_cell_bynum (look, extcol, irow);
	if (lookext[0] != EOS && patmatch (lookext, filext) == 0)
	    continue;

	if (status < MATCH_EXT)
	    status = MATCH_EXT;

	/* Open the reference file if it is not already open */

	if (ref == NULL) {
	    looktype = get_tab_cell_bynum (look, typecol, irow);
	    ref = open_reffile (fname, looktype, "r");

	    if (ref == NULL)
	        continue;
	}

	if (status < MATCH_TYPE)
	    status = MATCH_TYPE;

	/* Check the header keywords against the lookup table */

	for (ikey = 0; ikey < NKEY; ikey ++) {
	    lookkey = get_tab_cell_bynum (look, keycol[ikey], irow);

	    if (lookkey[0] != EOS) {
	        lookval = get_tab_cell_bynum (look, valcol[ikey], irow);

		get_hdr_reffile (ref, lookkey, fileval, SZ_KEYVAL);
		strlcase (fileval);

		if (strcmp (lookval, fileval) != 0)
		    break;
	    }
	}

	if (ikey == NKEY) {
	    status = MATCH_KEY;
	    break;
	}
    }

    if (ref != NULL)
        close_reffile (ref);

    /* Print a warning message explaining the reason why 
     * no match was found in the catalog file
     */

    switch (status) {
    case MATCH_ZIP:
        cdbs_warn (BADEXT, "%s", fname);
        irow = ERR;
        break;
    case MATCH_EXT:
        if (lod_error != NULL) {
	    cdbs_warn (NOLOAD, "%s", lod_error);
	    lod_error = NULL;
	}

        cdbs_warn (BADTYPE, "%s", fname);
        irow = ERR;
        break;
    case MATCH_TYPE:
        cdbs_warn (BADKEY, "%s", fname);
        irow = ERR;
        break;
    case MATCH_KEY:
        break;
    }

    /* Return a value of ERR if there was no match.
     * Otherwise return the number of the matched row
     */

    return (irow);
}

/* NAME_LOOKUP -- Retrieve the name of the lookup file from the table header
 *
 * i: look	table descriptor
 */

char *name_lookup (table *look)
{
    return (get_tab_key (look, filename));
}

/* OPEN_LOOKUP -- Read a lookup file into a table structure
 *
 * i: fname	lookup file name
 * i: columns	list of column names
 */

table *open_lookup (char *fname, char *columns)
{
    FILE	*fd;
    table 	*look;
    int		irow, ifield, ikey, keyfield;
    char	collist[SZ_LINE+1], buf[SZ_FNAME+1];
    char	*colname[MAXFIELD], *field[MAXFIELD];
    char	*keypair[MAXFIELD], *keyval[NPAIR];

    /* Open the lookup file */

    fd = fopen (fname, "r");
    if (fd == NULL)
	cdbs_error (NOTOPEN, "%s", fname);

    look = create_table ();
    put_tab_key (look, filename, fname);

    /* Create columns required for searching */

    add_tab_col (look, extname);
    add_tab_col (look, typename);

    for (ikey = 0; ikey < NKEY; ikey ++) {
	add_tab_col (look, keyname[ikey]);
	add_tab_col (look, valname[ikey]);
    }

    /* Split the list of column names and look for the keyword field */

    safecopy (collist, columns, SZ_LINE);
    strsplit (',', collist, colname, MAXFIELD);

    for (keyfield = 0; keyfield < MAXFIELD; keyfield ++) {
	if (strcmp (colname[keyfield], "keyword") == 0)
	    break;
    }

    /* Read the lookup file into the table */

    for (irow = 0; rdlines (fd, buf, SZ_LINE) == OK; irow ++) {
	strsplit (' ', buf, field, MAXFIELD);

	/* Read each field from the line into the table */

	for (ifield = 0; 
	     ifield < MAXFIELD && colname[ifield][0] != EOS; 
	     ifield ++) {

	    if (ifield == keyfield) {
		/* The keyword field is split into a keyword list
		 * and further split into keyword = value pairs.
		 * The values are converted to lower case for a case
		 * independent match.
		 */

		strlcase (field[ifield]);
		strsplit (',', field[ifield], keypair, MAXFIELD);

		for (ikey = 0; ikey < NKEY; ikey ++) {
		    strsplit ('=', keypair[ikey], keyval, NPAIR);
		    put_tab_cell (look, keyname[ikey], irow, keyval[0]);
		    put_tab_cell (look, valname[ikey], irow, keyval[1]);
		}

		if (keypair[ikey][0] != EOS)
		    cdbs_error (MAXKEY, "%s", fname);

	    } else {
		/* All other fields are copied as is to the table */

		put_tab_cell (look, colname[ifield], irow, field[ifield]);
	    }
	}
    }

    fclose (fd);
    return (look);
}
