#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "system.h"
#include "c_iraf.h"
#include "xtables.h"
#include "ximio.h"
#include "config.h"
#include "error.h"
#include "cdbslib.h"
#include "loadfile.h"
#include "reffile.h"

#define  NBITS_BYTE		8

#define  EXTRA_KEYS		"datamin|datamax|bitpix|datatype|simple"

#define  ORDINARY		0
#define  DATAMIN		1
#define  DATAMAX		2
#define  BITPIX			3
#define  DATATYPE		4
#define  SIMPLE			5

#define  MODELIST		"r|r+|w|a"

static int		count_reffile (reffile *);
static int		ext_reffile (char *);
static int		fits_reffile (char *);
static int		image_reffile (IRAFPointer, char *, char *, int);
static void 		reclose_reffile (reffile *);
static int 		reopen_reffile (reffile *, int);
static void		root_reffile (char *, char *, int);

/* ACCESS_REFFILE -- Check access status of a reference file
 *
 * i: fname	file name
 * i: type	file type
 * i: mode	access mode
 */

int access_reffile (char *fname, char *ftype, char *mode)
{
    int		status;
    reffile	*ref;

    /* Currently the onlt reliable way to check access 
     * to a reference file is to try to open it
     */

    ref = open_reffile (fname, ftype, mode);

    if (ref == NULL) {
        status = NO;
    } else {
        close_reffile (ref);
	status = YES;
    }

    return (status);
}

/* CLOSE_REFFILE -- Close the reference file
 *
 * i: ref	reference file descriptor
 */

void close_reffile (reffile *ref) 
{
    reclose_reffile (ref);
    free (ref);
}

/* GET_BODY_REFFILE -- Read a value from the body of a reference file
 *
 * u: ref	reference file descriptor
 * i: name	column name
 * i: row	row number (zero indexed)
 * o: value	keyword value
 * i: maxch	max length of value
 */

int get_body_reffile (reffile *ref, char *name, int row, 
		      char *value, int maxch)
{
    char	*str, *uname;
    int		status;
    IRAFPointer	cptr;

    /* Switch to reading body */

    status = reopen_reffile (ref, row+1);
    if (status == ERR)
        return (status);

    /* Read the value from the table body */

    switch (ref->type) {
    case IMAGE:
	status = image_reffile ((IRAFPointer) ref->ptr, name, value, maxch);
	break;
    case TABLE:
	c_tbcfnd1 ((IRAFPointer) ref->ptr, name, &cptr);
	if (cptr == 0) {
	    value[0] = EOS;
	    status = ERR;

	} else {
	    c_tbegtt ((IRAFPointer) ref->ptr, cptr, row+1, value, maxch);
	    if (c_iraferr () != 0) {
		value[0] = EOS;
		status = ERR;
	    }
	}
	break;
    case LOADFILE:
	uname = strclone (name);
	strucase (uname);

	str = get_tab_cell ((table *) ref->ptr, uname, row);
	if (str != NULL) {
	    safecopy (value, str, maxch);
	} else {
	    value[0] = EOS;
	    status = ERR;
	}

	free (uname);
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    return (status);
}

/* GET_HDR_REFFILE -- Read a header keyword from a reference file
 *
 * u: ref	reference file descriptor
 * i: name	keyword name
 * o: value	keyword value
 * i: maxch	max length of value
 */

int get_hdr_reffile (reffile *ref, char *name, char *value, int maxch)
{
    char	*str, *uname;
    int		status;

    /* Switch to reading header */

    status = reopen_reffile (ref, 0);
    if (status == ERR)
        return (status);

    /* Read header keyword */

    switch (ref->type) {
    case IMAGE:
	status = image_reffile ((IRAFPointer) ref->ptr, name, value, maxch);
	break;
    case TABLE:
	c_tbhgtt ((IRAFPointer) ref->ptr, name, value, maxch);
	if (c_iraferr () != 0) {
	    value[0] = EOS;
	    status = ERR;
	}
	break;
    case LOADFILE:
	uname = strclone (name);
	strucase (uname);

	str = get_tab_key ((table *) ref->ptr, uname);
	if (str != NULL) {
	    safecopy (value, str, maxch);
	} else {
	    value[0] = EOS;
	    status = ERR;
	}

	free (uname);
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    return (status);
}

/* NROW_REFFILE -- Number of rows in a reference file
 *
 * i: ref	reference file descriptor
 */

int nrow_reffile (reffile *ref)
{
    int		nrow;


    switch (ref->type) {
    case IMAGE:
	nrow = c_imgeti ((IRAFPointer) ref->ptr, "GCOUNT");
	if (c_iraferr (0) != 0)
	    nrow = count_reffile (ref);
	break;

    case TABLE:
        /* Switch to reading table body */
        if (reopen_reffile (ref, 1) == ERR) {
	    cdbs_error (NOTOPEN, "%s", name_reffile (ref));

	} else {
	    nrow = c_tbpsta ((IRAFPointer) ref->ptr, TBL_NROWS);
	    if (c_iraferr (0) != 0)
	        cdbs_error (BADNROW, "%s", name_reffile (ref));
	}
	break;

    case LOADFILE:
	nrow = nrow_table ((table *) ref->ptr);
	break;
    }

    return (nrow);
}

/* OPEN_REFFILE -- Open a reference file of the specified type
 *
 * i: name	reference file name
 * i: type	reference file type
 */

reffile *open_reffile (char *fname, char *type, char *mode)
{
    reffile	*ref;

    /* Allocate and intialize the descriptor */

    ref = (reffile *) malloc (sizeof (reffile));

    root_reffile (ref->name, fname, SZ_FNAME);
    ref->type = strabbrev (type, FILETYPE_NAMES);
    ref->mode = strmatch (mode, MODELIST);
    ref->primary = fits_reffile (ref->name);
    ref->smudge = NO;
    ref->ptr = NULL;

    /* Set the row number */

    if ((ref->row = ext_reffile (fname)) == ERR)
        ref->row = 0;

    /* Call reopen to actually open the file */

    if (reopen_reffile (ref, ref->row) == ERR) {
        free (ref);
	ref = NULL;
    }

    /* Return the descriptor */

    return (ref);
}

/* PUT_BODY_REFFILE -- Write a value to the reference file body 
 *
 * u: ref	reference file descriptor
 * i: name	column name
 * i: row	row number (zero indexed)
 * i: value	cell value
 */

void put_body_reffile (reffile *ref, char *name, int row, char *value)
{
    IRAFPointer	cptr;

    /* Switch to accessing body */

    if (reopen_reffile (ref, row+1) == ERR)
        cdbs_error (NOWRITE, ref->name);

    switch (ref->type) {
    case IMAGE:
	cdbs_error (NOWRITE, "%s.%s", ref->name, name);
	break;
    case TABLE:
	c_tbcfnd1 ((IRAFPointer) ref->ptr, name, &cptr);
	if (cptr == 0) {
	    cdbs_error (NOWRITE, "%s.%s", ref->name, name);
	} else {
	    c_tbeptt ((IRAFPointer) ref->ptr, cptr, row+1, value);
	}
	break;
    case LOADFILE:
	put_tab_cell ((table *) ref->ptr, name, row, value);
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    ref->row = row;
    ref->smudge = YES;
}

/* PUT_HDR_REFFILE -- Write a header keyword to a reference file
 *
 * u: ref	reference file descriptor
 * i: name	keyword name
 * i: value	keyword value
 */

void put_hdr_reffile (reffile *ref, char *name, char *value)
{
    /* Switch to accessing header */

    if (reopen_reffile (ref, 0) == ERR)
        cdbs_error (NOWRITE, ref->name);

    /* Write keyword value */

    switch (ref->type) {
    case IMAGE:
	c_imastr ((IRAFPointer) ref->ptr, name, value);
	break;
    case TABLE:
	c_tbhadt ((IRAFPointer) ref->ptr, name, value);
	break;
    case LOADFILE:
	put_tab_key ((table *) ref->ptr, name, value);
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    ref->smudge = YES;
}

/* PUT_HIST_REFFILE -- Write a history keyword to a reference file
 *
 * u: ref	reference file descriptor
 * i: name	keyword name
 * i: value	keyword value
 */

void put_hist_reffile (reffile *ref, char *name, char *value)
{
    /* Switch to accessing header */

    if (reopen_reffile (ref, 0) == ERR)
        cdbs_error (NOWRITE, ref->name);

    /* Write history keyword */

    switch (ref->type) {
    case IMAGE:
	c_imputh ((IRAFPointer) ref->ptr, name, value);
	break;
    case TABLE:
	c_tbhadt ((IRAFPointer) ref->ptr, name, value);
	break;
    case LOADFILE:
	put_tab_key ((table *) ref->ptr, name, value);
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    ref->smudge = YES;
}

/* RENAME_REFFILE -- Change the name of a reference file
 *
 * i: oldname	current file name
 * i: newname	new file name
 * i: type	file type
 */

int rename_reffile (char *oldname, char *newname, char *ftype)
{
    int		type, status;

    status = OK;
    type = strabbrev (ftype, FILETYPE_NAMES);

    switch (type) {
    case IMAGE:
	c_imrename (oldname, newname);
	if (c_iraferr() != 0)
	    status = ERR;
	break;
    case TABLE:
    case LOADFILE:
	if (rename (oldname, newname) != 0)
	    status = ERR;
	break;
    default:
	cdbs_error (FTYPE, "%d", type);
    }

    return (status);
}

/*----------------------------------------------------------------------
 *
 * Procedures below this line are not externally callable
 *
 *----------------------------------------------------------------------
 */

/* COUNT_REFFILE -- Count the number of groups in an image
 *
 * i: ref	reference file descriptor
 */


static int count_reffile (reffile *ref)
{
    IRAFPointer ptr;
    int		hi, lo, mid, step;
    char	fname[SZ_FNAME+1];

    /* The code in this function does not work currently because of
     * a bug in the iraf interface. Send an error message instead.
     */

    cdbs_error (BADNROW, "%s", name_reffile (ref));

    /* Find a group not in the image */

    hi = 0;
    for (step = 1; step > 0; step *= 2) {
        hi = hi + step;
	sprintf (fname, "%s[%d]", ref->name, hi);

	ptr = c_immap (fname, IRAF_READ_ONLY, 0);
	if (c_iraferr() != 0)
	    break;

	c_imunmap (ptr);
    }

    /* Use binary search to find last group in the image */

    lo = hi - step;
    while (hi - lo > 1) {
        mid = (hi + lo) / 2;
	sprintf (fname, "%s[%d]", ref->name, mid);

	ptr = c_immap (fname, IRAF_READ_ONLY, 0);
	if (c_iraferr() != 0) {
	    hi = mid;
	} else {
	    c_imunmap (ptr);
	    lo = mid;
	}
    }

    return (lo);
}

/* EXT_REFFILE -- Check to see if reference filename has an extension
 *
 * i: fname	reference file name
 */

static int ext_reffile (char *fname)
{
    long	ext;
    char	*lbrak, *rbrak;

    lbrak = strrchr (fname, '[');
    if (lbrak == NULL)
        return (ERR);

    ext = strtol (lbrak+1, &rbrak, 10);
    if (*rbrak != ']')
        return (ERR);

    return (ext);
}

/* FITS_REFFILE -- Check to see if reference file is a fits file
 *
 * i: fname	reference_file_name
 */

static int fits_reffile (char *fname)
{
    char	dir[SZ_FNAME+1], root[SZ_FNAME+1], ext[SZ_FNAME+1];

    breakfname (fname, dir, root, ext, SZ_FNAME);

    /* Determine filetype by checking extension. The extensions 
     * matched must agree with those in the fits kernel
     */

    if (strcmp (ext, ".fits") == 0) {
        return (YES);
    } else if (strcmp (ext, ".fit") == 0) {
        return (YES);
    } else if (ext[3] == 'f' && ext[4] == EOS) {
        return (YES);
    }

    return (NO);
}

/* IMAGE_REFFILE -- Routine to read image header keywords
 *
 * i: im	image descriptor
 * i: name	keyword name
 * o: value	keyword value
 * i: maxch	max value length
 */

static int image_reffile (IRAFPointer im, char *name, char *value, int maxch)
{
    char	buf[SZ_STRING+1], datatype[SZ_STRING+1];
    double	pixval;
    int		pixtype, bitpix, nbytes;

    switch (strmatch (name, EXTRA_KEYS)) {
    case ORDINARY:
	c_imgstr (im, name, value, maxch);

	if (c_iraferr (0) != 0) {
	    value[0] = EOS;
	    return (ERR);
	}
	break;

    case DATAMIN:
	pixval = c_imgmin (im);
	sprintf (buf, "%f", pixval);
	safecopy (value, buf, maxch);
	break;

    case DATAMAX:
	pixval = c_imgmax (im);
	sprintf (buf, "%f", pixval);
	safecopy (value, buf, maxch);
	break;

    case BITPIX:
	pixtype = c_imgtypepix (im);
	bitpix = IRAFSize[pixtype] * NBITS_BYTE;

	sprintf (buf, "%d", bitpix);
	safecopy (value, buf, maxch);
	break;

    case DATATYPE:
	pixtype = c_imgtypepix (im);
	nbytes = IRAFSize[pixtype];

	switch (pixtype) {
	case IRAF_USHORT:
	    safecopy (datatype, "UNSIGNED", SZ_STRING);
	    break;
	case IRAF_SHORT:
	case IRAF_LONG:
	case IRAF_INT:
	    safecopy (datatype, "INTEGER", SZ_STRING);
	    break;
	case IRAF_REAL:
	case IRAF_DOUBLE:
	    safecopy (datatype, "REAL", SZ_STRING);
	    break;
	default:
	    safecopy (datatype, "REAL", SZ_STRING);
	    break;
	}

	sprintf (buf, "%s*%d", datatype, nbytes);
	safecopy (value, buf, maxch);
	break;

    case SIMPLE:
	safecopy (value, "F", maxch);
	break;

    default:
	cdbs_error (BADLIST, "%s", name);
    }

    return (OK);
}

/* RECLOSE_REFFILE -- Close the current row so a new one can be opened
 *
 * u: ref	reference file descriptor
 */

static void reclose_reffile (reffile *ref)
{
    if (ref->ptr == NULL)
        return;

    switch (ref->type) {
    case IMAGE:
        c_imunmap ((IRAFPointer) ref->ptr);
	break;
    case TABLE:
	c_tbtclo ((IRAFPointer) ref->ptr);
	break;
    case LOADFILE:
	if (ref->smudge == YES)
	    write_loadfile (ref->name, (table *) ref->ptr);

	free_table ((table *) ref->ptr);
	ref->smudge = NO;
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    ref->ptr = NULL;
}

/* REOPEN_REFFILE -- Open a new row (group, extension) in a reference file
 *
 * u: ref	reference file descriptor
 * i: row	row number (zero for header)
 */

static int reopen_reffile (reffile *ref, int row)
{
    char	fname[SZ_FNAME+1];

    /* Don't need to open row that is already open */

    if (ref->ptr != NULL && ref->row == row)
        return (OK);

    /* The first row of a file without a primary header 
     * is the same as the zero-th row
     */

    if (ref->primary == NO && ref->row == 0 && row == 1) {
        ref->row = row;
	return (OK);
    }

    /* Don't need to re-open a table or loadfile to access a new row */

    if (ref->ptr != NULL && ref->type != IMAGE && 
	ref->row > 0 && row > 0) {
        ref->row = row;
        return (OK);
    }

    /* Close the reference file if not already closed */

    reclose_reffile (ref);

    /* Add the extension to the file name if it has a primary header */

    if ((ref->primary == YES && row == 0) || ref->type == IMAGE) {
        sprintf (fname, "%s[%d]", ref->name, row);
    } else {
        safecopy (fname, ref->name, SZ_FNAME);
    }

    /* Call the appropriate iraf routine to open the file */

    switch (ref->type) {
    case IMAGE:
	ref->ptr = (void *) c_immap (fname, (IRAFIOMode) ref->mode, 0);
	if (c_iraferr() != 0) {
	    ref->ptr = NULL;
	    return (ERR);
	}
	break;
    case TABLE:
        /* call c_tbtacc to work around problem in c error handling */
        if (c_tbtacc (fname) == NO)
	    return (ERR);

	ref->ptr = (void *) c_tbtopn (fname, (IRAFIOMode) ref->mode, 0);
	if (c_iraferr() != 0) {
	    ref->ptr = NULL;
	    return (ERR);
	}
	break;
    case LOADFILE:
	ref->ptr = (void *) read_loadfile (fname);
	if (ref->ptr == NULL) {
	    return (ERR);
	}
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    /* Update row number in descriptor */

    ref->row = row;
    return (OK);
}

/* ROOT_REFFILE -- Extract root (filename minus extension) from filename)
 *
 * o: root	root name
 * i: fname	file name
 * i: maxch	max length of root name
 */

static void root_reffile (char *root, char *fname, int maxch)
{
    char	*lbrak;

    /* Copy filename over to root name */

    safecopy (root, fname, maxch);

    /* If filename has an extension look for the start and truncate name */

    if (ext_reffile (root) != ERR) {
        lbrak = strrchr (root, '[');
	if (lbrak != NULL)
	    lbrak = EOS;
    }

}
