#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "getobsmode.h"

/* GETOBSMODE -- Get the observation mode for a given obervation
 *
 * i: rootname	observation set name
 * u: index	which obsmode should the function generate?
 * o: obsmode	observation mode
 *
 * The output string should be at least 64 characters long to hold 
 * the observation mode. If no observation is found corresponding 
 * to the rootname, the function returns a value of 0 and the obsmode 
 * string is empty. Some intruments (the wfpc and wfpc2) have more 
 * than one observation mode per rootname.  To support them, index, 
 * the second argument to the function, should contain the number of
 * the obsmode requested. It should range from 1 to 4 for the wfpc 
 * and wfpc2 and should be 1 for the other instruments. The function
 * increments the value of index after each call. It is important that 
 * the first time the function is called for a particular rootname, 
 * index should be should be set to 1. Otherwise, results will be
 * incorrect.
 */

int getobsmode (char *rootname, int *index, char *obsmode) 
{
    char	*root;
    int		ic, jc, nc;

    static char	oldobs[SZ_FIELD+1];

    /* Copy the rootname, remove any extension, (which shouldn't
     * really be there), and convert string to upper case
     */

    jc = -1;
    root = (char *) malloc (strlen (rootname) + 1);

    for (ic = 0; rootname[ic] != '\0'; ic ++) {
	if (rootname[ic] == '.')
	    jc = ic;

	root[ic] = toupper (rootname[ic]);
    }

    if (jc >= 0) {
	root[jc] = '\0';
    } else {
	root[ic] = '\0';
    }

    /* Call the instrument-specific function to construct the
     * observation mode based on the first letter of the rootname
     * which indicates the instrument
     */

    switch (root[0]) {
    case 'N':
	if (*index == 1) {
	    getnicmode (root, obsmode);
	    *index += 1;
	} else {
	    obsmode[0] = '\0';
	}
	break;

    case 'O':
	if (*index == 1) {
	    getstismode (root, obsmode);
	    *index += 1;
	} else {
	    obsmode[0] = '\0';
	}
	break;

    case 'U':
	if (*index == 1) {
	    getwfpc2mode (root, obsmode);
	    strcpy (oldobs, obsmode);
	}

	if (*index <= 4) {
	    sprintf (obsmode, "%s,%d", oldobs, *index);
	    *index += 1;
	} else {
	    obsmode[0] = '\0';
	}
	break;

    case 'V':
	if (*index == 1) {
	    gethspmode (root, obsmode);
	    *index += 1;
	} else {
	    obsmode[0] = '\0';
	}
	break;

    case 'W':
	if (*index == 1) {
	    getwfpcmode (root, obsmode);
	    strcpy (oldobs, obsmode);
	}

	if (*index <= 4) {
	    nc = strlen (oldobs);
	    if (nc < 2) {
		obsmode[0] = '\0';
	    } else if (strncmp (&oldobs[nc-2], "WF", 2) == 0) {
		sprintf (obsmode, "%s,%d", oldobs, *index);
	    } else if (strncmp (&oldobs[nc-2], "PC", 2) == 0) {
		sprintf (obsmode, "%s,%d", oldobs, *index+4);
	    } else {
		obsmode[0] = '\0';
	    }
	    *index += 1;
	} else {
	    obsmode[0] = '\0';
	}
	break;

    case 'X':
	if (*index == 1) {
	    getfocmode (root, obsmode);
	    *index += 1;
	} else {
	    obsmode[0] = '\0';
	}
	break;
    default:
	obsmode[0] = '\0';
    }

    free (root);
    return (obsmode[0] != '\0');
}
