#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "loadfile.h"

/* The functions in this file read and write a loadfile into a table 
 * structure. A table structure is an in-memory structure designed to hold 
 * the contents of a loadfile. The calls used to access a table are 
 * described in a comment at the top of table.c. The calls used to read
 * and write the loadfile are:
 *
 * tab = read_loadfile (fname)			Read a loadfile
 * write_loadfile (fname, tab)			Write a loadfile
 *
 * Please note that the table structure allocated by read_loadfile must 
 * be freed by free_table. Write_loadfile does not free the structure or 
 * modify the table in any way.
 *
 * The loadfile is divided into two parts: a header and body. The body is 
 * further subdivided into rows. Each divison has a terminator. The 
 * termintaor for the heading is as ENDHEADER line. For the body it is an 
 * ENDFILE line. For each row it is an ENDROW line. A terminator is a single
 * word, in either case, starting in the first column, as the sole word on 
 * the line. Comments can be placed anywhere in the loadfile before the 
 * ENDFILE line. Comments are any line whose first non-white character is a 
 * sharp (#). Blank lines are treated as comments, except that they can
 * also follow the ENDFILE line.
 * 
 * Most lines in the loadfile are keyword-value pairs separated by an 
 * equals sign. If found in the header division, they are written to the 
 * table header, if found in the body division, they are written to a 
 * table cell. The COMMENT keyword is handled specially. The COMMENT keyword 
 * can be continued over more than one line and is only terminated by a blank 
 * line, comment or a terminator. Since the COMMENT keyword will otherwise 
 * swallow the other keywords, it should be placed last in the header or in 
 * a row of the body. All rows after the first row in the table body must 
 * contain the same keywords as the first row. It is OK if a keyword is 
 * missing from a row after the first row, but it is an error if a new 
 * keyword is introduced, as this breaks the table structure.
 *
 * All characters are converted to upper case when read into the table and 
 * superfluous blanks are deleted. Superfluous blanks are leading and 
 * trailing blanks and more than one blank between words. So when 
 * retrieving information from the table, specify keyword names and column 
 * names in upper case with only one blank between words.
 */

#define  SZ_ERROR	79
#define	 SZ_BUFFER	80
#define  SZ_KEYWORD	19
#define  NO		0
#define  YES		1

#define  HEADER		1
#define  BODY		2

#define  UNKNOWN	0
#define  LASTLINE	1
#define  BLANK		2
#define  COMMENT	3
#define  KEYWORD	4
#define  ENDROW		5
#define  ENDHEADER	6
#define  ENDFILE	7

#define  TOK_MAXLEN	10		/* max length of any reserved word */

typedef struct {
    FILE	*file;
    char	*buffer;
    int		token;
    int		line;
    int		row;
    int		comment;
} parser;

char	*lod_error = NULL;

char	*load_message[] = {
    "Cannot open loadfile",			/* 0 */
    "Syntax error in loadile"			/* 1 */
};

static void 	error_load_parser (parser *);
static void 	free_load_parser (parser *);
static parser * init_load_parser (char *);
static int	put_load_info (parser *, table *, int);
static int	which_load_token (parser *);

/* READ_LOADFILE -- Read a loadfile into a table structure
 *
 * i: fname	loadfile name
 */

table * read_loadfile (char *fname)
{
    table	*tab;
    parser	*load;

    /* Initialize information used by the parser */

    load = init_load_parser (fname);
    if (load == NULL)
	return (NULL);

    /* Create an empty table */

    tab = create_table ();
    /* Read info up to the ENDHEADER card into the table keywords */

    load->comment = NO;
    while (which_load_token (load) != ENDHEADER) {
	switch (load->token) {
	case BLANK:
	case COMMENT:
	    break;
	case KEYWORD:
	    if (put_load_info (load, tab, HEADER) == BAD) {
		error_load_parser (load);
		free_table (tab);
		return (NULL);
	    }
	    break;
	default:
	    error_load_parser (load);
	    free_table (tab);
	    return (NULL);
	}
    }

    /* Put info up to ENDFILE card into the table cells */

    load->row = 0;
    load->comment = NO;
    while (which_load_token (load) != ENDFILE) {
	switch (load->token) {
	case BLANK:
	case COMMENT:
	    break;
	case KEYWORD:
	    if (put_load_info (load, tab, BODY) == BAD) {
		error_load_parser (load);
		free_table (tab);
		return (NULL);
	    }
	    break;
	case ENDROW:
	    load->comment = NO;
	    load->row ++;
	    break;
	default:
	    error_load_parser (load);
	    free_table (tab);
	    return (NULL);
	}
    }

    /* Only blank lines are allowed after ENDFILE */

    while (which_load_token (load) == BLANK)
	;

    if (load->token != LASTLINE) {
	error_load_parser (load);
	free_table (tab);
	return (NULL);
    }

    free_load_parser (load);
    return (tab);
}

/* WRITE_LOADFILE -- Write a loadfile from the info in a table
 *
 * i: fname	loadfile name
 * i: tab	pointer to table
 */

int write_loadfile (char *fname, table *tab)
{
    FILE	*file;
    textbuf	*text;
    int		ikey, icol, comment, row, line, nc;
    char	*name, *value;

    /* Open the output file */

    file = fopen (fname, "w");
    if (file == NULL) {
	lod_error = load_message[0];
	return (BAD);
    }

    /* Write the table header keywords */

    comment = BAD;
    for (ikey = 0; ikey < nkey_table(tab); ikey ++) {
	name = get_tab_keyname_bynum (tab, ikey);
	value = get_tab_keyval_bynum (tab, ikey);

	if (strcmp (name, "COMMENT") == 0) {
	    comment = ikey;
	} else {
	    fprintf (file, "%s = %s\n", name, value);
	}
    }

    /* Write the comment */

    if (comment != BAD) {
	value = get_tab_keyval_bynum (tab, comment);
	fprintf (file, "COMMENT = %s", value);

	nc = strlen (value);
	if (nc == 0 || value[nc-1] != '\n')
	    fprintf (file, "\n");
    }

    fprintf (file, "ENDHEADER\n\n");

    /* Write the table columns */

    comment = get_tab_colnum (tab, "COMMENT");
    for (row = 0; row < nrow_table(tab); row ++) {
	for (icol = 0; icol < ncol_table(tab); icol ++) {
	    name = get_tab_colname (tab, icol);
	    value = get_tab_cell_bynum (tab, icol, row);

	    if (icol != comment) {
		fprintf (file, "%s = %s\n", name, value);
	    }
	}

	/* Write the row level comment */

	if (comment != BAD) {
	    value = get_tab_cell_bynum (tab, comment, row);
	    fprintf (file, "COMMENT = %s", value);

	    nc = strlen (value);
	    if (nc == 0 || value[nc-1] != '\n')
	        fprintf (file, "\n");
	}

	/* Write the end of row marker */

	fprintf (file, "ENDROW\n\n");
    }

    fprintf (file, "ENDFILE\n");
    fclose (file);

    return (GOOD);
}

/*----------------------------------------------------------------------
 *
 * Procedures below this line are not externally callable
 *
 *----------------------------------------------------------------------
 */

/* ERROR_LOAD_PARSER -- Syntax error handling routine
 *
 * u: load	parsing structure
 *
 */

static void error_load_parser (parser * load)
{
    static char	syntax_message[SZ_ERROR+1];

    sprintf (syntax_message, "%s at line %d", load_message[1], load->line);
    lod_error = syntax_message;

    free_load_parser (load);
}

/* FREE_LOAD_PARSER -- Free the loadfile parser information
 *
 * u: load 	parsing structure
 */

static void free_load_parser (parser * load)
{
    if (load->file != NULL)
	fclose (load->file);

    free (load->buffer);
    free (load);
}

/* INIT_LOAD_PARSER -- Initialize the loadfile parser information
 *
 * i: fname	loadfile name
 */

static parser * init_load_parser (char * fname)
{
    parser 	*load;

    load = (parser *) malloc (sizeof (parser));

    load->token = UNKNOWN;
    load->line = 0;
    load->row = 0;
    load->comment = NO;
    load->buffer = (char *) malloc (SZ_BUFFER+1);

    load->file =  fopen (fname, "r");
    if (load->file == NULL) {
	lod_error = load_message[0];
	free_load_parser (load);
	load = NULL;
    }

    return (load);
}

/* PUT_LOAD_INFO -- Put a value from the loadfile in a table
 *
 * u: load	parser information
 * u: table	pointer to table
 * i: section	section of the table to add info to (header or body)
 */

static int put_load_info (parser *load, table *tab, int section)
{
    int		status;
    char	*eq, *name, *value;
    static char	comment[] = "COMMENT";


    if (load->comment == NO) {
	eq = strchr (load->buffer, '=');
	if (eq == NULL)
	    return (BAD);
	
	*eq = '\0';
	
	name = load->buffer;
	value = eq + 1;

	strclean (name);
	strucase (name);

	if (strcmp (name, comment) == 0) {
	    load->comment = YES;
	} else {
	    strclean (value);
	}

	if (section == HEADER) {
	    status = put_tab_key (tab, name, value);
	} else if (section == BODY) {
	    status = put_tab_cell (tab, name, load->row, value);
	}

    } else {
	value = load->buffer;

	if (section == HEADER) {
	    status = cat_tab_key (tab, comment, value);
	} else if (section == BODY) {
	    status = cat_tab_cell (tab, comment, load->row, value);
	}
    }

    return (status);
}

/* WHICH_LOAD_TOKEN -- Read line and determine token type
 *
 * u: load	parsing information
 */

static int which_load_token (parser *load)
{
    char		keyword[SZ_KEYWORD+1];
    int			ic, jc, kc, nc;

    if (fgets (load->buffer, SZ_BUFFER, load->file) == NULL) {
	load->token = LASTLINE;

    } else {
	load->line ++;

	/* Look for: first non-blank character (jc)
	 *           last non-blank character  (kc)
	 *	     newline character         (nc)
	 */

	jc = kc = nc = -1;
	for (ic = 0; load->buffer[ic] != '\0'; ic ++) {
	    if (load->buffer[ic] == '\n') {
	        nc = ic;

	    } else if (load->buffer[ic] > ' ') {
	        kc = ic;
		if (jc < 0)
		    jc = ic;
	    }
	}

	/* Preliminary determination of token type */

	if (nc < 0) {
	    /* Line too long: newline not found */
	    load->token = UNKNOWN;

	} else if (load->comment) {
	    /* All lines following a comment are keywords 
	     * until an ENDHEADER or ENDROW is seen 
	     */
	    load->token = KEYWORD;

	} else if (jc < 0) {
	    /* BLANK card: all whitespace */
	    load->token = BLANK;

	} else if (load->buffer[jc] == '#') {
	    /* COMMENT card: sharp is first nonwhite char on line */
	    load->token = COMMENT;

	} else {
	    load->token = KEYWORD;
	}

	/* Check for reserved keywords */

	if (load->token == KEYWORD) {
	    /* Convert keyword to upper case */

	    for (ic = 0; ic <= SZ_KEYWORD && load->buffer[ic] != '\0'; ic ++)
		keyword[ic] = toupper (load->buffer[ic]);

	    if (kc+1 < ic) {
	        keyword[kc+1] = '\0';
	    } else {
	        keyword[ic] = '\0';
	    }

	    /* Reserved keywords have different token types */

	    if (strcmp (keyword, "ENDROW") == 0) {
		load->token = ENDROW;
	    } else if (strcmp (keyword, "ENDHEADER") == 0) {
		load->token = ENDHEADER;
	    } else if (strcmp (keyword, "ENDFILE") == 0) {
		load->token = ENDFILE;
	    } else {
		load->token = KEYWORD;
	    }
	}
    }

    return (load->token);
}
