#include <string.h>
#include <stdlib.h>
#include "config.h"
#include "cdbslib.h"
#include "loadfile.h"
#include "hashtab.h"

static void	key_hash (hashtab *, textbuf *, table *, int, 
			  unsigned int *, unsigned int *);
static int	match_hash (textbuf *, table *, int, int);

/* FIND_HASH -- Find a value in a hash table
 *
 * u: hash	hash table
 * i: obsmode	list of observation mode fields
 * i: tab	load table
 * i: row	row number of value to match
 */

int find_hash (hashtab *hash, textbuf *obsmode, table *tab, int row)
{
    unsigned int	key, step;
    int			idx;

    /* Calculate hash key and step size */

    key_hash (hash, obsmode, tab, row, &key, &step);

    /* Loop until an empty spot is found in the hash table
     * or until a row is found which matches the input row
     */

    for (idx = key % hash->size; YES; idx = (idx + step) % hash->size) {
	if (hash->val[idx] == -1) {
	    /* Found an empty spot. This means there is no
	     * match in the hash table. Fill the spot with 
	     * the current value.
	     */

	    hash->key[idx] = key;
	    hash->val[idx] = row;
	    return (NO);

	} else if (hash->key[idx] == key) {
	    /* If the keys match, we must do a complete match
	     * to rule out the possibility of key collision.
	     */

	    if (match_hash (obsmode, tab, row, hash->val[idx]) == YES)
		return (YES);
	}
    }
}

/* FREE_HASH -- Free memory allocated for hash table 
 *
 * i: hash	hash table
 */

void free_hash (hashtab *hash)
{
    free (hash->key);
    free (hash->val);
    free (hash);
}

/* INIT_HASH -- Allocate and initialize hash table
 *
 * i: nval	max number of hash table values
 */

hashtab *init_hash (int nval)
{
    int		idx;
    hashtab	*hash;

    /* Create a hash table whose size is a power of two
     * It should be at least twice the size as the max number of values
     */

    hash = (hashtab *) malloc (sizeof (hashtab));

    for (hash->size = 1; hash->size < 2 * nval; hash->size *= 2)
	;

    hash->key = (unsigned int *) malloc (hash->size * sizeof (unsigned int));
    hash->val = (int *) malloc (hash->size * sizeof (int));

    for (idx = 0; idx < hash->size; idx ++) {
	hash->key[idx] = 0;
	hash->val[idx] = -1;
    }

    return (hash);
}

/* KEY_HASH -- Compute hash table key and step size
 *
 * i: hash	hash table
 * i: obsmode	list of observation mode fields
 * i: tab	load table
 * i: row	row number in load table
 * o: key	hash table key
 * o: step	hash table step size
 */

static void key_hash (hashtab *hash, textbuf *obsmode, table *tab, int row, 
		      unsigned int *key, unsigned int *step)
{
    int		icol, ic;
    char	*keyword, *value, *str;

    *key = 0;
    *step = 0;

    /* The obsmode values in the row are concatenated to form the key */

    for (icol = 0; keyword = get_text(obsmode, icol); icol ++) {
	value = get_tab_cell (tab, keyword, row);

	if (value != NULL) {
	    str = nospaces (value);
	    for (ic = 0; str[ic] != EOS; ic ++) {
		*key = (*key * 2) + str[ic];
		*step = *step + str[ic];
	    }
	}
    }

    /* Ensure that step size is an odd number so we 
     * hit all spots in the hash table.
     */

    *step = *step - (*step % 2) + 1;
}

/* MATCH_HASH -- Detemine if two rows have matching obsmode values
 *
 * i: obsmode	list of observation mode fields
 * i: table	load table
 * i: irow	first row to match
 * i: jrow	second row to match
 */

static int match_hash (textbuf *obsmode, table *tab, int irow, int jrow)
{
    int		icol;
    char	*keyword, *istr, *jstr;

    for (icol = 0; keyword = get_text (obsmode, icol); icol ++) {
	istr = get_tab_cell (tab, keyword, irow);
	jstr = get_tab_cell (tab, keyword, jrow);

	if (istr != NULL && jstr != NULL) {
	    if (strcmp (istr, jstr) != 0) 
		return (NO);
	}
    }

    return (YES);
}
