#include <string.h>
#include <stdlib.h>
#include "loadfile.h"

/* A text structure consists of two arrays. The string array holds lines of
 * text and the line array hold the index to the first char in each line
 * of text. Since the lines of text are packed, the only way to change a
 * line is to write the changed line at the end of the string array. The
 * arrays are implemented by buffer structures, which are unbounded arrays.
 */

static char *text_message[] = {
    "Text line index out of bounds",		/* 0 */
    "cat_text: internal error"			/* 1 */
};

/* ADD_TEXT -- Store a new string in a text buffer
 *
 * u: text	pointer to text buffer
 * i: str	string to add to buffer
 */

int add_text (textbuf *text, char * str)
{
    int		index, to;

    index = put_char_buf (&text->string, str);
    to = put_int_buf (&text->line, index);
    text->nline = to + 1;

    return (to);
}

/* CAT_TEXT -- Concatenate a string to an existing line in the text buffer
 *
 * u: text	pointer to text buffer
 * i: to	line to concatenate to
 * i: str	string to concatenate
 */

int cat_text (textbuf *text, int to, char *str)
{
    int		index, top, len;
    char	*str1, *str2;

    if (to < 0 || to >= text->nline) {
	lod_error = text_message[0];
	return (BAD);
    }

    /* If the line we are concatenating to is not at the end of the char
     * buffer, we cannot concatenate to it. So we must rewrite the string 
     * in the buffer to ensure it is at the end of the buffer. We make a
     * temporary copy of the string because a put may cause the char buffer 
     * to be reallocated, causing all old pointers to be invalid.
     */

    index = text->line.info[to];
    top = top_char_buf (&text->string, index);

    if (top == BAD) {
	str1 = get_text (text, to);

	len = strlen (str1) + 1;
	str2 = (char *) malloc (len);
	strcpy (str2, str1);

	put_text (text, to, str2);
	free (str2);
    }

    /* We backspace over the null char at the end of the line and 
     * then copy the new string into the buffer. Since the two 
     * strings are contiguous , this concatenates them. 
     */

    text->string.next --;
    put_char_buf (&text->string, str);

    return (to);
}

/* COPY_TEXT -- Create a copy of a string in a text buffer
 *
 * u: text	pointer to text buffer
 * i: from	line to copy from
 * i: to	line to copy to
 */

int copy_text (textbuf *text, int from, int to) 
{
    int 	index;

    if (from < 0 || from >= text->nline) {
	lod_error = text_message[0];
	return (BAD);
    }

    if (to < 0 || to > text->nline) {
	lod_error = text_message[0];
	return (BAD);
    }

    index = text->line.info[from];

    if (to == text->nline) {
	put_int_buf (&text->line, index);
	text->nline ++;
    } else {
	text->line.info[to] = index;
    }

    return (to);
}

/* FIND_TEXT -- Search for a string in the text buffer
 *
 * i: text	pointer to text buffer
 * i: str	string to search for
 */

int find_text (textbuf *text, char *str)
{
    int		line, index;

    for (line = 0; line < text->nline; line ++) {
	index = text->line.info[line];

	if (strcmp (str, &text->string.info[index]) == 0)
	    return (line);
    }

    return (BAD);
}

/* FREE_TEXT -- Free memory used by a text buffer
 *
 * u: text	pointer to text buffer
 */

void free_text (textbuf *text)
{
    free (text->string.info);
    free (text->line.info);
}

/* GET_TEXT -- Get a string from a text buffer
 *
 * i: text	pointer to text buffer
 * i: from	line to get from
 */
char *get_text (textbuf *text, int from)
{
    int		index;

    if (from < 0 || from >= text->nline) {
	lod_error = text_message[0];
	return (NULL);
    }

    index = text->line.info[from];
    return (&text->string.info[index]);
}


/* INIT_TEXT -- Initialize a text buffer
 *
 * u: text	pointer to text buffer
 * i: line_size	size of line buffer
 * i: str_size	size of string buffer
 */

void init_text (textbuf *text, int line_size, int str_size) 
{
    init_int_buf (&text->line, line_size);
    init_char_buf (&text->string, str_size);
    text->nline = 0;
}

/* PUT_TEXT -- Put a string in a line in the text buffer
 *
 * u: text	pointer to text buffer
 * i: to	line to put to
 * i: str	string to add to buffer
 */

int put_text (textbuf *text, int to, char *str)
{
    int		index;

    if (to < 0 || to > text->nline) {
	lod_error = text_message[0];
	return (BAD);
    }

    if (to == text->nline)
	return (add_text (text, str));

    index = put_char_buf (&text->string, str);
    text->line.info[to] = index;

    return (to);
}
